 
#include "bess.h"
#include "bessstubs.h"
#include "bessproto.h"
#include "bess.pt"

#define DYN_CONFIG(arg) arg

#define REPERR(st, err)                                                 \
  {                                                                     \
    if (st->error_notify) {                                             \
      char *_ebuf_ = malloc (strlen ("Error: ") + strlen (err));        \
      sprintf ( _ebuf_, "Error: %s", err);                              \
      (*st->error_notify) (st, __FILE__, __LINE__, _ebuf_);             \
      free (_ebuf_);                                                    \
    } else {                                                            \
      fprintf (stderr, "\nError (%s, l.%d): %s\n",                      \
               __FILE__, __LINE__, err);                                \
    }                                                                   \
  }

#include "listwrap.h"

#define GENWRAP_TYPE   TVP
#define GENWRAP_PREFIX tv
#include "listwrap.c"
#define GENWRAP_TYPE   TEP
#define GENWRAP_PREFIX te
#include "listwrap.c"
#define GENWRAP_TYPE   TLP
#define GENWRAP_PREFIX tl
#include "listwrap.c"
#define GENWRAP_TYPE   TFP
#define GENWRAP_PREFIX tf
#include "listwrap.c"
#define GENWRAP_TYPE   TFUP
#define GENWRAP_PREFIX tfu
#include "listwrap.c"
#define GENWRAP_TYPE   TEUP
#define GENWRAP_PREFIX teu
#include "listwrap.c"
#define GENWRAP_TYPE   TSP
#define GENWRAP_PREFIX ts
#include "listwrap.c"
#define GENWRAP_TYPE   TRP
#define GENWRAP_PREFIX tr
#include "listwrap.c"
#define GENWRAP_TYPE   MFRP
#define GENWRAP_PREFIX mfr
#include "listwrap.c"
#define GENWRAP_TYPE   MERP
#define GENWRAP_PREFIX mer
#include "listwrap.c"



typedef struct BESS {
  bess_error_notify_func      error_notify;
  MESHP                       mesh;
  GENWRAP_LIST_TYPE (MFRP)    mfrl;
  GENWRAP_LIST_TYPE (MERP)    merl;  
  double                      crease_ang;
  double                      corner_ang;
  GENWRAP_LIST_TYPE (TVP)     tvl;
  GENWRAP_LIST_TYPE (TEP)     tel;
  GENWRAP_LIST_TYPE (TLP)     tll;
  GENWRAP_LIST_TYPE (TFP)     tfl;
  GENWRAP_LIST_TYPE (TSP)     tsl;
  GENWRAP_LIST_TYPE (TRP)     trl;
}              BESS;

typedef struct MFR {
  MFP               mf;
  struct TF        *tf;
}              MFR;

#define ME_IS_MANIFOLD(mer)      (mer->num_inc_t == 2)
#define ME_IS_SHEET(mer)         (mer->num_inc_t == 1)
#define ME_IS_NON_MANIFOLD(mer)  (mer->num_inc_t > 2)

typedef struct MER {
  MEP                     me;
  struct TE              *te;
  int                     bounds_tf;
  int                     num_inc_t;
}              MER;

typedef struct TR {
  int                           id;
  GENWRAP_LIST_TYPE (TSP)       tsl;
  MRID                          mrid;
}              TR;



typedef struct TS {
  int                           id;
  GENWRAP_LIST_TYPE (TFUP)      tful; /* using */
  TRP                           tr; /* used by  */
}              TS;

typedef struct TF {
  int                           id;
  GENWRAP_LIST_TYPE (TLP)       tll;
  GENWRAP_LIST_TYPE (TFUP)      tful;
  GENWRAP_LIST_TYPE (MFRP)      mfrl;
  GENWRAP_LIST_TYPE (MERP)      merl;
}              TF;

typedef struct TFU {
  int                    id;
  struct TF             *tf; /* references f */
  TOE                    o; /* orientation of f */
  struct TS             *ts; /* owning shell */
}              TFU;

typedef struct TV {
  int                      id;
  MVP                      mv;
  GENWRAP_LIST_TYPE (TEP)  tel;
}              TV;

typedef struct TE {
  int                        id;
  TVP                        tv[2];
  GENWRAP_LIST_TYPE (TEUP)   teul;  /* list of edgeuses */
  GENWRAP_LIST_TYPE (MERP)   merl;  /* list of mesh edges */
  int                        num_inc_f;
}              TE;


typedef struct TEU {
  int                    id;
  struct TE             *te; /* references e */
  TOE                    o; /* orientation of e */
  struct TL             *tl; /* owning loop */
}              TEU;

typedef struct TL {
  int                        id;
  GENWRAP_LIST_TYPE (TEUP)   teul; /* list of edgeuses */
  struct TF                 *tf;  /* owning face */
}              TL;


#define GENID(id) (++id)

static TR *
new_tr (MRID mrid)
{
  TR *r;
  static int id = 0;
  
  r = NEWP_OBJ (TR);
  if (!r) return NULL;

  r->id   = GENID (id);
  r->tsl  = tsp_l_make ();
  r->mrid = mrid;

  return r;
}

static TS *
new_ts (TRP tr)
{
  TS *s;
  static int id = 0;
  
  s = NEWP_OBJ (TS);
  if (!s) return NULL;

  s->id    = GENID (id);
  s->tful  = tfup_l_make ();
  s->tr    = tr;

  return s;
}



static TL *
new_tl (TF *f)
{
  TL *l;
  static int id = 0;

  l = NEWP_OBJ (TL);
  if (!l) return NULL;

  l->id       = GENID (id);
  l->teul     = teup_l_make ();
  l->tf       = f;
  tlp_l_add (l->tf->tll, l);

  return l;
}

static TF *
new_tf (void)
{
  TF *f;
  static int id = 0;
  
  f = NEWP_OBJ (TF);
  if (!f) return NULL;

  f->id    = GENID (id);
  f->tll   = tlp_l_make ();
  f->tful  = tfup_l_make ();
  f->mfrl  = mfrp_l_make ();
  f->merl  = merp_l_make ();

  return f;
}

static TFU *
new_tfu (TF *f, TOE o, TS *s)
{
  TFU *fu;
  static int id = 0;
  
  fu = NEWP_OBJ (TFU);
  if (!fu) return NULL;

  fu->id   = GENID (id);
  fu->tf   = f;
  fu->o    = o;
  fu->ts   = s;

  return fu;
}


static TV *
new_tv (MVP mv)
{
  TV *v;
  static int id = 0;

  v = NEWP_OBJ (TV);
  if (!v) return NULL;

  v->id  = GENID (id);
  v->mv  = mv;
  BESS_STUB_MV_SET_AP (v->mv, v);
  v->tel = tep_l_make ();

  return v;
}

static TE *
new_te (void)
{
  TE *e;
  static int id = 0;

  e = NEWP_OBJ (TE);
  if (!e) return NULL;

  e->id          = GENID (id);
  e->tv[0]       = 
  e->tv[1]       = NULL;
  e->teul        = teup_l_make ();
  e->merl        = merp_l_make ();
  e->num_inc_f   = 0;

  return e;
}

static TEU *
new_teu (TE *e, TOE o)
{
  TEU *eu;
  static int id = 0;

  eu = NEWP_OBJ (TEU);
  if (!eu) return NULL;

  eu->id      = GENID (id);
  eu->te      = e;
  eu->o       = o;
  eu->tl      = NULL;

  return eu;
}

static MFR *
new_mfr (MFP t)
{
  MFR *mfr;

  mfr = NEWP_OBJ (MFR);
  if (!mfr) return NULL;

  mfr->mf  = t;
  mfr->tf  = NULL;
  return mfr;
}

static MER *
new_mer (MEP e)
{
  MER *er;
  er = NEWP_OBJ (MER);
  if (!er) return NULL;
  er->me          = e;
  er->te          = NULL;
  er->bounds_tf   = FALSE;
  er->num_inc_t   = 0;

  return er;
}




static void
get_mfrs (BESSP bess)
{
  MFP  t;
  MFR *mfr;
  t = NULL;
  while ((t = BESS_STUB_NEXT_MF (bess->mesh, t))) {
    mfr = new_mfr (t);
    mfrp_l_add (bess->mfrl, mfr);
    BESS_STUB_MF_SET_AP (t, mfr);
  }
}


static void
get_mers (BESSP bess)
{
  MEP  e;
  MER *er;
  e = NULL;
  while ((e = BESS_STUB_NEXT_ME (bess->mesh, e))) {
    er = new_mer (e);
    merp_l_add (bess->merl, er);
    BESS_STUB_ME_SET_AP (e, er);
  }
}


static void
failed_to_retrieve_normal (BESSP bess, MFP mf)
{
  char buf[1024];
  MVP mv[6];
  LOC l[6];
  int supplied = BESS_STUB_MF_GET_MV (mf, mv);
  int i;
  FOR (i, supplied) {
    BESS_STUB_MV_GET_LOC (mv[i], &l[i]);
  }
  sprintf (buf,
           "Failed to retrieve normal: {%g,%g,%g},{%g,%g,%g},{%g,%g,%g}",
           l[0].x, l[0].y, l[0].z,
           l[1].x, l[1].y, l[1].z,
           l[2].x, l[2].y, l[2].z);
  REPERR (bess, buf);
}

static int
acceptable_crease_angle (BESSP bess, MFP t, MFP ot, MEP me, double ada)
{
  LOC n0, n1;
  double cada = cos (ada);
  double ndot;
  double txi[3], teta[3];
  double otxi[3], oteta[3];
  MEP tme[3];
  MEP otme[3];
  int i;

  if (ada >= 2*M_PI) return TRUE; /* if angle >= 360 deg, this check is disabled */

  BESS_STUB_MF_GET_ME (t, tme);
  if        (me == tme[0]) {
    txi[0]  = 1.0; txi[1]  = 0.5; txi[0]  = 0.0;
    teta[0] = 0.0; teta[1] = 0.0; teta[2] = 0.0;
  } else if (me == tme[1]) {
    txi[0]  = 1.0; txi[1]  = 0.5; txi[2]  = 0.0;
    teta[0] = 0.0; teta[1] = 0.5; teta[2] = 1.0;
  } else {
    txi[0]  = 0.0; txi[1]  = 0.0; txi[0]  = 0.0;
    teta[0] = 1.0; teta[1] = 0.5; teta[2] = 0.0;
  }
  BESS_STUB_MF_GET_ME (ot, otme);
  if        (me == otme[0]) {
    otxi[0]  = 1.0; otxi[1]  = 0.5; otxi[0]  = 0.0;
    oteta[0] = 0.0; oteta[1] = 0.0; oteta[2] = 0.0;
  } else if (me == otme[1]) {
    otxi[0]  = 1.0; otxi[1]  = 0.5; otxi[2]  = 0.0;
    oteta[0] = 0.0; oteta[1] = 0.5; oteta[2] = 1.0;
  } else {
    otxi[0]  = 0.0; otxi[1]  = 0.0; otxi[0]  = 0.0;
    oteta[0] = 1.0; oteta[1] = 0.5; oteta[2] = 0.0;
  }
  ndot = 0;
  FOR (i, 3) {
    if (!BESS_STUB_MF_GET_NORMAL (t, txi[i], teta[i], &n0)) {
      failed_to_retrieve_normal (bess, t);
      return FALSE;
    }
    if (!BESS_STUB_MF_GET_NORMAL (ot, otxi[i], oteta[i], &n1)) {
      failed_to_retrieve_normal (bess, ot);
      return FALSE;
    }
    ndot += VECMAC_DOT (n0, n1);
  }
  return (ndot/3 > cada);
}

static MVP
other_v (MVP amv, MVP mv[2])
{
  if      (amv == mv[0]) return mv[1];
  else if (amv == mv[1]) return mv[0];
  else                   return NULL;
}

static int
acceptable_corner_angle (MEP me, MEP nme, MVP mv, double ada)
{
  double cada = cos (ada);
  double ndot;
  LOC d0, d1;
  MVP emv[3], nemv[3];

  BESS_STUB_ME_GET_MV (me, emv);
  BESS_STUB_ME_GET_MV (nme, nemv);
  
  BESS_STUB_ME_GET_TANGENT (me, mv, &d0);
  if (mv == emv[0]) {
    VECMAC_SCALAR (d0, *=, -1);
  }
  BESS_STUB_ME_GET_TANGENT (nme, mv, &d1);
  if (mv == nemv[1]) {
    VECMAC_SCALAR (d1, *=, -1);
  }

  VECMAC_NORMALIZE (d0);
  VECMAC_NORMALIZE (d1);
  ndot = VECMAC_DOT (d0, d1);
  return (ndot > cada);
}


static void
attach_f (BESSP bess, TFP f)
{
  tfp_l_add (bess->tfl, f);
}

static void
attach_r (BESSP bess, TRP r)
{
  trp_l_add (bess->trl, r);
}

static void
attach_s (BESSP bess, TSP s)
{
  tsp_l_add (bess->tsl, s);
}

static void
attach_l (BESSP bess, TLP l)
{
  tlp_l_add (bess->tll, l);
}

static void
attach_e (BESSP bess, TEP e)
{
  tep_l_add (bess->tel, e);
}

static void
attach_v (BESSP bess, TVP v)
{
  tvp_l_add (bess->tvl, v);
}

/*
   Expects a manifold edge!
 */
static MFP
other_mf (MEP me, MFP mf)
{
  MFP amf;
  amf = NULL;
  while ((amf = BESS_STUB_ME_GET_NEXT_MF (me, amf))) {
    if (amf != mf) return amf;
  }
  return NULL;
}


static TVP
mv_get_tv (MVP mv) 
{
  return (TVP) BESS_STUB_MV_GET_AP (mv); 
}

static MERP
me_get_mer (MEP me)
{
  return (MERP) BESS_STUB_ME_GET_AP (me);
}

static MFRP
mf_get_mfr (MFP mf)
{
  return (MFRP) BESS_STUB_MF_GET_AP (mf);
}

static void
start_f (BESSP bess, MFR *smfr)
{
  TF *tf;
  static STACK mfrs = NULL;
  MFR *mfr, *omfr;
  int i;
  MEP me[3];
  MER *mer;
  MFP omf;
  
  tf = new_tf ();
  attach_f (bess, tf);
  if (!mfrs) mfrs = make_stack ();
  if (!stack_empty (mfrs)) {
    REPERR (bess, "stack is not empty");
  }

  push_stack (mfrs, smfr);
  while (!stack_empty (mfrs)) {
    void *surfm;
    mfr = (MFR *) pop_stack (mfrs);
    if (!mfr->tf) { /* not yet assigned to a face */
      mfr->tf = tf;
      mfrp_l_add (tf->mfrl, mfr);
    }
    surfm = BESS_STUB_MF_GET_SURF (mfr->mf);
    BESS_STUB_MF_GET_ME (mfr->mf, me);
    FOR (i, 3) {
      mer = me_get_mer (me[i]);
      mer->num_inc_t = BESS_STUB_ME_GET_NUM_INC_MF (me[i]);
      if (ME_IS_MANIFOLD (mer)) {
        omf = other_mf (me[i], mfr->mf);
        omfr = mf_get_mfr (omf);
        if (!omfr->tf) { /* not yet assigned to any face */
          void *surfom;
          surfom = BESS_STUB_MF_GET_SURF (omfr->mf);
          if (acceptable_crease_angle (bess,
                                       mfr->mf, omf, mer->me,
                                       bess->crease_ang) &&
              ((surfm == NULL || surfom == NULL) || /* no suggestion for either */
               (surfom == surfm)  /* or both have the same surface tag */
               )
              ) {
            push_stack (mfrs, omfr);
          } else {
            mer->bounds_tf = TRUE;
            merp_l_add (tf->merl, mer);
          }
        } else if (omfr->tf != tf) {
          mer->bounds_tf = TRUE;
          merp_l_add (tf->merl, mer);
        }
      } else {
        mer->bounds_tf = TRUE;
        merp_l_add (tf->merl, mer);
      }
    } /* FOR (i, 3) */
  } /* while */
}

static void
extract_faces (BESSP bess)
{
  int all_done;
  MFR *tr;
  DYN_CONFIG (static int doit = TRUE;)
  DYN_CONFIG (if (!doit) return;)
  
  get_mfrs (bess);
  get_mers (bess);

  while (1) {
    all_done = TRUE;
    tr = NULL;
    while ((tr = (MFR *) mfrp_l_next (bess->mfrl, tr))) {
      if (!tr->tf) {
        start_f (bess, tr);
        all_done = FALSE;
      }
    }
    if (all_done)
      return;
  }
}

static void
v_set_e (TVP tv, TEP te)
{
  tep_l_add (tv->tel, te);
}

static void
e_set_v (TEP te, TVP tv, BESSP bess)
{
  if (te->tv[0] && te->tv[1]) {
    if ((te->tv[0] != tv) && (te->tv[1] != tv)) {
      REPERR (bess, "setting tv to a full te");
    }
  } else {
    if      (!te->tv[0]) te->tv[0] = tv;
    else                 te->tv[1] = tv;
    v_set_e (tv, te);
  }
}


static void
e_via_v (MVP mv, BESSP bess, MERP mer, STACK mers)
{
  TVP tv;
  MEP me;
  MERP omer, cmer = NULL;
  int num_bounding_at_v = 0;

  tv = mv_get_tv (mv);
  me = NULL;
  while ((me = BESS_STUB_MV_GET_NEXT_ME (mv, me))) {
    omer = me_get_mer (me);
    if (omer->bounds_tf && omer != mer) { /* do not consider self */
      cmer = omer; /* possible continuation */
      num_bounding_at_v++;
    }
  }
  if (num_bounding_at_v == 0) { /* wire edge */
    if (!tv) { tv = new_tv (mv); attach_v (bess, tv); }
    e_set_v (mer->te, tv, bess);
  } else if (num_bounding_at_v == 1) {  /* possible continuation */
    if (!cmer->te) { /* not yet assigned to any te */
      if (mer->num_inc_t != cmer->num_inc_t) { /* change in adjacent faces; split */
        if (!tv) { tv = new_tv (mv); attach_v (bess, tv); }
        e_set_v (mer->te, tv, bess);
      } else {
        if (acceptable_corner_angle (mer->me, cmer->me, mv, bess->corner_ang)) {
          push_stack (mers, cmer);
        } else {
          if (!tv) { tv = new_tv (mv); attach_v (bess, tv); }
          e_set_v (mer->te, tv, bess);
        }
      }
    } else {
      if (cmer->te != mer->te) { /* different te's */
        if (!tv) { tv = new_tv (mv); attach_v (bess, tv); }
        e_set_v (mer->te, tv, bess);
      } else {                  /* the same te */
        if (tv) {/* cycle edge */
          e_set_v (mer->te, tv, bess);
        }
      }
    }
  } else { /* branching vertex */
    if (!tv) { tv = new_tv (mv); attach_v (bess, tv); }
    e_set_v (mer->te, tv, bess);
  }
}

static void
start_e (BESSP bess, MERP smer)
{
  MERP mer;
  TEP te;
  MVP mv[3];
  static STACK mers = NULL;
  int i;
  
  te = new_te ();
  attach_e (bess, te);
  if (!mers) mers = make_stack ();
  if (!stack_empty (mers)) {
    REPERR (bess, "stack is not empty");
  }

  push_stack (mers, smer);
  while (!stack_empty (mers)) {
    mer = (MERP) pop_stack (mers);
    mer->te = te; merp_l_add (te->merl, mer);
    BESS_STUB_ME_GET_MV (mer->me, mv);
    FOR (i, 2) {
      e_via_v (mv[i], bess, mer, mers);
    } /* FOR (i, 3) */
  } /* while */
}

static void
extr_es (TFP tf, BESSP bess) 
{
  MERP mer;
  int all_done;

  while (1) {
    all_done = TRUE;
    mer = NULL;
    while ((mer = merp_l_next (tf->merl, mer))) {
      if (!mer->te) {
        start_e (bess, mer);
        all_done = FALSE;
      }
    }
    if (all_done)
      return;
  }
}




static MERP
next_mer (MERP mer,
          MVP common_mv,
          TEP te)
{
  MERP omer;
  MEP me;
  
  me = NULL;
  while ((me = BESS_STUB_MV_GET_NEXT_ME (common_mv, me))) {
    omer = me_get_mer (me);
    if (omer->te == te) {       /* same te */
      if (mer) {                /* continuation */
        if ((omer != mer) &&             /* do not back up */
            (common_mv != te->tv[1]->mv) /* for a cyclic edge */
            ) return omer;
      } else {                  /* start */
        return omer;
      }
    }
  }
  return NULL;
}

static void
order_mers (TEP te)
{
  MERP mer;
  LIST merl;
  MVP nmv;
  MVP emv[3];
  
  merl = merp_l_make ();
  mer = next_mer (NULL, te->tv[0]->mv, te);
  assert (mer);
  nmv = te->tv[0]->mv;
  do {
    BESS_STUB_ME_GET_MV (mer->me, emv);
    nmv  = other_v (nmv, emv);
    add_to_tail (merl, mer);
  } while ((mer  = next_mer (mer, nmv, te)));

  { LIST al = merl; merl = te->merl; te->merl = al; } /* swap lists */
  free_list (merl, NOT_NODES);
}


static void
complete_closed_e (TEP te, BESSP bess)
{
  TVP cv;
  MVP mv[3];

  if ((!te->tv[0]) && (!te->tv[1])) {
    /* closed edge */
    MERP mer;
    mer = merp_l_next (te->merl, NULL);
    BESS_STUB_ME_GET_MV (mer->me, mv);
    cv = mv_get_tv (mv[0]);
    if (!cv) { cv = new_tv (mv[0]); attach_v (bess, cv); }
    e_set_v (te, cv, bess);
    e_set_v (te, cv, bess);
  }
}

static void
order_e (TEP te, BESSP bess)
{
  if ((!te->tv[0]) && (!te->tv[1]))
    complete_closed_e (te, bess);
  if (te->tv[0] == NULL || te->tv[1] == NULL) {
    REPERR (bess, "One or more TV not set");
  }
  order_mers (te);
}



static void
classify_te (TEP te, BESSP bess)
{
  MERP mer;
  
  mer = merp_l_next (te->merl, NULL);
  te->num_inc_f = mer->num_inc_t;

  mer = NULL;
  while ((mer = merp_l_next (te->merl, mer))) {
    if (te->num_inc_f != mer->num_inc_t) {
      char buf[256];
      MVP emv[3];
      LOC l[2];
      BESS_STUB_ME_GET_MV (mer->me, emv);
      BESS_STUB_MV_GET_LOC (emv[0], &l[0]);
      BESS_STUB_MV_GET_LOC (emv[1], &l[1]);
      sprintf (buf, "Wrong incidences: %d (TE %d) vs. %d (ME {%g %g %g} {%g %g %g})",
               te->num_inc_f, te->id, mer->num_inc_t,
               l[0].x, l[0].y, l[0].z, l[1].x, l[1].y, l[1].z);
      
      te->num_inc_f = min (te->num_inc_f, mer->num_inc_t);
      REPERR (bess, buf);
    }
  }
}


static void
extract_edges (BESSP bess)
{
  TFP tf;
  TEP te;
  DYN_CONFIG (static int doit = TRUE;)
  DYN_CONFIG (if (!doit) return;)
  
  tf = NULL;
  while ((tf = tfp_l_next (bess->tfl, tf))) {
    extr_es (tf, bess);
  }

  te = NULL;
  while ((te = tep_l_next (bess->tel, te))) {
    order_e (te, bess);
  }

  te = NULL;
  while ((te = tep_l_next (bess->tel, te))) {
    classify_te (te, bess);
  }
}


static void
find_tes_bounding_tf (TFP tf, dict_dictionary ted)
{
  MERP mer;

  mer = NULL;
  while ((mer = merp_l_next (tf->merl, mer))) {
    if (dict_find_obj_with_int_name (ted, mer->te->id) == NULL) {
      dict_add_obj_int_name (ted, mer->te, mer->te->id);
    }
  }
}


static bess_entity_class_enum
f_class (TFP tf)
{
  TEUP teu;
  TLP tl;
  bess_entity_class_enum cl = BESS_CLASS_MANIFOLD;
  tl = NULL;
  while ((tl = tlp_l_next (tf->tll, tl))) {
    teu = NULL;
    while ((teu = teup_l_next (tl->teul, teu))) {
      if (cl == BESS_CLASS_MANIFOLD && bess_te_is_sheet (teu->te))
        cl = BESS_CLASS_SHEET;
      if (bess_te_is_non_manifold (teu->te)) return BESS_CLASS_NON_MANIFOLD;
    }
  }
  return cl;
}

static bess_entity_class_enum
s_class (TSP ts)
{
  TFUP tfu;
  TFP tf;
  bess_entity_class_enum tscl = BESS_CLASS_MANIFOLD;
  bess_entity_class_enum tfcl;

  tfu = NULL;
  while ((tfu = bess_ts_get_next_tfu (ts, tfu))) {
    tf = tfu->tf;
    tfcl = f_class (tf);
    if (tscl == BESS_CLASS_MANIFOLD && tfcl == BESS_CLASS_SHEET)
      tscl = BESS_CLASS_SHEET;
    if (tfcl == BESS_CLASS_NON_MANIFOLD)
      return BESS_CLASS_NON_MANIFOLD;
  }
  return tscl;
}


#define P fprintf (fp, 
#define EP )
#define INDENT "      "
#define NL "\n"
#define BR "<br>\n"
#define PAR "<p>"


static int
print_tv (TVP v, BESSP bess, FILE *fp)
{
  TEP te;
  LOC l;
  BESS_STUB_MV_GET_LOC (v->mv, &l);
  P INDENT "TV %d <%g %g %g>" NL, v->id, l.x, l.y, l.z EP;
  P INDENT INDENT "TE: (" EP;
  te = NULL;
  while ((te = tep_l_next (v->tel, te))) {
    P "%d ", te->id EP;
  }
  P ")" NL EP;
  return TRUE;
}

static int
html_tv (TVP v, BESSP bess, FILE *fp)
{
  TEP te;
  LOC l;
  
  BESS_STUB_MV_GET_LOC (v->mv, &l);
  P "<a name=\"TV_%d\"></a>", v->id EP;
  P "<h3>TV %d</h3>", v->id EP;
  P "loc={%g %g %g}" BR, l.x, l.y, l.z EP;
  P "TE: (" EP;
  {
    GENWRAP_LIST_TYPE (TEP) tel;
    tel = tep_l_ordered_copy (v->tel);
    te = NULL;
    while ((te = tep_l_next (tel, te))) {
      P "<a href=\"#TE_%d\"> %d </a> ", te->id, te->id EP;
    }
    tep_l_free (tel);
  }
  P ")" BR EP;
  P BR "<a href=\"#bess\">Top</a>" PAR NL EP;
  return TRUE;
}

static int
print_te (TEP e, BESSP bess, FILE *fp)
{
  TEUP teu;
  
  P INDENT "TE %d (%s)" NL, e->id,
  (e->num_inc_f == 1
   ? "sheet"
   : (e->num_inc_f == 2
      ? "manifold"
      : "non-manifold"
      )) EP;
  P INDENT INDENT "TV {%d, %d}" NL, e->tv[0]->id, e->tv[1]->id EP;

  P INDENT INDENT "TL: (" EP;
  teu = NULL;
  while ((teu = teup_l_next (e->teul, teu))) {
    P "%d ", teu->tl->id EP;
  }
  P ")" NL EP;
  
  return TRUE;
}

static int
html_te (TEP e, BESSP bess, FILE *fp)
{
  TEUP teu;
  
  P "<a name=\"TE_%d\"></a>", e->id EP;
  P "<h3>TE %d</h3>", e->id EP;
  P "%s" BR, 
  (e->num_inc_f == 1
   ? "sheet"
   : (e->num_inc_f == 2
      ? "manifold"
      : "non-manifold"
      )) EP;
  P "TV {<a href=\"#TV_%d\">%d<a/>, <a href=\"#TV_%d\">%d</a>}" BR,
  e->tv[0]->id, e->tv[0]->id, e->tv[1]->id, e->tv[1]->id EP;

  P "TL: (" EP;
  {
    GENWRAP_LIST_TYPE (TEUP) teul;
    teul = teup_l_ordered_copy (e->teul);
    teu = NULL;
    while ((teu = teup_l_next (teul, teu))) {
      P "<a href=\"#TL_%d\">%d </a>", teu->tl->id, teu->tl->id EP;
    }
    teup_l_free (teul);
  }
  P ")" BR EP;
  if (bess_te_is_non_manifold (e)) {
    GENWRAP_LIST_TYPE (TFP) tfl;
    TFP tf;
    P "Radial ordering: <" EP;
    tfl = bess_tf_in_radial_order_around (e);
    tf = NULL;
    while ((tf = tfp_l_next (tfl, tf))) {
      P "<a href=\"#TF_%d\">%d</a> ", tf->id,  tf->id EP;
    }
    P ">" BR EP;
  }
  P BR "<a href=\"#bess\">Top</a>" PAR NL EP;
  
  return TRUE;
}

static int
print_tl (TLP l, BESSP bess, FILE *fp)
{
  TEUP teu;
  char o;
  
  P INDENT "TL %d" NL, l->id EP;
  P INDENT INDENT "TE: (" EP;
  teu = NULL;
  while ((teu = teup_l_next (l->teul, teu))) {
    if      (teu->o == BESS_ORIENT_SAME)     o = ' ';
    else if (teu->o == BESS_ORIENT_REVERSED) o = '-';
    else                               o = '?';
    P "%c%d ", o, teu->te->id EP;
  }
  P ")" NL EP;
  P INDENT INDENT "TF: %d" NL, l->tf->id EP;
  
  return TRUE;
}

static int
html_tl (TLP l, BESSP bess, FILE *fp)
{
  TEUP teu;
  char o;
  
  P "<a name=\"TL_%d\"></a>", l->id EP;
  P PAR "<H3>TL %d</h3>", l->id EP;
  P "TE: (" EP;
  {
    GENWRAP_LIST_TYPE (TEUP) teul;
    teul = teup_l_ordered_copy (l->teul);
    teu = NULL;
    while ((teu = teup_l_next (teul, teu))) {
      if      (teu->o == BESS_ORIENT_SAME)     o = ' ';
      else if (teu->o == BESS_ORIENT_REVERSED) o = '-';
      else                               o = '?';
      P "<a href=\"#TE_%d\">%c%d</a> ", teu->te->id, o, teu->te->id EP;
    }
    teup_l_free (teul);
  }
  P ")" BR EP;
  P "TF: <a href=\"#TF_%d\">%d</a>" BR, l->tf->id,  l->tf->id EP;
  P BR "<a href=\"#bess\">Top</a>" PAR NL EP;
  
  return TRUE;
}

static int
print_tf (TFP f, BESSP bess, FILE *fp)
{
  TLP l;
  TFUP fu;
  
  P INDENT "TF %d" NL, f->id EP;
  P INDENT INDENT "TL: (" EP;
  l = NULL;
  while ((l = tlp_l_next (f->tll, l))) {
    P "%d ", l->id EP;
  }
  P ")" NL EP;
  P INDENT INDENT "TS: (" EP;
  fu = NULL;
  while ((fu = tfup_l_next (f->tful, fu))) {
    P "%d ", fu->ts->id EP;
  }
  P ")" NL EP;
  
  return TRUE;
}

static int
tf_euler_characteristic (TFP tf)
{
  MVP mvs[3];
  MFRP mfr;
  MEP mes[3];
  int i, chi = 0;
  
  dict_dictionary mvd = dict_create_dictionary ();
  dict_dictionary med = dict_create_dictionary ();
  dict_dictionary mfd = dict_create_dictionary ();

  mfr = NULL;
  while ((mfr = mfrp_l_next (tf->mfrl, mfr))) {
    dict_add_obj_handle_name (mfd, mfr->mf, mfr->mf);
    BESS_STUB_MF_GET_ME (mfr->mf, mes);
    FOR (i, 3) {
      dict_add_obj_handle_name (med, mes[i], mes[i]);
    }
    BESS_STUB_MF_GET_MV (mfr->mf, mvs);
    FOR (i, 3) {
      dict_add_obj_handle_name (mvd, mvs[i], mvs[i]);
    }
  }

  chi = dict_count (mvd) - dict_count (med) + dict_count (mfd);

  dict_destroy_dictionary (mvd);
  dict_destroy_dictionary (med);
  dict_destroy_dictionary (mfd);

  return chi;
}

static int
html_tf (TFP f, BESSP bess, FILE *fp)
{
  TLP l;
  TFUP fu;
  int mfcount;
  int chi;
  bess_entity_class_enum cl;

  chi =  tf_euler_characteristic (f);

  mfcount = list_length (f->mfrl);
  
  P "<a name=\"TF_%d\"></a>", f->id EP;
  P INDENT "<h3>TF %d</h3>", f->id EP;
  P INDENT INDENT "%d mesh facets" BR, mfcount EP;
  P "Euler characteristic, chi: %d" BR, chi EP;
  cl = f_class (f);
  switch (cl) {
  case BESS_CLASS_MANIFOLD:
    P "TF is manifold" BR EP;
    break;
  case BESS_CLASS_SHEET:
    P "TF is sheet" BR EP;
    break;
  case BESS_CLASS_NON_MANIFOLD:
    P "TF is non-manifold" BR EP;
    break;
  }
  P INDENT INDENT "TL: (" EP;
  {
    GENWRAP_LIST_TYPE (TLP) tll;
    tll = tlp_l_ordered_copy (f->tll);
    l = NULL;
    while ((l = tlp_l_next (tll, l))) {
      P "<a href=\"#TL_%d\">%d</a> ", l->id, l->id EP;
    }
    tlp_l_free (tll);
  }
  P ")" BR EP;
  P INDENT INDENT "TS: (" EP;
  {
    GENWRAP_LIST_TYPE (TFUP) tful;
    tful = tfup_l_ordered_copy (f->tful);
    fu = NULL;
    while ((fu = tfup_l_next (tful, fu))) {
      P "<a href=\"#TS_%d\">%d</a> ", fu->ts->id, fu->ts->id EP;
    }
    tfup_l_free (tful);
  }
  P ")" BR EP;
  if (bess_tf_is_sheet (f)) {
    P "TF is sheet" BR NL EP;
  }
  P BR "<a href=\"#bess\">Top</a>" PAR NL EP;
  
  return TRUE;
}


static int
ts_euler_characteristic_alt (TSP ts)
{
  TFUP tfu;
  TLP tl;
  TEUP teu;
  int chi = 0;
  int addv = 0;
  int adde = 0;
  int ntv, nte, ntf;
  
  dict_dictionary tvd = dict_create_dictionary ();
  dict_dictionary ted = dict_create_dictionary ();
  dict_dictionary tfd = dict_create_dictionary ();

  tfu = NULL;
  while ((tfu = tfup_l_next (ts->tful, tfu))) {
    dict_add_obj_handle_name (tfd, tfu->tf, tfu->tf);
    if (list_empty (tfu->tf->tll)) {
      addv++;
    } else {
      int numtl = 0;
      tl = NULL;
      while ((tl = tlp_l_next (tfu->tf->tll, tl))) {
        numtl++;
        teu = NULL;
        while ((teu = teup_l_next (tl->teul, teu))) {
          dict_add_obj_handle_name (ted, teu->te, teu->te);
          dict_add_obj_handle_name (tvd, teu->te->tv[0], teu->te->tv[0]);
          dict_add_obj_handle_name (tvd, teu->te->tv[1], teu->te->tv[1]);
        }
      }
      adde += (numtl - 1);
    }
  }

  ntv = dict_count (tvd);
  nte = dict_count (ted);
  ntf = dict_count (tfd);
  
  chi = (+ (
            + ntv
            + addv                /* for empty loops */
            )
         - (+ nte
            + adde /* for more than one loop per face */
            )
         + ntf
         );
         
  dict_destroy_dictionary (tvd);
  dict_destroy_dictionary (ted);
  dict_destroy_dictionary (tfd);

  return chi;
}

static int
ts_euler_characteristic (TSP ts)
{
  TFUP tfu;
  MVP mvs[3];
  MFRP mfr;
  MEP mes[3];
  int i, chi = 0;
  
  dict_dictionary mvd = dict_create_dictionary ();
  dict_dictionary med = dict_create_dictionary ();
  dict_dictionary mfd = dict_create_dictionary ();

  tfu = NULL;
  while ((tfu = tfup_l_next (ts->tful, tfu))) {
    mfr = NULL;
    while ((mfr = mfrp_l_next (tfu->tf->mfrl, mfr))) {
      dict_add_obj_handle_name (mfd, mfr->mf, mfr->mf);
      BESS_STUB_MF_GET_ME (mfr->mf, mes);
      FOR (i, 3) {
        dict_add_obj_handle_name (med, mes[i], mes[i]);
      }
      BESS_STUB_MF_GET_MV (mfr->mf, mvs);
      FOR (i, 3) {
        dict_add_obj_handle_name (mvd, mvs[i], mvs[i]);
      }
    }
  }
  chi = dict_count (mvd) - dict_count (med) + dict_count (mfd);

  dict_destroy_dictionary (mvd);
  dict_destroy_dictionary (med);
  dict_destroy_dictionary (mfd);

  return chi;
}

static int
ts_is_closed (TSP ts)
{
  TFUP tfu, atfu;
  int num_uses;
  
  tfu = NULL;
  while ((tfu = tfup_l_next (ts->tful, tfu))) {
    num_uses = 0;
    atfu = NULL;
    while ((atfu = tfup_l_next (tfu->tf->tful, atfu))) {
      if (atfu->ts == ts) num_uses++;
    }
    if (num_uses < 2) return TRUE;
  }

  return FALSE;
}

static int
print_ts (TSP s, BESSP bess, FILE *fp)
{
  char o;
  TFUP tfu;
  int chi;
  int closed;
  
  P INDENT "TS %d" NL, s->id EP;
  chi =  ts_euler_characteristic (s);
  closed = ts_is_closed (s);
  if (closed)
    P "TS is closed" NL EP;
  else
    P "TS is open" NL EP;
#if 0
  {
    int chialt;
    chialt = ts_euler_characteristic_alt (s);
    if (chi != chialt)
      P INDENT INDENT "Inconsistent, chi: %d, chialt: %d" NL, chi, chialt EP;
  }
#endif
  P INDENT INDENT "Euler characteristic, chi: %d" NL, chi EP;
  if (closed)
    P "Genus [= (2 - chi) / 2]: %d" NL, (2 - chi) / 2 EP;
  else
    P "Genus [= 1 - chi]: %d" NL, (1 - chi) EP;

  P INDENT INDENT "TF: (" EP;
  tfu = NULL;
  while ((tfu = tfup_l_next (s->tful, tfu))) {
    if      (tfu->o == BESS_ORIENT_SAME)     o = ' ';
    else if (tfu->o == BESS_ORIENT_REVERSED) o = '-';
    else                               o = '?';
    P "%c%d ", o, tfu->tf->id EP;
  }
  P ")" NL EP;

  if (s->tr) {
    P INDENT INDENT "TR: %d", s->tr->id EP;
  }
  
  return TRUE;
}

static int
html_ts (TSP s, BESSP bess, FILE *fp)
{
  char o;
  TFUP tfu;
  int chi;
  int closed;
  bess_entity_class_enum cl;
  
  P "<a name=\"TS_%d\"></a>", s->id EP;
  P "<h3>TS %d</h3>", s->id EP;
  chi    =  ts_euler_characteristic (s);
  closed = ts_is_closed (s);
  if (closed)
    P "TS is closed" BR EP;
  else
    P "TS is open" BR EP;
  cl = s_class (s);
  switch (cl) {
  case BESS_CLASS_MANIFOLD:
    P "TS is manifold" BR EP;
    break;
  case BESS_CLASS_SHEET:
    P "TS is sheet" BR EP;
    break;
  case BESS_CLASS_NON_MANIFOLD:
    P "TS is non-manifold" BR EP;
    break;
  }
#if 0
  {
    int chialt;
    chialt = ts_euler_characteristic_alt (s);
    if (chi != chialt)
      P INDENT INDENT "Inconsistent, chi: %d, chialt: %d" NL, chi, chialt EP;
  }
#endif
  P "Euler characteristic, chi: %d" BR, chi EP;
  if (closed)
    P "Genus [= (2 - chi) / 2]: %d" BR, (2 - chi) / 2 EP;
  else
    P "Genus [= 1 - chi]: %d" BR, (1 - chi) EP;

  P "TF: (" EP;
  {
    GENWRAP_LIST_TYPE (TFUP) tful;
    tful = tfup_l_ordered_copy (s->tful);
    tfu = NULL;
    while ((tfu = tfup_l_next (tful, tfu))) {
      if      (tfu->o == BESS_ORIENT_SAME)     o = ' ';
      else if (tfu->o == BESS_ORIENT_REVERSED) o = '-';
      else                               o = '?';
      P "<a href=\"#TF_%d\">%c%d</a> ", tfu->tf->id, o, tfu->tf->id EP;
    }
    tfup_l_free (tful);
  }
  P ")" BR EP;

  if (s->tr) {
    P "TR: <A HREF=\"#TR_%d\">%d</a>" BR, s->tr->id, s->tr->id EP;
  }

  P BR "<a href=\"#bess\">Top</a>" PAR NL EP;
  
  return TRUE;
}

static int
print_tr (TRP r, BESSP bess, FILE *fp)
{
  TSP ts;
  
  P INDENT "TR %d" NL, r->id EP;
  P INDENT INDENT "TS: (" EP;
  ts = NULL;
  while ((ts = tsp_l_next (r->tsl, ts))) {
    P "%d ", ts->id EP;
  }
  P ")" NL EP;
  
  return TRUE;
}

static int
html_tr (TRP r, BESSP bess, FILE *fp)
{
  TSP ts;
  
  P "<a name=\"TR_%d\"></a>", r->id EP;
  P INDENT "TR %d" BR, r->id EP;
  P INDENT INDENT "TS: (" EP;
  {
    GENWRAP_LIST_TYPE (TSP) tsl;
    tsl = tsp_l_ordered_copy (r->tsl);
    ts = NULL;
    while ((ts = tsp_l_next (tsl, ts))) {
      P "<a href=\"#TS_%d\">%d</a> ", ts->id, ts->id EP;
    }
    tsp_l_free (tsl);
  }
  P ")" NL EP;
  P BR "<a href=\"#bess\">Top</a>" PAR NL EP;
  
  return TRUE;
}

int
bess_print (BESSP bess, FILE *fp)
{
  TVP v;
  TEP e;
  TLP l;
  TFP f;
  TSP s;
  TRP r;
  
  P "BESS:\n" EP;
  P INDENT "totals:" NL EP;
  P INDENT INDENT
    "#tv %d, #te %d, #tl %d, #tf %d, #ts %d, #tr %d" NL,
    list_length (bess->tvl),
    list_length (bess->tel),
    list_length (bess->tll),
    list_length (bess->tfl),
    list_length (bess->tsl),
    list_length (bess->trl)
      EP;
  P INDENT
    "limit crease angle = %g, corner angle = %g" NL,
    180 / M_PI * bess->crease_ang, 180 / M_PI * bess->corner_ang EP;

#define PENT(ent)                                               \
  ent = NULL;                                                   \
  while ((ent = t##ent##p_l_next (bess->t##ent##l, ent))) {     \
    print_t##ent (ent, bess, fp);                               \
  }

  PENT (v);
  PENT (e);
  PENT (l);
  PENT (f);
  PENT (s);
  PENT (r);
 
  return TRUE;
}


int
bess_html (BESSP bess, FILE *fp, char *options)
{
  TVP v;
  TEP e;
  TLP l;
  TFP f;
  TSP s;
  TRP r;
  int include_tv = TRUE, include_te = TRUE, include_tl = TRUE, include_tf = TRUE, include_ts = TRUE, include_tr = TRUE;

  if (options) {
    tokens_parser parser = tokens_new_parser ();
    if (tokens_parse_line (parser, options)) {
      int i;
      for (i = 1; i <= tokens_total_of_tokens (parser); i++) {
        if (tokens_token_and_keyword_equiv (tokens_token_as_string (parser, i), "NOTV"))
          include_tv = FALSE;
        if (tokens_token_and_keyword_equiv (tokens_token_as_string (parser, i), "NOTE"))
          include_te = FALSE;
        if (tokens_token_and_keyword_equiv (tokens_token_as_string (parser, i), "NOTL"))
          include_tl = FALSE;
        if (tokens_token_and_keyword_equiv (tokens_token_as_string (parser, i), "NOTF"))
          include_tf = FALSE;
        if (tokens_token_and_keyword_equiv (tokens_token_as_string (parser, i), "NOTS"))
          include_ts = FALSE;
        if (tokens_token_and_keyword_equiv (tokens_token_as_string (parser, i), "NOTR"))
          include_tr = FALSE;
      }
    }
    tokens_delete_parser (parser);
  }
  
  P "<html>" NL EP;
  P "<head>" NL EP;
  P "<title> BESS topological data structure. (C) 1997,1998 Petr Krysl </title>" NL EP;
  P "</head>" NL EP;
#if 0
  P "<BASE target=\"bess.topo.viewer.window\">" NL EP;
#endif

  P "<body>" NL EP;
  P "<a name=\"bess\"></a>" EP;
  P PAR "<center> <em> <font size=+2>B</font>Rep for <font size=+2>E</font>volving <font size=+2>S</font>hape <font size=+2>S</font>imulations </em> </center>" BR EP;
 P PAR "<center> <em> Copyright (C) 1997,1998 Petr Krysl </em> </center>" BR EP;
  P PAR "<h1>BESS topological data structure:</h1>" EP;
  P PAR "totals:" NL EP;
  P "#tv %d, #te %d, #tl %d, #tf %d, #ts %d, #tr %d" BR,
    list_length (bess->tvl),
    list_length (bess->tel),
    list_length (bess->tll),
    list_length (bess->tfl),
    list_length (bess->tsl),
    list_length (bess->trl)
      EP;
  P INDENT
    "limit crease angle = %g, corner angle = %g" BR,
    180 / M_PI * bess->crease_ang, 180 / M_PI * bess->corner_ang EP;

  P "Mesh is %s" BR, (BESS_STUB_IS_ORIENTABLE (bess->mesh) ? "orientable" : "<em>not</em> orientable") EP;

#undef PENT
#define PENT(ent)                                                       \
  if (include_t##ent) {                                                 \
    ent = NULL;                                                         \
    while ((ent = t##ent##p_l_next (bess->t##ent##l, ent))) {           \
      html_t##ent (ent, bess, fp);                                      \
    }                                                                   \
  } else {                                                              \
    P "Data omitted on request" EP;                                     \
  }
  
  P PAR EP;
  P "<h2> Contents</h2>" EP;
  P "<ol> " EP;
  P "<li> <a href=\"#vertices\">Vertices</a>" EP;
  P "<li> <a href=\"#edges\">Edges</a>" EP;
  P "<li> <a href=\"#loops\">Loops</a>" EP;
  P "<li> <a href=\"#faces\">Faces</a>" EP;
  P "<li> <a href=\"#shells\">Shells</a>" EP;
  P "<li> <a href=\"#regions\">Regions</a>" EP;
  P "</ol> " EP;

  P "<h2><a name=\"vertices\">Vertices</a></h2>" EP;
  PENT (v);
  P "<h2><a name=\"edges\">Edges</a></h2>" EP;
  PENT (e);
  P "<h2><a name=\"loops\">Loops</a></h2>" EP;
  PENT (l);
  P "<h2><a name=\"faces\">Faces</a></h2>" EP;
  PENT (f);
  P "<h2><a name=\"shells\">Shells</a></h2>" EP;
  PENT (s);
  P "<h2><a name=\"regions\">Regions</a></h2>" EP;
  PENT (r);
 
  P "</body>" NL EP;
  P "</html>" NL EP;
  return TRUE;
}


static void
conn_te_teu (TEP te, TEUP teu)
{
  teu->te = te;
  teup_l_add (te->teul, teu);
}

static int
e_used_in_l (TEP te, TLP tl)
{
  TEUP teu;
  teu = NULL;
  while ((teu = teup_l_next (te->teul, teu))) {
    if (teu->tl == tl) return TRUE;
  }
  return FALSE;
}


static void
conn_tl_teu (TLP tl, TEUP teu)
{
  teu->tl = tl;
  teup_l_add (tl->teul, teu);
}


static MEP
me_leaving_mv (MFP mf, MVP mv)
{
  int ei, vi;
  MVP tmv[6];
  MEP tme[6];
  BESS_STUB_MF_GET_MV (mf, tmv);
  if      (tmv[0] == mv) vi = 0;
  else if (tmv[1] == mv) vi = 1;
  else                   vi = 2;
  ei = (vi + 1) % 3;
  BESS_STUB_MF_GET_ME (mf, tme);
  return tme[ei];
}



static TEP
any_te_leaving_tv (TVP tv, TFP tf)
{
  MEP me, eme;
  MERP mer;
  MFP mf;
  TFP atf;
  MFRP mfr;
  
  me = NULL;
  while ((me = BESS_STUB_MV_GET_NEXT_ME (tv->mv, me))) {
    mf = NULL;
    while ((mf = BESS_STUB_ME_GET_NEXT_MF (me, mf))) {
      mfr = mf_get_mfr (mf);
      atf = mfr->tf;
      if (atf == tf) {
        eme = me_leaving_mv (mf, tv->mv);
        if (eme == me) {
          mer = me_get_mer (me);
          if (mer->te)
            return mer->te;
        }
      }
    }
  }
  assert (NULL);
  return NULL;
}


static GENWRAP_LIST_TYPE (TEP)
tes_leaving_tv (TVP tv, TFP tf)
{
  MEP me, eme;
  MERP mer;
  MFP mf;
  TFP atf;
  MFRP mfr;
  GENWRAP_LIST_TYPE (TEP) tel = make_list ();
  
  me = NULL;
  while ((me = BESS_STUB_MV_GET_NEXT_ME (tv->mv, me))) {
    mf = NULL;
    while ((mf = BESS_STUB_ME_GET_NEXT_MF (me, mf))) {
      mfr = mf_get_mfr (mf);
      atf = mfr->tf;
      if (atf == tf) {
        eme = me_leaving_mv (mf, tv->mv);
        if (eme == me) {
          mer = me_get_mer (me);
          if (mer->te) add_to_tail (tel, mer->te);
        }
      }
    }
  }
  return tel;
}


static MVP
via_mv (MERP mer, TVP tv)
{
  MVP mv[3];
  BESS_STUB_ME_GET_MV (mer->me, mv);
  if (tv->mv == mv[0]) return mv[1];
  else                 return mv[0];
}

static int
uses_mvs_in_order (MFP mf, MVP v0, MVP v1)
{
  MVP mv[6];
  BESS_STUB_MF_GET_MV (mf, mv);
  if      (mv[0] == v0 && mv[1] == v1) return TRUE;
  else if (mv[1] == v0 && mv[2] == v1) return TRUE;
  else if (mv[2] == v0 && mv[0] == v1) return TRUE;
  else                                 return FALSE;
}

static MVP 
third (MFP mf, MVP v0, MVP v1)
{
  MVP mv[6];
  BESS_STUB_MF_GET_MV (mf, mv);
  if      (mv[0] != v0 && mv[0] != v1) return mv[0];
  else if (mv[1] != v0 && mv[1] != v1) return mv[1];
  else if (mv[2] != v0 && mv[2] != v1) return mv[2];
  else                                 return NULL;
}


static int
used_by_tf (TEP te, TVP tv, TFP tf)
{
  TOE toe;
  TEUP teu;
  if (te->tv[0] == tv)  toe = BESS_ORIENT_SAME;
  else                  toe = BESS_ORIENT_REVERSED;
  teu = NULL;
  while ((teu = teup_l_next (te->teul, teu))) {
    if (teu->tl->tf == tf && teu->o == toe)
      return TRUE;
  }
  return FALSE;
}


MEP
mv_get_shared_me (MVP v, MVP otherv)
{
  MEP me;
  MVP omv[3];
  me = NULL;
  while ((me = BESS_STUB_MV_GET_NEXT_ME (v, me))) {
    BESS_STUB_ME_GET_MV (me, omv);
    if (   (omv[0] == v && omv[1] == otherv)
        || (omv[1] == v && omv[0] == otherv)
        ) return me;
  }
  return NULL;
}


static void
build_mf_at_mv_list (MVP mv, LIST l)
{
  MEP me;
  MFP mf;
  
  flush_list (l, NOT_NODES);

  me = NULL; 
  while ((me = BESS_STUB_MV_GET_NEXT_ME (mv, me))) {
    mf = NULL;
    while ((mf = BESS_STUB_ME_GET_NEXT_MF (me, mf))) {
      if (!in_list (l, mf)) add_to_tail (l, mf);
    }
  }
}

static TEP
cw_te (TVP tv, TFP tf, MERP via_mer)
{
  static LIST mfl = NULL;
  MFP mf;
  MFRP mfr;
  MERP mer;
  MEP me;
  MVP viamv, imv, tvmv = tv->mv, tmv;
  TEP te = NULL;

  if (!mfl) mfl = make_list ();
  flush_list (mfl, NOT_NODES);

  build_mf_at_mv_list (tv->mv, mfl);

  viamv = via_mv (via_mer, tv);
  imv   = viamv;
 restart:
  mf = NULL;
  while ((mf = (MFP) get_list_next (mfl, mf))) {
    mfr = mf_get_mfr (mf);
    if (mfr->tf == tf) {
      if (uses_mvs_in_order (mf, imv, tvmv)) {
        tmv = third (mf, imv, tvmv);
        assert (tmv);
        me  = mv_get_shared_me (tvmv, tmv);
        mer = me_get_mer (me);
        if (mer->te) {
          if (!used_by_tf (mer->te, tv, tf)) {
            te = mer->te;
          }
          goto exit;
        } 
        imv = tmv;
        goto restart;
      }
    }
  }
 exit:
  return te;
}

static TVP
other_tv (TEP te, TVP tv)
{
  if (tv == te->tv[0]) return te->tv[1];
  else                 return te->tv[0];
}

static MERP 
last_mer (TEP te, TVP tv)
{
  if (tv == te->tv[0]) return (MERP) get_list_tail (te->merl);
  else                 return (MERP) get_list_head (te->merl);
}

static dict_dictionary
connected_tes (dict_dictionary ted, TEP te)
{
  int i;
  TEP yate;
  dict_dictionary cted = dict_create_dictionary ();
  
  STACK s = make_stack ();
  push_stack (s, te);
  
  while ((te = (TEP) pop_stack (s))) {
    if (dict_find_obj_with_int_name (cted, te->id) == NULL) {
      dict_add_obj_int_name (cted, te, te->id);
      FOR (i, 2) {
        yate = NULL;
        while ((yate = tep_l_next (te->tv[i]->tel, yate))) {
          if (dict_find_obj_with_int_name (ted, yate->id)) 
            push_stack (s, yate);
        }
      }
    }
  }
  free_stack (s, NOT_NODES);
  return cted;
}

static dict_dictionary
tvd_on_tl (dict_dictionary cted)
{
  TEP te;
  int i;
  dict_dictionary tvd = dict_create_dictionary ();
  dict_iterator di = dict_get_iterator ();
  dict_reset_iterator (di);
  while ((te = (TEP) dict_iter_objs (cted, di))) {
    FOR (i, 2) {
      if (dict_find_obj_with_int_name (tvd, te->tv[i]->id) == NULL) {
        dict_add_obj_int_name (tvd, te->tv[i], te->tv[i]->id);
      }
    }
  }
  dict_del_iterator (di);
  return tvd;
}  


static void
start_l (BESSP bess, TFP tf, TEP ste, dict_dictionary ted)
{
  dict_dictionary cted;
  dict_dictionary tvd;
  dict_iterator di = dict_get_iterator ();
  TLP tl;
  TEP te;
  TVP tv;
  TEUP teu;
  TOE toe;
  GENWRAP_LIST_TYPE (TEP) tel;

  tl = new_tl (tf); 
  attach_l (bess, tl);
  
  cted = connected_tes (ted, ste);
  tvd  = tvd_on_tl (cted);

  dict_reset_iterator (di);
  while ((tv = (TVP) dict_iter_objs (tvd, di))) {
    tel = tes_leaving_tv (tv, tf);
    te = NULL;
    while ((te = tep_l_next (tel, te))) {
      if (te->tv[0] == tv)  toe = BESS_ORIENT_SAME;
      else                  toe = BESS_ORIENT_REVERSED;
      teu = new_teu (te, toe); conn_te_teu (te, teu); conn_tl_teu (tl, teu);
    }
    free_list (tel, NOT_NODES);
  }
  
  dict_del_iterator (di);
  dict_destroy_dictionary (tvd);
  dict_destroy_dictionary (cted);
}

static int
e_used_in_l_on_f (TEP te, TFP tf)
{
  TLP tl;
  tl = NULL;
  while ((tl = tlp_l_next (tf->tll, tl))) {
    if (e_used_in_l (te, tl)) return TRUE;
  }
  return FALSE;
}


static void
gen_tls (TFP tf, dict_dictionary ted, BESSP bess)
{
  TEP te;
  int all_done;
  static dict_iterator diter = NULL;
  if (!diter) diter = dict_get_iterator ();

  while (1) {
    all_done = TRUE;
    dict_reset_iterator (diter);
    while ((te = (TEP) dict_iter_objs (ted, diter))) {
      if (!e_used_in_l_on_f (te, tf)) {
        start_l (bess, tf, te, ted);
        all_done = FALSE;
      }
    }
    if (all_done) 
      return;
  }
}

static void
extr_ls (TFP tf, BESSP bess)
{
  dict_dictionary ted = dict_create_dictionary ();

  find_tes_bounding_tf (tf, ted);

  gen_tls (tf, ted, bess);
  
  dict_destroy_dictionary (ted);
}

static void
extract_loops (BESSP bess)
{
  TFP tf;
  DYN_CONFIG (static int doit = TRUE;)
  DYN_CONFIG (if (!doit) return;)
      
  tf = NULL;
  while ((tf = tfp_l_next (bess->tfl, tf))) {
    extr_ls (tf, bess); 
  }
}



static int
tf_used_by_any_ts (TFP tf)
{
  return (list_length (tf->tful) >= 1);
}

static void
conn_tf_tfu (TFP tf, TFUP tfu)
{
  tfu->tf = tf;
  tfup_l_add (tf->tful, tfu);
}


static void
conn_ts_tfu (TSP ts, TFUP tfu)
{
  tfu->ts = ts;
  tfup_l_add (ts->tful, tfu);
}


static void
conn_tr_ts (TRP tr, TSP ts)
{
  ts->tr = tr;
  tsp_l_add (tr->tsl, ts);
}


static TFUP
any_tfu_used_in_ts (TFP tf, TSP ts)
{
  TFUP tfu;
  tfu = NULL;
  while ((tfu = tfup_l_next (tf->tful, tfu))) {
    if (tfu->ts == ts) return tfu;
  }
  return NULL;
}


static int
tfu_used_in_ts (TFP tf, TOE o, TSP ts)
{
  TFUP tfu;
  tfu = NULL;
  while ((tfu = tfup_l_next (tf->tful, tfu))) {
    if (tfu->ts == ts && tfu->o == o) return TRUE;
  }
  return FALSE;
}

/*
   Works under the assumption that teu->te is a manifold edge
 */
static TFP 
tf_across_teu_manifold (TFP tf, TEUP teu)
{
  TEP te = teu->te;
  TEUP ateu;
  ateu = NULL;
  while ((ateu = teup_l_next (te->teul, ateu))) {
    if (ateu->tl->tf != tf) {
      return ateu->tl->tf;
    }
  }
  return FALSE;
}


static void
start_manifold_s (TFP stf, TRP tr, BESSP bess)
{
  TSP ts;
  TFUP tfu, atfu;
  TFP tf = stf, atf;
  TLP tl;
  TEUP teu;
  static STACK s = NULL;

  ts = new_ts (tr);
  attach_s (bess, ts);
  
  if (!s) s = make_stack ();
  if (!stack_empty (s)) {
    REPERR (bess, "stack is not empty");
  }

  tfu = new_tfu (tf, BESS_ORIENT_SAME, ts);
  conn_tf_tfu (tf, tfu);
  conn_ts_tfu (ts, tfu);
  push_stack (s, tf);
  while (!stack_empty (s)) {
    tf = (TFP) pop_stack (s);
    tfu = any_tfu_used_in_ts (tf, ts);
    tl = NULL;
    while ((tl = tlp_l_next (tf->tll, tl))) {
      teu = NULL;
      while ((teu = teup_l_next (tl->teul, teu))) {
        if ((atf = tf_across_teu_manifold (tf, teu))) {
          atfu = any_tfu_used_in_ts (atf, ts);
          if (atfu) {
            /* RAW check orientation; if incompatible, report
               an error: nonsense shell
             */
          } else {
            atfu = new_tfu (atf, tfu->o, ts);
            conn_tf_tfu (atf, atfu);
            conn_ts_tfu (ts, atfu);
            push_stack (s, atf);
          }
        }
      }
    }
  }
}


static void
start_sheet_s (TFP stf, TRP tr, BESSP bess)
{
  TSP ts;
  TFUP tfu;
  TFP tf = stf, atf;
  TLP tl;
  TEUP teu;
  static STACK s = NULL;

  ts = new_ts (tr);
  attach_s (bess, ts);
  
  if (!s) s = make_stack ();
  if (!stack_empty (s)) {
    REPERR (bess, "stack is not empty");
  }

  tfu = new_tfu (tf, BESS_ORIENT_SAME, ts);
  conn_tf_tfu (tf, tfu); conn_ts_tfu (ts, tfu);
  tfu = new_tfu (tf, BESS_ORIENT_REVERSED, ts);
  conn_tf_tfu (tf, tfu); conn_ts_tfu (ts, tfu);
  push_stack (s, tf);
  while (!stack_empty (s)) {
    tf = (TFP) pop_stack (s);
    tl = NULL;
    while ((tl = tlp_l_next (tf->tll, tl))) {
      teu = NULL;
      while ((teu = teup_l_next (tl->teul, teu))) {
        if ((atf = tf_across_teu_manifold (tf, teu))) {
          if (!any_tfu_used_in_ts (atf, ts)) {
            tfu = new_tfu (atf, BESS_ORIENT_SAME, ts);
            conn_tf_tfu (atf, tfu); conn_ts_tfu (ts, tfu);
            tfu = new_tfu (atf, BESS_ORIENT_REVERSED, ts);
            conn_tf_tfu (atf, tfu); conn_ts_tfu (ts, tfu);
            push_stack (s, atf);
          }
        }
      }
    }
  }
}

static TFUP
random_tfu (TSP ts)
{
  int randn = rand_0_to_1 () * 13;
  int i;
  TFUP tfu;
  
  i = -1;
  while (TRUE) {
    tfu = NULL;
    while ((tfu = tfup_l_next (ts->tful, tfu))) {
      if (++i == randn) return tfu;
    }
  }
}

static MFP
random_mf (TFUP tfu)
{
  int randn = rand_0_to_1 () * 61;
  int i;
  MFRP mfr;
  TFP tf = tfu->tf;
  
  i = -1;
  while (TRUE) {
    mfr = NULL;
    while ((mfr = mfrp_l_next (tf->mfrl, mfr))) {
      if (++i == randn) return mfr->mf;
    }
  }
}


static void
centroid (MFP mf, LOC *centroid)
{
  int j;
  MVP mv[6];
  int supplied;
  LOC l;

  VECMAC_SCALAR ((*centroid), =, 0);
  supplied = BESS_STUB_MF_GET_MV (mf, mv);
  FOR (j, 3) {
    BESS_STUB_MV_GET_LOC (mv[j], &l);
    VECMAC_2_VEC ((*centroid), +=, 0.333333 * l);
  }
}


static void
compute_direction (MFP mf, bess_orientation_enum o, LOC *l, VEC *direction)
{
  double randu = rand_0_to_1 (), randv = rand_0_to_1 ();
  VEC n;
  G3D_plane_t plane;
  BESS_STUB_MF_GET_NORMAL (mf, 0.3333, 0.3333, &n);
  G3D_plane_consts_p_and_n (&plane, l, &n);
  *direction = plane.n;
  VECMAC_2_VEC ((*direction), +=, randu * plane.u);
  VECMAC_2_VEC ((*direction), +=, randv * plane.v);
  if (o == BESS_ORIENT_REVERSED) {
    VECMAC_SCALAR ((*direction), *=, -1); /* reverse to obtain normal to the ts */
  }
}

static void
start_and_direction (TSP ts, LOC *start, VEC *direction)
{
  TFUP tfu = random_tfu (ts);
  MFP mf = random_mf (tfu);
  
  centroid (mf, start);
  compute_direction (mf, tfu->o, start, direction);
}

static int
do_toggle_ts (TSP ts);

static void
orient_manifold_ts (BESSP bess, TSP ts)
{
#define NRETRY 7
#define NTESTS 3
  int num_intersect[NRETRY];
  int i, k;
  LIST tfl;
  TFUP tfu;
  LOC start;
  VEC direction;
  int result, prev_result;
#define IS_EVEN(num) (num / 2 * 2 == num)
  
  tfl = make_list ();
  tfu = NULL;
  while ((tfu = bess_ts_get_next_tfu (ts, tfu))) {
    add_to_tail (tfl, tfu->tf);
  }

  FOR (k, NTESTS) { 
    FOR (i, NRETRY) {
      start_and_direction (ts, &start, &direction);
      num_intersect[i] = bess_num_intersections (bess, tfl, &start, &direction);
    }
    prev_result = IS_EVEN (num_intersect[0]);
    for (i = 1; i < NRETRY; i++) {
      result = IS_EVEN (num_intersect[i]);
      if (prev_result != result) goto next_test; /* inconsistent results; try another test */
      prev_result = result;
    }
    if (IS_EVEN (num_intersect[0]))
      do_toggle_ts (ts);
    return;
  next_test:
    ;
  }
  
  free_list (tfl, NOT_NODES);
}

static void
orient_manifold_shells (BESSP bess)
{
  TSP ts;
  ts = NULL;
  while ((ts = bess_next_ts (bess, ts))) {
    if (bess_ts_is_manifold (ts)) 
      orient_manifold_ts (bess, ts);
  }
}

static void
fill_tfd (dict_dictionary tfd, TFP tf)
{
  TLP tl;
  TEUP teu;
  TEUP oteu;
  STACK s = make_stack ();
  push_stack (s, tf);
  while ((tf = (TFP) pop_stack (s))) {
    if (dict_find_obj_with_handle_name (tfd, tf) == NULL) {
      dict_add_obj_handle_name (tfd, tf, tf);
      tl = NULL;
      while ((tl = tlp_l_next (tf->tll, tl))) {
        teu = NULL;
        while ((teu = teup_l_next (tl->teul, teu))) {
          oteu = NULL;
          while ((oteu = teup_l_next (teu->te->teul, oteu))) {
            push_stack (s, oteu->tl->tf);
          }
        }
      }
    }
  }
  free_stack (s, NOT_NODES);
}

static dict_dictionary 
collect_unused_conn_tf (BESSP bess)
{
  TFP tf;
  
  tf = NULL;
  while ((tf = tfp_l_next (bess->tfl, tf))) {
    if (!tf_used_by_any_ts (tf)) {
      dict_dictionary tfd = dict_create_dictionary ();
      fill_tfd (tfd, tf);
      return tfd;
    }
  }
  return NULL;
}

static int
connected_tfs_are_manifold (dict_dictionary tfd)
{
  TFP tf;
  int result = TRUE;
  dict_iterator di = dict_get_iterator ();
  dict_reset_iterator (di);
  while ((tf = (TFP) dict_iter_objs (tfd, di))) {
    if (!bess_tf_is_manifold (tf)) {
      result = FALSE;
      goto exit;
    }
  }
 exit:
  dict_del_iterator (di);
  return result;
}


/* If all the shells are either sheet or manifold, any given TF may be used
   no more than in a single shell.  One could optimize based on this.
 */
static void
extract_manifold_or_sheet_shells (BESSP bess)
{
  TFP tf;
  int all_done;
  dict_dictionary tfd;

  while (1) {
    all_done = TRUE;
    tfd = collect_unused_conn_tf (bess);
    if (tfd) {
      tf = dict_next_obj (tfd, NULL);
      if (connected_tfs_are_manifold (tfd)) {
        start_manifold_s (tf, NULL, bess);
      } else {
        start_sheet_s (tf, NULL, bess);
      }
      dict_destroy_dictionary (tfd);
      all_done = FALSE;
    }
    if (all_done) return;
  }
}


static LIST
mf_around_te_at_start (TEP te)
{
  LIST l = make_list ();
  MVP mv[3];
  MFP mf;
  MERP mer = (MERP) get_list_head (te->merl);
  BESS_STUB_ME_GET_MV (mer->me, mv);
  assert (mv[0] == te->tv[0]->mv || mv[1] == te->tv[0]->mv);
  mf = NULL;
  while ((mf = BESS_STUB_ME_GET_NEXT_MF (mer->me, mf))) {
    add_to_tail (l, mf);
  }
  return l;
}

static GENWRAP_LIST_TYPE (TFP)
mfl_to_tfl (LIST mfl)
{
  MFP mf;
  MFRP mfr;
  GENWRAP_LIST_TYPE (TFP) tfl = make_list ();
  mf = NULL;
  while ((mf = (MFP) get_list_next (mfl, mf))) {
    mfr = (MFRP) BESS_STUB_MF_GET_AP (mf);
    add_to_tail (tfl, mfr->tf);
  }
  return tfl;
}

static MEP 
common_me (MEP me0[3], MEP me1[3])
{
  int i;
  FOR (i, 3) {
    if (me0[i] == me1[0]) return me0[i];
    if (me0[i] == me1[1]) return me0[i];
    if (me0[i] == me1[2]) return me0[i];
  }
  return NULL;
}

static void
v_on_e_in_t (MFP mf, MEP me, MVP v[2])
{
  MVP tv[6];
  MEP te[3];
  BESS_STUB_MF_GET_MV (mf, tv);
  BESS_STUB_MF_GET_ME (mf, te);
  if        (me == te[0]) {
    v[0] = tv[2];
    v[1] = tv[0];
  } else if (me == te[1]) {
    v[0] = tv[0];
    v[1] = tv[1];
  } else if (me == te[2]) {
    v[0] = tv[1];
    v[1] = tv[2];
  } else {
    v[0] = v[1] = NULL;
  }
}

static int
compatible_joint (MFP mf0, MEP cme, MFP mf1)
{
  MVP mv[2], omv[2];

  v_on_e_in_t (mf0, cme, mv);
  v_on_e_in_t (mf1, cme, omv);
  return (mv[0] == omv[1] && mv[1] == omv[0]);
}

typedef struct radinfo_t {
  MFP      mf;
  double   angle;
}              radinfo_t;


static int
cmp_angle (radinfo_t *ri0, radinfo_t *ri1)
{
  return (ri0->angle < ri1->angle);
}

static void
eval_along_te (TEP te, LOC *along_te)
{
  LOC l[2];
  MVP mv[3];
  MERP mer = (MERP) get_list_head (te->merl);
  BESS_STUB_ME_GET_MV (mer->me, mv);
  BESS_STUB_MV_GET_LOC (mv[0], &l[0]);
  BESS_STUB_MV_GET_LOC (mv[1], &l[1]);
  if (mv[0] == te->tv[0]->mv) {
    VECMAC_3_VEC ((*along_te), =, l[1], -, l[0]);
  } else {
    VECMAC_3_VEC ((*along_te), =, l[0], -, l[1]);
  }
  VECMAC_NORMALIZE ((*along_te));
}


static double
radang (MFP mf0, MFP mf1, MEP me, LOC *in_plane, LOC *normal)
{
  LOC n;
  double cosa, a, corra;
  BESS_STUB_MF_GET_NORMAL (mf1, 0.3333, 0.3333, &n);
  if (compatible_joint (mf0, me, mf1)) { VECMAC_SCALAR (n, *=, -1); }
  cosa = VECMAC_DOT ((*normal), n);
  corra = a = acos (cosa);
  if (VECMAC_DOT (n, (*in_plane)) < 0) { 
    corra = (2 * M_PI - a);
  }
  return corra;
}

static LIST
radord (TEP te, LIST mfl)
{
  LOC sn, along_te = {0,0,0}, in_plane = {0,0,0};
  MFP smf = (MFP) get_list_head (mfl);
  MFP mf;
  LIST ril = make_list ();
  MERP mer = (MERP) get_list_head (te->merl);
  radinfo_t *ri;

  BESS_STUB_MF_GET_NORMAL (smf, 0.3333, 0.3333, &sn);
  eval_along_te (te, &along_te);
  VECMAC_CROSS (in_plane, =, along_te, x, sn);
  
  mf = NULL;
  while ((mf = (MFP) get_list_next (mfl, mf))) {
    ri = NEWP_OBJ (radinfo_t);
    assert (ri);
    ri->mf = mf;
    if (mf == smf) ri->angle = 0;
    else           ri->angle = radang (smf, mf, mer->me, &in_plane, &sn);
    add_to_tail (ril, ri);
  }
  sort_list_func (ril, cmp_angle);
  flush_list (mfl, NOT_NODES);
  ri = NULL;
  while ((ri = (radinfo_t *) get_list_next (ril, ri))) {
    add_to_tail (mfl, ri->mf);
  }
  free_list (ril, AND_NODES);
  return mfl;
}

static LIST
mf_in_radial_order_around (TEP te)
{
  LIST mfl = mf_around_te_at_start (te);
  return radord (te, mfl);
}

LIST 
bess_tf_in_radial_order_around (TEP te)
{
  GENWRAP_LIST_TYPE (TFP) tfl;
  LIST mfl = mf_in_radial_order_around (te);
  tfl = mfl_to_tfl (mfl);       /* RAW should make sure that the AP
                                   in each mf really stores the pointer
                                   to the appropriate tf */
  free_list (mfl, NOT_NODES);
  return tfl;
}


static void
tf_cw_ccw (GENWRAP_LIST_TYPE (TFP) radl,
                              TFP tf, TEP te, int which_occurence,
                              TFP *cw_tf, TFP *ccw_tf)
{
  int i = 0;
  TFP atf = NULL;

  *cw_tf  = NULL;
  *ccw_tf = NULL;

  atf = NULL;
  while ((atf = tfp_l_next (radl, atf))) {
    if (atf == tf) {
      i++;
      if (i == which_occurence) {
        *ccw_tf = tfp_l_next (radl, atf);
        if (*ccw_tf == NULL) *ccw_tf = tfp_l_head (radl);
        if (*cw_tf == NULL)  *cw_tf  = tfp_l_tail (radl);
        return;
      }
    }
    *cw_tf = atf;
  }
  
  /* no such tf occurrence */
  *cw_tf  = NULL;
  *ccw_tf = NULL;
  return;
}


static void
mf_cw_ccw (LIST mfradl, MFP mf, /* IN */
           MFP *cw_mf, MFP *ccw_mf /* OUT */
           )
{
  MFP amf = NULL;

  *cw_mf  = NULL;
  *ccw_mf = NULL;

  amf = NULL;
  while ((amf = (MFP) get_list_next (mfradl, amf))) {
    if (amf == mf) {
      *ccw_mf = (MFP) get_list_next (mfradl, amf);
      if (*ccw_mf == NULL) *ccw_mf = (MFP) get_list_head (mfradl);
      if (*cw_mf == NULL)  *cw_mf  = (MFP) get_list_tail (mfradl);
      return;
    }
    *cw_mf = amf;
  }
  
  /* no such mf in mfradl */
  *cw_mf  = NULL;
  *ccw_mf = NULL;
  return;
}

void
bess_tf_cw_ccw (LIST radl,
                TFP tf, TEP te, int which_occurence,
                              TFP *cw_tf, TFP *ccw_tf)
{
  LIST l = copy_list (radl, NOT_NODES);
  *cw_tf = *ccw_tf = NULL;
  if (radl && te && tf && which_occurence > 0)
    tf_cw_ccw (l, tf, te, which_occurence, cw_tf, ccw_tf);
  free_list (l, NOT_NODES);
}


static void
mv_along_te (MEP me, TEP te, MVP emv[2])
{
  MVP mv[3];
  BESS_STUB_ME_GET_MV (me, mv);
  if (mv[0] == te->tv[0]->mv) {
    emv[0] = mv[0]; emv[1] = mv[1];
  } else {
    emv[0] = mv[1]; emv[1] = mv[0];
  }
}

static bess_orientation_enum
mf_uses_te_in_o (MVP tmv[6], MVP emv[2])
{
  if (
      (emv[0] == tmv[0] && emv[1] == tmv[1]) ||
      (emv[0] == tmv[1] && emv[1] == tmv[2]) ||
      (emv[0] == tmv[2] && emv[1] == tmv[0])
      )
    return BESS_ORIENT_SAME;
  else
    return BESS_ORIENT_REVERSED;
}


static void
tfu_data_across_non_manifold_te (TEP te, MFP mf, TOE mfo,
                                 TOE *tfuo, TFP *tf)
{
  MFP cw_mf, ccw_mf, mf_across;
  MVP emv[2];
  MVP tmv[6];
  TOE te_use_o;
  MFRP mfr;
  MERP mer = (MERP) get_list_head (te->merl); /* first MER on TE */
  LIST mfradl = mf_in_radial_order_around (te);

  mf_cw_ccw (mfradl, mf, &cw_mf, &ccw_mf);
  assert (cw_mf != NULL && ccw_mf != NULL);

  BESS_STUB_MF_GET_MV (mf, tmv);
  mv_along_te (mer->me, te, emv);
  te_use_o = mf_uses_te_in_o (tmv, emv);

  if (mfo == BESS_ORIENT_SAME) {
    if (te_use_o == BESS_ORIENT_SAME) {
      mf_across = cw_mf;
    } else { /* BESS_ORIENT_REVERSED */
      mf_across = ccw_mf;
    }
  } else { /* BESS_ORIENT_REVERSED */
    if (te_use_o == BESS_ORIENT_SAME) {
      mf_across = ccw_mf;
    } else { /* BESS_ORIENT_REVERSED */
      mf_across = cw_mf;
    }
  }

  mfr = (MFRP) BESS_STUB_MF_GET_AP (mf_across);
  *tf = mfr->tf;

  if (compatible_joint (mf, mer->me, mf_across))
    *tfuo = mfo;
  else
    *tfuo = (bess_orientation_enum) ((-1) * mfo);

  free_list (mfradl, NOT_NODES);
}

static int
mf_is_on_tf (MFP mf, TFP tf)
{
  MFRP mfr = BESS_STUB_MF_GET_AP (mf);
  return (mfr->tf == tf);
}

static void
start_non_manifold_s (TFP stf, TOE so, TRP tr, BESSP bess)
{
  TSP ts;
  TFUP tfu, atfu;
  TOE atfuo;
  TFP tf = stf, atf;
  TLP tl;
  MFP mf;
  LIST mfradl;
  TEUP teu;
  static STACK S = NULL;
#define TFUCONN(tfu, tf, ts) {conn_tf_tfu (tf, tfu); conn_ts_tfu (ts, tfu);}

  ts = new_ts (tr);
  attach_s (bess, ts);
  
  if (!S) S = make_stack ();
  if (!stack_empty (S)) {
    REPERR (bess, "stack is not empty");
  }

  tfu = new_tfu (tf, so, ts); TFUCONN (tfu, tf, ts);
  push_stack (S, tfu);
  while (!stack_empty (S)) {
    tfu = (TFUP) pop_stack (S);
    tf  = tfu->tf;
    tl = NULL;
    while ((tl = tlp_l_next (tf->tll, tl))) {
      teu = NULL;
      while ((teu = teup_l_next (tl->teul, teu))) {
        mfradl = mf_in_radial_order_around (teu->te);
        mf = NULL;
        while ((mf = (MFP) get_list_next (mfradl, mf))) {
          if (mf_is_on_tf (mf, tf)) {
            tfu_data_across_non_manifold_te (teu->te, mf, tfu->o,
                                             &atfuo, &atf);
            if (!tfu_used_in_ts (atf, atfuo, ts)) {
              atfu = new_tfu (atf, atfuo, ts); TFUCONN (atfu, atf, ts);
              push_stack (S, atfu);
            }
          }
        } /* over radial mf's */
        free_list (mfradl, NOT_NODES);
      } /* over teu's */
    } /* over loops */
  } /* !stack_empty */
}


static TOE
o_of_unused_tfu (TFP tf)
{
  TFUP tfu;
  int had_same = FALSE, had_reversed = FALSE;
  tfu = NULL;
  while ((tfu = tfup_l_next (tf->tful, tfu))) {
    if (tfu->o == BESS_ORIENT_REVERSED) had_reversed = TRUE;
    if (tfu->o == BESS_ORIENT_SAME)     had_same     = TRUE;
  }
  if (had_same)
    if (had_reversed)
      return BESS_ORIENT_UNSPECIFIED;
    else
      return BESS_ORIENT_REVERSED;
  else
    return BESS_ORIENT_SAME;
}


/* If all the shells are either sheet or manifold, any given TF may be used
   no more than in a single shell.  One could optimize based on this.
 */
static void
extract_non_manifold_shells (BESSP bess)
{
  TFP tf;
  TOE o;
  int all_done;
  DYN_CONFIG (static int doit = TRUE;)
  DYN_CONFIG (if (!doit) return;)

  while (1) {
    all_done = TRUE;
    tf = NULL;
    while ((tf = tfp_l_next (bess->tfl, tf))) {
      if ((o = o_of_unused_tfu (tf)) != BESS_ORIENT_UNSPECIFIED) {
        start_non_manifold_s (tf, o, NULL, bess);
        all_done = FALSE;
        break;
      }
    }
    if (all_done) return;
  }
}


int
bess_tf_is_manifold (TFP tf) { return (f_class (tf) == BESS_CLASS_MANIFOLD); }

int
bess_tf_is_non_manifold (TFP tf) { return (f_class (tf) == BESS_CLASS_NON_MANIFOLD); }

int
bess_tf_is_sheet (TFP tf) { return (f_class (tf) == BESS_CLASS_SHEET); }

static int
all_f_are_manifold (BESSP bess)
{
  TFP tf;
  tf = NULL;
  while ((tf = tfp_l_next (bess->tfl, tf))) {
    if (!bess_tf_is_manifold (tf)) return FALSE;
  }
  return TRUE;
}

static int
all_f_are_manifold_or_sheet (BESSP bess)
{
  TFP tf;
  tf = NULL;
  while ((tf = tfp_l_next (bess->tfl, tf))) {
    if (bess_tf_is_non_manifold (tf)) return FALSE;
  }
  return TRUE;
}

static int
can_query_regions (BESSP bess)
{
  TFP tf;
  MFRP mfr;
  MRID mridplus;
  MRID mridminus;
  
  tf = tfp_l_next (bess->tfl, NULL);
  mfr = mfrp_l_next (tf->mfrl, NULL);
  return (BESS_STUB_MF_GET_REGIONS (mfr->mf, &mridplus, &mridminus) > 0);
}

static int
have_tr (BESSP bess, MRID mrid)
{
  TRP tr;
  tr = NULL;
  while ((tr = trp_l_next (bess->trl, tr))) {
    if (tr->mrid == mrid) return TRUE;
  }
  return FALSE;
}

static int
have_all_incident_tr (TFP tf, BESSP bess)
{
  MFRP mfr;
  MRID mridplus;
  MRID mridminus;
  int whichr;
  
  mfr = mfrp_l_next (tf->mfrl, NULL);
  whichr = BESS_STUB_MF_GET_REGIONS (mfr->mf,
                                    &mridplus, &mridminus);
  if (whichr < 1) {
    REPERR (bess, "Face does not bound regions?");
    return TRUE;
  }
  if (mridplus) {
    if (!have_tr (bess, mridplus)) return FALSE;
  }
  if (mridminus) {
    if (!have_tr (bess, mridminus)) return FALSE;
  }
  return TRUE;
}

static int
tf_has_tr_on_plus_side (TFP tf, TRP tr)
{
  MFRP mfr;
  MRID mridplus;
  MRID mridminus;
  int  whichr;
  
  mfr = mfrp_l_next (tf->mfrl, NULL);
  whichr = BESS_STUB_MF_GET_REGIONS (mfr->mf,
                                     &mridplus, &mridminus);
  if (whichr > 0 && mridplus) {
    if (tr->mrid == mridplus) return TRUE;
  }
  return FALSE;
}

static int
tf_has_tr_on_minus_side (TFP tf, TRP tr)
{
  MFRP mfr;
  MRID mridplus;
  MRID mridminus;
  int whichr;
  
  mfr = mfrp_l_next (tf->mfrl, NULL);
  whichr = BESS_STUB_MF_GET_REGIONS (mfr->mf,
                                     &mridplus, &mridminus);
  if (whichr && mridminus) {
    if (tr->mrid == mridminus) return TRUE;
  }
  return FALSE;
}




static void
start_general_s (TFP stf, TRP tr, BESSP bess)
{
  TSP ts;
  TFP tf = stf, atf;
  TLP tl;
  TEUP teu, ateu;
  TEP te;
  static STACK s = NULL;

  ts = new_ts (tr);
  attach_s (bess, ts);
  conn_tr_ts (tr, ts);
  
  if (!s) s = make_stack ();
  if (!stack_empty (s)) {
    REPERR (bess, "stack is not empty");
  }

  push_stack (s, tf);
  while (!stack_empty (s)) {
    tf = (TFP) pop_stack (s);
    if (tf_has_tr_on_plus_side (tf, tr)) {    
      TFUP tfu = new_tfu (tf, BESS_ORIENT_REVERSED, ts);       
      conn_tf_tfu (tf, tfu); conn_ts_tfu (ts, tfu);                              
    }
    if (tf_has_tr_on_minus_side (tf, tr)) {    
      TFUP tfu = new_tfu (tf, BESS_ORIENT_SAME, ts);       
      conn_tf_tfu (tf, tfu); conn_ts_tfu (ts, tfu);                              
    }
    tl  = NULL;
    while ((tl = tlp_l_next (tf->tll, tl))) {
      teu = NULL;
      while ((teu = teup_l_next (tl->teul, teu))) {
        te = teu->te;
        ateu = NULL;
        while ((ateu = teup_l_next (te->teul, ateu))) {
          atf = ateu->tl->tf;
          if (atf != tf) {
            if ((!any_tfu_used_in_ts (atf, ts)) &&
                (tf_has_tr_on_minus_side (tf, tr) ||
                 tf_has_tr_on_plus_side  (tf, tr))) {
              push_stack (s, atf);
            }
          }
        }
      }
    }
  }
}

static int
tf_used_by_any_ts_in_tr (TFP tf, TRP tr)
{
  TFUP tfu;
  tfu = NULL;
  while ((tfu = tfup_l_next (tf->tful, tfu))) {
    if (tfu->ts->tr == tr) return TRUE;
  }
  return FALSE;
}

static void
gather_shells (TRP tr, BESSP bess)
{
  TFP tf;
  int all_done;
  
  while (1) {
    all_done = TRUE;
    tf = NULL;
    while ((tf = tfp_l_next (bess->tfl, tf))) {
      if (   tf_has_tr_on_plus_side (tf, tr)
          || tf_has_tr_on_minus_side (tf, tr)) {
        if (!tf_used_by_any_ts_in_tr (tf, tr)) {
          start_general_s (tf, tr, bess);
          all_done = FALSE;
        }
      }
    }
    if (all_done) return;
  } 
}


static void
start_r (TFP stf, BESSP bess)
{
  TRP tr;
  MFRP mfr;
  MRID mridplus;
  MRID mridminus;
  int whichr;
  
  mfr = mfrp_l_next (stf->mfrl, NULL);
  whichr = BESS_STUB_MF_GET_REGIONS (mfr->mf,
                                     &mridplus, &mridminus);
  if (whichr > 0 && mridplus) {
    if (!have_tr (bess, mridplus)) {
      tr = new_tr (mridplus); attach_r (bess, tr);
      gather_shells (tr, bess);
      return;
    }
  }
  if (whichr > 0 && mridminus) {
    if (!have_tr (bess, mridminus)) {
      tr = new_tr (mridminus); attach_r (bess, tr);
      gather_shells (tr, bess);
      return;
    }
  }
}



static void
extract_all_shells (BESSP bess)
{
  TFP tf;
  int all_done;

  while (1) {
    all_done = TRUE;
    tf = NULL;
    while ((tf = tfp_l_next (bess->tfl, tf))) {
      if (!have_all_incident_tr (tf, bess)) {
        start_r (tf, bess);
        all_done = FALSE;
      }
    }
    if (all_done) return;
  } 
}




static void
extract_shells (BESSP bess)
{
  DYN_CONFIG (static int doit = TRUE;)
  DYN_CONFIG (if (!doit) return;)
  
  if (can_query_regions (bess)) {
    extract_all_shells (bess);
  } else {
    if (all_f_are_manifold_or_sheet (bess)) {
      extract_manifold_or_sheet_shells (bess);
      orient_manifold_shells (bess);
    } else {
      extract_non_manifold_shells (bess);
    }
  }
}


/*
   Create a new topology model.
   There are no data at this point in the model, the mesh
   has not been set.
 */
BESSP
bess_new (void)
{
  BESSP bess;
  bess = NEWP_OBJ (BESS);
  if (!bess) return NULL;

  bess->mesh       = NULL;
  bess->mfrl       = mfrp_l_make ();
  bess->merl       = merp_l_make ();
  bess->crease_ang = 45 * M_PI / 180;     /* radians */
  bess->corner_ang = 60 * M_PI / 180;     /* radians */
  bess->tvl        = tvp_l_make ();
  bess->tel        = tep_l_make ();
  bess->tll        = tlp_l_make ();
  bess->tfl        = tfp_l_make ();
  bess->tsl        = tsp_l_make ();
  bess->trl        = trp_l_make ();
  
  return bess;
}

/*
   Build a topological model.
   The mesh data are supplied on input, and it is converted
   to a BRep.  Note that the associated pointers stored
   by the mesh interface are used only within this function.
   In particular, any value stored by the surface mesh entities
   as the associated pointer is bound to be overwritten here.
 */
int
bess_build (BESSP bess, MESHP mesh)
{
  bess->mesh = mesh;
  
  extract_faces (bess);
  extract_edges (bess);
  extract_loops (bess);
  extract_shells (bess);
  
  return TRUE;
}

int
bess_config (BESSP bess, bess_opts_t *opts)
{
  switch (opts->option) {
  case BESS_CREASE_ANGLE:
    bess->crease_ang = opts->value.d * M_PI / 180;
    break;
  case BESS_CORNER_ANGLE:
    bess->corner_ang = opts->value.d * M_PI / 180;
    break;
  default:
    REPERR (bess, "Unrecognized option");
    return FALSE;
  }
  return TRUE;
}


TVP 
bess_next_tv (BESSP bess, TVP v)
{
  return (TV *) tvp_l_next (bess->tvl, v);
}

TEP 
bess_next_te (BESSP bess, TEP e)
{
  return (TE *) tep_l_next (bess->tel, e);
}

TLP
bess_next_tl (BESSP bess, TLP l)
{
  return (TLP) tlp_l_next (bess->tll, l);
}

TFP 
bess_next_tf (BESSP bess, TFP f)
{
  return (TF *) tfp_l_next (bess->tfl, f);
}

TSP 
bess_next_ts (BESSP bess, TSP s)
{
  return (TS *) tsp_l_next (bess->tsl, s);
}

TRP
bess_next_tr (BESSP bess, TRP tr)
{
  return (TRP) trp_l_next (bess->trl, tr);
}

TVP 
bess_find_tv (BESSP bess, int id)
{
  TV *v;
  v = NULL;
  while ((v = tvp_l_next (bess->tvl, v))) {
    if (v->id == id) {
      return v;
    }
  }
  return NULL;
}

TEP 
bess_find_te (BESSP bess, int id)
{
  TE *e;
  loop_through_list (bess->tel, e, TE *) {
    if (e->id == id) {
      return e;
    }
  }
  return NULL;
}

TLP 
bess_find_tl (BESSP bess, int id)
{
  TLP l;
  loop_through_list (bess->tll, l, TLP) {
    if (l->id == id) {
      return l;
    }
  }
  return NULL;
}

TFP 
bess_find_tf (BESSP bess, int id)
{
  TF *f;
  loop_through_list (bess->tfl, f, TF *) {
    if (f->id == id) {
      return f;
    }
  }
  return NULL;
}

TSP 
bess_find_ts (BESSP bess, int id)
{
  TS *s;
  loop_through_list (bess->tsl, s, TS *) {
    if (s->id == id) {
      return s;
    }
  }
  return NULL;
}

TFUP
bess_ts_get_next_tfu (TSP ts, TFUP tfu)
{
  return tfup_l_next (ts->tful, tfu);
}

TFP
bess_tfu_get_tf (TFUP tfu)
{
  return tfu->tf;
}

TOE
bess_tfu_get_o (TFUP tfu)
{
  return tfu->o;
}

TSP 
bess_tfu_get_ts (TFUP tfu)
{
  return tfu->ts;
}

TFUP
bess_tf_get_next_tfu (TFP tf, TFUP tfu)
{
  return tfup_l_next (tf->tful, tfu);
}

LIST
bess_te_get_mel (TEP e)
{
  LIST_OF (MEP) mel = make_list ();
  MER *er;
  loop_through_list (e->merl, er, MER *) {
    add_to_tail (mel, er->me);
  }
  return mel;
}

TEP
bess_teu_get_te (TEUP teu)
{
  return teu->te;
}

TLP
bess_teu_get_tl (TEUP teu)
{
  return teu->tl;
}

TOE
bess_teu_get_o (TEUP teu)
{
  return teu->o;
}

LIST
bess_tf_get_mfl (TFP f)
{
  LIST_OF (MFP) mfl = make_list ();
  MFR *tr;
  loop_through_list (f->mfrl, tr, MFR *) {
    add_to_tail (mfl, tr->mf);
  }
  return mfl;
}


int
bess_tv_get_id (TVP v)
{
  return v->id;
}

int
bess_te_get_id (TEP e)
{
  return e->id;
}

int
bess_tf_get_id (TFP f)
{
  return f->id;
}


int
bess_tl_get_id (TLP l)
{
  return l->id;
}

int
bess_ts_get_id (TSP s)
{
  return s->id;
}

int
bess_teu_get_id (TEUP teu)
{
  return (teu->o == BESS_ORIENT_SAME ? teu->te->id : -teu->te->id);
}

int
bess_tfu_get_id (TFUP tfu)
{
  return (tfu->o == BESS_ORIENT_SAME ? tfu->tf->id : -tfu->tf->id);
}

MVP
bess_tv_get_mv (TVP tv)
{
  return tv->mv;
}


void
bess_te_get_tv (TEP te, TVP tv[2])
{
  tv[0] = te->tv[0];
  tv[1] = te->tv[1];
}


static void
destroy_tv (TVP tv)
{
  free_list (tv->tel, NOT_NODES);
  NEWP_FREE_OBJ (tv);
}

static void
destroy_teu (TEUP teu)
{
  NEWP_FREE_OBJ (teu);
}

static void
destroy_te (TEP te)
{
  TEUP teu;
  teu = NULL;
  while ((teu = teup_l_next (te->teul, teu))) {
    destroy_teu (teu);
  }
  free_list (te->teul, NOT_NODES);
  free_list (te->merl, NOT_NODES);
  NEWP_FREE_OBJ (te);
}

static void
destroy_tl (TLP tl)
{
  NEWP_FREE_OBJ (tl);
}

static void
destroy_tfu (TFUP tfu)
{
  NEWP_FREE_OBJ (tfu);
}

static void
destroy_tf (TFP tf)
{
  TFUP tfu;
  tfu = NULL;
  while ((tfu = tfup_l_next (tf->tful, tfu))) {
    destroy_tfu (tfu);
  }
  free_list (tf->tll, NOT_NODES);
  free_list (tf->tful, NOT_NODES);
  free_list (tf->mfrl, NOT_NODES);
  free_list (tf->merl, NOT_NODES);
  NEWP_FREE_OBJ (tf);
}


static void
destroy_ts (TSP ts)
{
  free_list (ts->tful, NOT_NODES);
  NEWP_FREE_OBJ (ts);
}

static void
destroy_tr (TRP tr)
{
  free_list (tr->tsl, NOT_NODES);
  NEWP_FREE_OBJ (tr);
}

static void
destroy_mfr (MFRP mfr)
{
  NEWP_FREE_OBJ (mfr);
}

static void
destroy_mer (MERP mer)
{
  NEWP_FREE_OBJ (mer);
}

void
bess_destroy (BESSP bess)
{
  TVP tv;
  TEP te;
  TLP tl;
  TFP tf;
  TSP ts;
  TRP tr;
  MFRP mfr;
  MERP mer;

  tv = NULL;
  while ((tv = tvp_l_next (bess->tvl, tv))) {
    destroy_tv (tv);
  }
  
  
  te = NULL;
  while ((te = tep_l_next (bess->tel, te))) {
    destroy_te (te);
  }
  
  
  tl = NULL;
  while ((tl = tlp_l_next (bess->tll, tl))) {
    destroy_tl (tl);
  }
  
  
  tf = NULL;
  while ((tf = tfp_l_next (bess->tfl, tf))) {
    destroy_tf (tf);
  }
  
  
  ts = NULL;
  while ((ts = tsp_l_next (bess->tsl, ts))) {
    destroy_ts (ts);
  }
  
  tr = NULL;
  while ((tr = trp_l_next (bess->trl, tr))) {
    destroy_tr (tr);
  }
  
  mfr = NULL;
  while ((mfr = mfrp_l_next (bess->mfrl, mfr))) {
    destroy_mfr (mfr);
  }
  
  mer = NULL;
  while ((mer = merp_l_next (bess->merl, mer))) {
    destroy_mer (mer);
  }
  
  free_list (bess->tvl, NOT_NODES);
  free_list (bess->tel, NOT_NODES);
  free_list (bess->tll, NOT_NODES);
  free_list (bess->tfl, NOT_NODES);
  free_list (bess->tsl, NOT_NODES);
  free_list (bess->trl, NOT_NODES);

  free_list (bess->mfrl, NOT_NODES);
  free_list (bess->merl, NOT_NODES);

  NEWP_FREE_OBJ (bess);
}


TEUP
bess_tl_get_next_teu (TLP tl, TEUP teu)
{
  return teup_l_next (tl->teul, teu);
}

TFP 
bess_tl_get_tf (TLP tl)
{
  return tl->tf;
}

int
bess_te_is_manifold (TEP te) { return te->num_inc_f == 2; }

int
bess_te_is_sheet (TEP te) { return te->num_inc_f == 1; }

int
bess_te_is_non_manifold (TEP te) { return te->num_inc_f > 2; }

void
bess_mv_get_loc (MVP mv, LOC *l)
{
  BESS_STUB_MV_GET_LOC (mv, l);
}

int
bess_me_get_loc (MEP me, LOC l[3])
{
  MVP mv[3];
  int supplied = BESS_STUB_ME_GET_MV (me, mv);
  BESS_STUB_MV_GET_LOC (mv[0], &l[0]);
  BESS_STUB_MV_GET_LOC (mv[1], &l[1]);
  if (supplied == 3)  {
    BESS_STUB_MV_GET_LOC (mv[2], &l[2]);
  }
  return supplied;
}

int
bess_mf_get_loc (MFP mf, LOC l[6])
{
  MVP mv[6];
  int supplied = BESS_STUB_MF_GET_MV (mf, mv);
  BESS_STUB_MV_GET_LOC (mv[0], &l[0]);
  BESS_STUB_MV_GET_LOC (mv[1], &l[1]);
  BESS_STUB_MV_GET_LOC (mv[2], &l[2]);
  if (supplied == 6) {
    BESS_STUB_MV_GET_LOC (mv[3], &l[3]);
    BESS_STUB_MV_GET_LOC (mv[4], &l[4]);
    BESS_STUB_MV_GET_LOC (mv[5], &l[5]);
  }
  return supplied;
}

int
bess_mf_get_mv (MFP mf, MVP mv[6])
{
  return BESS_STUB_MF_GET_MV (mf, mv);
}

int
bess_mf_get_normal (MFP mf, double xi, double eta, LOC *normal)
{
  return BESS_STUB_MF_GET_NORMAL (mf, xi, eta, normal);
}

int
bess_mv_get_id (MVP mv)
{
  return BESS_STUB_MV_GET_ID (mv);
}

int
bess_me_get_id (MEP me)
{
  return BESS_STUB_ME_GET_ID (me);
}

int
bess_mf_get_id (MFP mf)
{
  return BESS_STUB_MF_GET_ID (mf);
}

int
bess_me_get_mv (MEP me, MVP mv[3])
{
  return BESS_STUB_ME_GET_MV (me, mv);
}

MFP
bess_next_mf (BESSP bess, MFP mf)
{
  return BESS_STUB_NEXT_MF (bess->mesh, mf);
}

MEP
bess_next_me (BESSP bess, MEP me)
{
  return BESS_STUB_NEXT_ME (bess->mesh, me);
}


MVP
bess_next_mv (BESSP bess, MVP mv)
{
  return BESS_STUB_NEXT_MV (bess->mesh, mv);
}

MEP
bess_mv_get_next_me (MVP mv, MEP me)
{
  return BESS_STUB_MV_GET_NEXT_ME (mv, me);
}

MFP
bess_me_get_next_mf (MEP me, MFP mf)
{
  return BESS_STUB_ME_GET_NEXT_MF (me, mf);
}

TEUP
bess_te_get_next_teu (TEP te, TEUP teu)
{
  return teup_l_next (te->teul, teu);
}

void *
bess_mv_set_ap (MVP mv, void *ap)
{
  return BESS_STUB_MV_SET_AP (mv, ap);
}

void *
bess_mv_get_ap (MVP mv)
{
  return BESS_STUB_MV_GET_AP (mv);
}

  
void *
bess_me_set_ap (MEP me, void *ap)
{
  return BESS_STUB_ME_SET_AP (me, ap);
}

void *
bess_me_get_ap (MEP me)
{
  return BESS_STUB_ME_GET_AP (me);
}

  
int
bess_tf_get_num_tl (TFP tf)
{
  return list_length (tf->tll);
}

TLP
bess_tf_get_next_tl (TFP tf, TLP tl)
{
  return tlp_l_next (tf->tll, tl);
}

int
bess_tl_get_num_teu (TLP tl)
{
  return list_length (tl->teul);
}

int
bess_tr_get_id (TRP tr)
{
  return tr->id;
}

MRID
bess_tr_get_mrid (TRP tr)
{
  return tr->mrid;
}

int
bess_tr_get_num_ts (TRP tr)
{
  return list_length (tr->tsl);
}

int
bess_ts_get_num_tfu (TSP ts)
{
  return list_length (ts->tful);
}

TRP
bess_ts_get_tr (TSP ts)
{
  return ts->tr;
}

TSP
bess_tr_get_next_ts (TRP tr, TSP ts)
{
  return tsp_l_next (tr->tsl, ts);
}

MESHP
bess_get_mesh (BESSP bess)
{
  return bess->mesh;
}

static int
do_toggle_ts (TSP ts)
{
  TFUP tfu;
  
  if (!ts) return FALSE;

  tfu = NULL;
  while ((tfu = tfup_l_next (ts->tful, tfu))) {
    if (tfu->o == BESS_ORIENT_SAME) tfu->o = BESS_ORIENT_REVERSED;
    else                            tfu->o = BESS_ORIENT_SAME;
  }
  return TRUE;
}

int
bess_toggle_ts_orient (BESSP bess, int tsid)
{
  TSP ts = bess_find_ts (bess, tsid);
  if (!ts) return FALSE;
  else     return do_toggle_ts (ts);
}

void
bess_mf_get_me (MFP mf, /* triangle */
                MEP me[3] /* edges */
                )
{
  BESS_STUB_MF_GET_ME (mf, me);
}

void * /* previously stored pointer */
bess_mf_set_ap (MFP t, /* triangle */
                void *ap /* pointer */
                )
{
  return BESS_STUB_MF_SET_AP (t, ap);
}

int
bess_me_get_num_inc_mf (MEP e)
{
  return BESS_STUB_ME_GET_NUM_INC_MF (e);
}

void * /* associated pointer */
bess_mf_get_ap (MFP t)
{
  return BESS_STUB_MF_GET_AP (t);
}


int
bess_make_ts (BESSP bess, int tfids[], int numtfids)
{
  TFP tf;
  TFUP tfu;
  int tfid;
  int i;
  TOE o;
  TSP ts = NULL;
  
  FOR (i, numtfids) {
    tfid = tfids[i];
    o    = (tfid < 0 ? BESS_ORIENT_REVERSED : BESS_ORIENT_SAME);
    tfid = abs (tfid);
    tf   = bess_find_tf (bess, tfid);
    if (!tf) {
      REPERR (bess, "Face not found");
      return FALSE;
    }
    if (!ts) { ts = new_ts (NULL); attach_s (bess, ts); }
    tfu = new_tfu (tf, o, ts);
    conn_tf_tfu (tf, tfu);
    conn_ts_tfu (ts, tfu);
  }
  return TRUE;
}



int
bess_copy_ts (BESSP bess, int tsid, int switch_orientation)
{
  TFP tf;
  TFUP tfu;
  TFUP newtfu;
  TSP newts = NULL;
  TSP ts;
  TOE o;
  
  ts = bess_find_ts (bess, tsid);
  if (!ts) {
    REPERR (bess, "Shell not found");
    return FALSE;
  }
  
  tfu = NULL;
  while ((tfu = bess_ts_get_next_tfu (ts, tfu))) {
    tf = tfu->tf;
    if (!newts) { newts = new_ts (NULL); attach_s (bess, newts); }
    o = (switch_orientation
         ? (bess_orientation_enum) ((-1) * tfu->o)
         : tfu->o);
    newtfu = new_tfu (tf, o, newts);
    conn_tf_tfu (tf, newtfu);
    conn_ts_tfu (newts, newtfu);
  }
  return TRUE;
}

int
bess_make_tr (BESSP bess, MRID mrid, int tsids[], int numtsids)
{
  TSP ts;
  TRP tr = NULL;
  int tsid;
  int i;
  
  FOR (i, numtsids) {
    tsid = tsids[i];
    ts   = bess_find_ts (bess, tsid);
    if (!ts) {
      REPERR (bess, "Shell not found");
      return FALSE;
    }
    if (!tr) { tr = new_tr (mrid);  attach_r (bess, tr); }
    conn_tr_ts (tr, ts);
  }
  return TRUE;
}


int
bess_tf_encloses_pt (BESSP bess, TFP tf, LOC *l)
{
  /* RAW */
  return FALSE;
}

static MVP
find_mv_at_start (BESSP bess,
                  LOC *start
                  )
{
  MVP mv, cmv = NULL;
  LOC l;
  double vltol = BESS_STUB_VLTOL  (bess->mesh);

  mv = NULL;
  while ((mv = BESS_STUB_NEXT_MV (bess->mesh, mv))) {
    BESS_STUB_MV_GET_LOC (mv, &l);
    if (fabs (l.x - start->x) < vltol &&
        fabs (l.y - start->y) < vltol &&
        fabs (l.z - start->z) < vltol)
      cmv = mv;
  }

  return cmv;
}


static int
num_intersections (
                   GENWRAP_LIST_TYPE (TFP) tfl,
                   LOC *start,
                   VEC *direction,
                   double vltol,
                   MVP mv_at_start)
{
  TFP tf;
  MFRP mfrp;
  MVP fmv[6];
  int had_mv_at_start = FALSE;
  int supported, i;
  int num_intersect = 0;
  G3D_tri_t tri;
  LOC s;
  double lambda;
  double dlambda = vltol / VECMAC_NORM ((*direction));
  
  tf = NULL;
  while ((tf = tfp_l_next (tfl, tf))) {
    mfrp = NULL;
    while ((mfrp = mfrp_l_next (tf->mfrl, mfrp))) {
      supported = BESS_STUB_MF_GET_MV (mfrp->mf, fmv);
      if (mv_at_start) {        /* check if mv_at_start is used by the MF */
        FOR (i, supported) {
          if (fmv[i] == mv_at_start) { /* done for this MF; incremented counter (if possible) */
            if (!had_mv_at_start) {
              num_intersect++;
            }
            had_mv_at_start = TRUE;
            goto next_mf;
          }
        }
      }
      /* need to check for intersections */
      FOR (i, 3) {
        BESS_STUB_MV_GET_LOC (fmv[i], &tri.loc[i]);
      }
      if (G3D_intersect_tri_and_line (&s, &lambda, start, direction, &tri, 0.0001)) {
        if (lambda >= 0 || (lambda < 0 && fabs (lambda) < dlambda)) 
          num_intersect++;
      }
    next_mf:
      ;
    }
  }

  return num_intersect;
}

int
bess_num_intersections (BESSP bess,
                        LIST_OF (TFP) tfl,
                        LOC *start,
                        VEC *direction
                        )
{
  MVP mv_at_start = find_mv_at_start (bess, start);
  VEC dir = *direction;
  VECMAC_NORMALIZE (dir);
  return num_intersections (tfl, start, &dir, BESS_STUB_VLTOL (bess->mesh), mv_at_start);
}

int
bess_ts_is_manifold (TSP ts)
{
  TFUP tfu;
  tfu = NULL;
  while ((tfu = tfup_l_next (ts->tful, tfu))) {
    if (!bess_tf_is_manifold (tfu->tf)) return FALSE;
  }
  return TRUE;
}


int
bess_mer_get_id (MERP mer)
{
  return 0; /* fake function: mer's don't have any id's */
}

int
bess_mfr_get_id (MFRP mfr)
{
  return 0; /* fake function: mfr's don't have any id's */
}
