#ifndef BESS_H
#  define BESS_H


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <limits.h>
#include <sys/stat.h>
#include <unistd.h>
#include <float.h>

#if HAVE_STRING_H
#      include <string.h>
#else
#  if HAVE_STRINGS_H
#         include <strings.h>
#  endif
#endif

#include "newP.h"              /* Ckit library */
#include "ioP.h"              /* Ckit library */
#include "listP.h"              /* Ckit library */
#include "dictP.h"              /* Ckit library */
#include "hashP.h"              /* Ckit library */
#define VECMAC_VECTOR_TYPE 1
#include "vecmacP.h" 
#define WANT_VECTP_ALIASES
#include "vectP.h" /* to be able to use vector library routines. */
#include "tokensP.h"            /* Ckit library */
#include "expevalP.h" /* to be able to use expeval library routines. */
#include "gaussP.h" /* to be able to use gauss library routines. */
#include "randP.h" /* to be able to use rand library routines. */

/* ELIXIR library */
#include "Esimple.h"

/* G3D library */
#include "g3d.h"

/* PiRaT library */
#include "pirat.h"

/* LOBS library */
#include "lobs3d.h"

#ifndef FOR
#  define FOR(i_, n_) for (i_ = 0; i_ < n_; i_++)
#endif

#ifndef min
#   define min(A,B) (((A) < (B)) ? (A) : (B))
#endif
#ifndef max
#   define max(A,B) (((A) > (B)) ? (A) : (B))
#endif

#undef NDEBUG
#include <assert.h>


typedef enum bess_orientation_enum {
  BESS_ORIENT_UNSPECIFIED =  0,
  BESS_ORIENT_SAME        =  1,
  BESS_ORIENT_REVERSED    = -1
}            bess_orientation_enum;


typedef enum bess_entity_class_enum {
  BESS_CLASS_MANIFOLD     =  0,
  BESS_CLASS_SHEET        =  1,
  BESS_CLASS_NON_MANIFOLD =  2
}            bess_entity_class_enum;

typedef enum bess_option_enum {
  BESS_CREASE_ANGLE = 0,
  BESS_CORNER_ANGLE = 1
}            bess_option_enum;


typedef struct bess_opts_t {
  bess_option_enum   option;
  union {
    double         d;
    float          f;
    int            i;
    long           l;
    void          *v;
  }                   value;
}              bess_opts_t;


typedef struct bess_vertex_t *bess_vertex_p;
typedef struct bess_edge_t *bess_edge_p;
typedef struct bess_edgeuse_t *bess_edgeuse_p;
typedef struct bess_loop_t *bess_loop_p;
typedef struct bess_face_t *bess_face_p;
typedef struct bess_faceuse_t *bess_faceuse_p;
typedef struct bess_shell_t *bess_shell_p;
typedef struct bess_shelluse_t *bess_shelluse_p;
typedef struct bess_region_t *bess_region_p;
typedef struct bess_topology_t *bess_topology_p;
typedef struct bess_mesh_triangle_ref_t *bess_mesh_triangle_ref_p;
typedef struct bess_mesh_edge_ref_t *bess_mesh_edge_ref_p;

typedef int (*bess_error_notify_func) (bess_topology_p,
                                        char *file,
                                        int line,
                                        char *msg);


#undef LOC
#define LOC      G3D_loc_t 
#undef VEC
#define VEC      G3D_loc_t

struct bess_region_t;
struct bess_shell_t;
struct bess_shelluse_t;
struct bess_face_t;
struct bess_edge_t;
struct bess_edgeuse_t;
struct bess_faceuse_t;
struct bess_loop_t;
struct bess_vertex_t;
#define BESS   bess_topology_t
#define BESSP  bess_topology_p
#define TR     bess_region_t
#define TS     bess_shell_t
#define TSU    bess_shelluse_t
#define TF     bess_face_t
#define TFU    bess_faceuse_t
#define TL     bess_loop_t
#define TEU    bess_edgeuse_t
#define TE     bess_edge_t
#define TV     bess_vertex_t
#define TRP    bess_region_p
#define TSP    bess_shell_p
#define TSUP   bess_shelluse_p
#define TFP    bess_face_p
#define TFUP   bess_faceuse_p
#define TLP    bess_loop_p
#define TEUP   bess_edgeuse_p
#define TEP    bess_edge_p
#define TVP    bess_vertex_p
#define TOE    bess_orientation_enum


#define MFR    bess_mesh_triangle_ref_t
#define MER    bess_mesh_edge_ref_t
#define MFRP   bess_mesh_triangle_ref_p
#define MERP   bess_mesh_edge_ref_p

typedef void *bess_mesh_region_id_t;
#define MRID  bess_mesh_region_id_t


#endif

