#include "bess.h"
#include "bessstubs.h"
#include "bessproto.h"
#include "bess2acis.h"


static LIST 
all_mfs_in_ts (TSP ts)
{
  LIST mfl = make_list ();
  TFUP tfu;
  TFP tf;
  
  tfu = NULL;
  while ((tfu = bess_ts_get_next_tfu (ts, tfu))) {
    tf = bess_tfu_get_tf (tfu);
    append_list (mfl, bess_tf_get_mfl (tf));
  }

  return mfl;
}
  
static MFP
other_mf (MEP me, MFP mf)
{
  MFP amf;
  amf = NULL;
  while ((amf = bess_me_get_next_mf (me, amf))) {
    if (amf != mf) return amf;
  }
  return NULL;
}

static LIST
ordered_mfl (BESSP bess, TSP ts)
{
  LIST mfl = all_mfs_in_ts (ts);
  MFP mf, omf;
  STACK s = make_stack ();
  LIST omfl = make_list ();
  MEP me[3];
  int i;
  
  mf = NULL;
  while ((mf = (MFP) get_list_next (mfl, mf))) {
    bess_mf_set_ap (mf, 0);
  }

  
  mf = (MFP) get_list_next (mfl, NULL);
  bess_mf_set_ap (mf, mf);
  push_stack (s, mf);
  while (!stack_empty (s)) {
    mf = (MFP) pop_stack (s);
    add_to_tail (omfl, mf);
    bess_mf_get_me (mf, me);
    FOR (i, 3) {
      if (bess_me_get_num_inc_mf (me[i]) <= 2) {
        omf = other_mf (me[i], mf);
        if (omf && bess_mf_get_ap (omf) == 0) {
          push_stack (s, omf);
          bess_mf_set_ap (omf, omf);
        }
      } 
    }
  }

  free_list (mfl, NOT_NODES);
  free_stack (s, NOT_NODES);

  return omfl;
}


static void
write_out (LIST mfl, FILE *stream)
{
  MFP mf;
  MVP mv[6];
  LOC l;
  int i, supplied;
  int started = FALSE;
  
  mf = NULL;
  while ((mf = (MFP) get_list_next (mfl, mf))) {
    supplied = bess_mf_get_mv (mf, mv);
    fprintf (stream, "wire w ");
    FOR (i, 3) {
      bess_mv_get_loc (mv[i], &l);
      fprintf (stream, "%g %g %g ", l.x, l.y, l.z);
    }
    bess_mv_get_loc (mv[0], &l);
    fprintf (stream, "%g %g %g ", l.x, l.y, l.z);
    fprintf (stream, "\n");
    fprintf (stream, "cover w\n");
    fprintf (stream, "2d w\n");
    if (!started) {
      fprintf (stream, "copy w as b\n");
      fprintf (stream, "delete w\n");
      started = TRUE;
    } else {
      fprintf (stream, "stitch b w\n");
    }
  }
}


int
bess_to_acis (BESSP bess, FILE *stream)
{
  TSP ts;

  ts = NULL;
  while ((ts = bess_next_ts (bess, ts))) {
    write_out (ordered_mfl (bess, ts), stream);
  }
  return TRUE;
}
