#include "bess.h"
#include "bessstubs.h"
#include "bessproto.h"
#include "bess2adlib.h"


static dict_dictionary surf_p_dict = NULL;
static LIST surf_p_list = NULL;
static int ptid = 0;

static void
add_pt (dict_dictionary d, LIST l, MVP mv)
{
  MVP amv;
  amv = dict_find_obj_with_handle_name (d, mv);
  if (!amv) {
    dict_add_obj_handle_name (d, mv, mv);
    bess_mv_set_ap (mv, (void *)(++ptid));
    add_to_tail (l, mv);
  }
}

static void
build_dict_of_surface_points (BESSP bess)
{
  TEP te;
  TFP tf;
  int supplied;
  int i;
  
  surf_p_dict = dict_create_dictionary ();
  surf_p_list = make_list ();
  ptid = 0;

  {
    MEP me;
    LIST mel;
    MVP mv[3];
    
    te = NULL;
    while ((te = bess_next_te (bess, te))) {
      mel = bess_te_get_mel (te);
      me = NULL;
      while ((me = (MEP) get_list_next (mel, me))) {
        supplied = bess_me_get_mv (me, mv);
        FOR (i, supplied) {
          add_pt (surf_p_dict, surf_p_list, mv[i]);
        }
      }
      free_list (mel, NOT_NODES);
    }
  }

  {
    MFP mf;
    LIST mfl;
    MVP mv[6];
    
    tf = NULL;
    while ((tf = bess_next_tf (bess, tf))) {
      mfl = bess_tf_get_mfl (tf);
      mf = NULL;
      while ((mf = (MFP) get_list_next (mfl, mf))) {
        supplied = bess_mf_get_mv (mf, mv);
        FOR (i, supplied) {
          add_pt (surf_p_dict, surf_p_list, mv[i]);
        }
      }
      free_list (mfl, NOT_NODES);
    }
  }
}


static int
count_te (BESSP bess)
{
  TEP te;
  int count = 0;
  
  te = NULL;
  while ((te = bess_next_te (bess, te))) {
    count++;
  }
  return count;
}

static int
count_tf (BESSP bess)
{
  TFP tf;
  int count = 0;
  
  tf = NULL;
  while ((tf = bess_next_tf (bess, tf))) {
    count++;
  }
  return count;
}

static int
count_tr (BESSP bess)
{
  TRP tr;
  int count = 0;
  
  tr = NULL;
  while ((tr = bess_next_tr (bess, tr))) {
    count++;
  }
  return count;
}


static void
write_pts (BESSP bess, FILE *stream, double default_h)
{
  LOC l;
  MVP mv;

  fprintf (stream, "%d # num of surface vertices\n", dict_count (surf_p_dict));

  mv = NULL;
  while ((mv = (MVP) get_list_next (surf_p_list, mv))) {
    bess_mv_get_loc (mv, &l);
    fprintf (stream, "%g %g %g %g\n", l.x, l.y, l.z, default_h);
  }
}

static void
write_edges (BESSP bess, FILE *stream)
{
  TEUP teu;
  TEP te;
  TLP tl;
  TFP tf;
  MEP me;
  MVP mv[3];
  LIST mel;
  int supplied;
  int i, id;
  
  fprintf (stream, "%d # num of topological edges\n", count_te (bess));

  te = NULL;
  while ((te = bess_next_te (bess, te))) {
    mel = bess_te_get_mel (te);
    fprintf (stream, "%d ", list_length (mel));
    teu = NULL;
    while ((teu = bess_te_get_next_teu (te, teu))) {
      tl = bess_teu_get_tl (teu);
      tf = bess_tl_get_tf (tl);
      fprintf (stream, "%d ", bess_tf_get_id (tf));
    }
    fprintf (stream, "# edge %d\n", bess_te_get_id (te));
    me = NULL;
    while ((me = (MEP) get_list_next (mel, me))) {
      supplied = bess_me_get_mv (me, mv);
      FOR (i, supplied) {
        id = (int) bess_mv_get_ap (mv[i]);
        fprintf (stream, "%d ", id);
      }
      fprintf (stream, "  -1 -1 -1 -1\n");
    }
  }
}

static void
write_faces (BESSP bess, FILE *stream)
{
  TFP tf;
  TLP tl;
  TEUP teu;
  TEP te;
  TOE o;
  int id;
  MFP mf;
  MVP mv[6];
  int i, supplied;
  LIST mfl;
  
  fprintf (stream, "%d # num of topological faces\n", count_tf (bess));

  tf = NULL;
  while ((tf = bess_next_tf (bess, tf))) {
    fprintf (stream, "%d 0 0 0 0 0 0 # face %d\n", bess_tf_get_num_tl (tf), bess_tf_get_id (tf));
    tl = NULL;
    while ((tl = bess_tf_get_next_tl (tf, tl))) {
      fprintf (stream, "%d\n", bess_tl_get_num_teu (tl));
      teu = NULL;
      while ((teu = bess_tl_get_next_teu (tl, teu))) {
        te = bess_teu_get_te (teu);
        id = bess_te_get_id (te);
        o  = bess_teu_get_o (teu);
        fprintf (stream, "%c%d ", (o == BESS_ORIENT_SAME ? ' ' : '-'), id);
      }
      fprintf (stream, "\n");
    }
    mfl = bess_tf_get_mfl (tf);
    fprintf (stream, "%d # num of mesh faces\n", list_length (mfl));
    mf = NULL;
    while ((mf = (MFP) get_list_next (mfl, mf))) {
      supplied = bess_mf_get_mv (mf, mv);
      FOR (i, supplied) {
        id = (int) bess_mv_get_ap (mv[i]);
        fprintf (stream, "%d ", id);
      }
      fprintf (stream, "\n");
    }
  }
}

static void
write_subbody (TRP tr, BESSP bess, FILE *stream)
{
  TSP ts;
  TOE o;
  TFUP tfu;
  TFP tf;
  int id;
  
  fprintf (stream, "%d %d %d # subbody (region) %d\n",
           bess_tr_get_num_ts (tr), 1, 1, bess_tr_get_id (tr));

  ts = NULL;
  while ((ts = bess_tr_get_next_ts (tr, ts))) {
    fprintf (stream, "%d # shell %d\n", bess_ts_get_num_tfu (ts), bess_ts_get_id (ts));
    tfu = NULL;
    while ((tfu = bess_ts_get_next_tfu (ts, tfu))) {
      tf = bess_tfu_get_tf (tfu);
      id = bess_tf_get_id (tf);
      o  = bess_tfu_get_o (tfu);
      fprintf (stream, "%c%d ", (o == BESS_ORIENT_SAME ? ' ' : '-'), id);
    }
    fprintf (stream, "\n");
  }
}


static void
write_bodies (BESSP bess, FILE *stream)
{
  TRP tr;
  
  fprintf (stream, "1 # body \n");
  fprintf (stream, "%d # num of subbodies (regions)\n",
           count_tr (bess));
  tr = NULL;
  while ((tr = bess_next_tr (bess, tr))) {
    write_subbody (tr, bess, stream);
  }
}

int
bess_to_adlib (BESSP bess, FILE *stream, double default_h)
{
  build_dict_of_surface_points (bess);
  
  write_pts (bess, stream, default_h);

  write_edges (bess, stream);

  write_faces (bess, stream);
  
  write_bodies (bess, stream);
  
  dict_destroy_dictionary (surf_p_dict);
  
  return TRUE;
}
