#include "bess.h"
#include "bessstubs.h"
#include "bessproto.h"
#include "bess2targe2.h"


static int
add_pt (dict_dictionary d, MVP mv)
{
  MVP amv;
  amv = dict_find_obj_with_handle_name (d, mv);
  if (!amv) {
    dict_add_obj_handle_name (d, mv, mv);
    return 1;
  }
  return 0;
}

static int
wtf (BESSP bess, TFP tf, FILE *stream)
{
  LIST mfl;
  MVP mv[6];
  MVP amv;
  MFP mf;
  LOC l;
  int supplied;
  int nnodes, ntris;
  dict_dictionary d = dict_create_dictionary ();
  
  mfl = bess_tf_get_mfl (tf);
  
  /* Create dictionary of points used by triangles of the surface */
  nnodes = 0;
  mf = NULL;
  while ((mf = (MFP) get_list_next (mfl, mf))) {
    supplied = bess_mf_get_mv (mf, mv);
    nnodes += add_pt (d, mv[0]);
    nnodes += add_pt (d, mv[1]);
    nnodes += add_pt (d, mv[2]);
  }
    
  /* write out the totals */
  ntris = list_length (mfl);
  fprintf (stream, "%d %d\n", nnodes, ntris);

  /* Write out the points */
  fprintf (stream, "*NODES\n");
  {
    dict_iterator diter = dict_get_iterator ();
    dict_reset_iterator (diter);
    while ((amv = dict_iter_objs (d, diter))) {
      bess_mv_get_loc (amv, &l);
      fprintf (stream, "%d %g %g\n", bess_mv_get_id (amv), l.x, l.y);
    }
    dict_del_iterator (diter);
  }

  /* Write out the triangles */
  fprintf (stream, "*TR3N\n");
  ntris = 0;
  mf = NULL;
  while ((mf = (MFP) get_list_next (mfl, mf))) {
    supplied = bess_mf_get_mv (mf, mv);
    ntris++;
    fprintf (stream, "%d %d %d %d 0 0 0 tf%d\n",
             ntris, 
             bess_mv_get_id (mv[0]),
             bess_mv_get_id (mv[1]),
             bess_mv_get_id (mv[2]),
             bess_tf_get_id (tf)
             );
  }

  /* Clean up */
  dict_destroy_dictionary (d);
  free_list (mfl, NOT_NODES);
  
  return TRUE;
}




static int
write_tf (BESSP bess, TFP tf, char *name)
{
  FILE *fp;
  int result;
  
  fp = fopen (name, "w");
  if (!fp) return FALSE;
  fprintf (fp, "%s\n", name);
  result = wtf (bess, tf, fp);
  fclose (fp);
  return result;
}


int
bess_to_targe2 (BESSP bess, int tsid, char *basename)
{
  TSP ts;
  TFUP tfu;
  TFP tf;
  char name[1024];
  
  ts = bess_find_ts (bess, tsid);
  if (!ts) return FALSE;
  
  tfu = NULL;
  while ((tfu = bess_ts_get_next_tfu (ts, tfu))) {
    tf = bess_tfu_get_tf (tfu);
    sprintf (name, "%s-f%d.tg2", basename, bess_tf_get_id (tf));
    if (!write_tf (bess, tf, name)) return FALSE;
  }
  
  return TRUE;
}
