#include "bess.h"
#include "bessstubs.h"
#include "bessproto.h"
#include "bess2vdt.h"

static int
write_pts (BESSP bess, FILE *stream)
{
  MVP mv;
  LOC l;
  mv = NULL;
  while ((mv = bess_next_mv (bess, mv))) {
    bess_mv_get_loc (mv, &l);
    fprintf (stream, "pt %d %g %g %g\n", bess_mv_get_id (mv), l.x, l.y, l.z);
  }
  return TRUE;
}


static void
write_bsf (MFP mf, TOE o, int surf_id, MRID mrid[2], FILE *stream)
{
  LOC l;
  MVP mv[6];
  int i, supplied;
  int id;
  
  fprintf (stream, "bsf ");
  supplied = bess_mf_get_mv (mf, mv);
  if (o == BESS_ORIENT_SAME) {
    FOR (i, 3) {
      id = (int) bess_mv_get_id (mv[i]);
      fprintf (stream, "%d ", id);
    }
    if (supplied == 6) {
      bess_mv_get_loc (mv[3], &l);
      fprintf (stream, "%g %g %g ", l.x, l.y, l.z);
      bess_mv_get_loc (mv[4], &l);
      fprintf (stream, "%g %g %g ", l.x, l.y, l.z);
      bess_mv_get_loc (mv[5], &l);
      fprintf (stream, "%g %g %g ", l.x, l.y, l.z);
    }
    fprintf (stream, " %d %d %d\n", surf_id, (int) mrid[0], (int) mrid[1]);
  } else {
    for (i = 2; i >= 0; i--) {
      id = (int) bess_mv_get_id (mv[i]);
      fprintf (stream, "%d ", id);
    }
    if (supplied == 6) {
      bess_mv_get_loc (mv[5], &l);
      fprintf (stream, "%g %g %g ", l.x, l.y, l.z);
      bess_mv_get_loc (mv[4], &l);
      fprintf (stream, "%g %g %g ", l.x, l.y, l.z);
      bess_mv_get_loc (mv[3], &l);
      fprintf (stream, "%g %g %g ", l.x, l.y, l.z);
    }
    fprintf (stream, " %d %d %d\n", surf_id, (int) mrid[1], (int) mrid[0]);
  }
}

static void
write_faces (BESSP bess, FILE *stream)
{
  TFP tf;
  TFUP tfu;
  TOE o, o_for_one=BESS_ORIENT_SAME;
  TSP ts;
  TRP tr;
  MFP mf;
  LIST mfl;
  int surf_id;
  MRID mrid[2];
  int numtfu;

  fprintf (stream, "! boundary facets\n");

  tf = NULL;
  while ((tf = bess_next_tf (bess, tf))) {
    mfl      = bess_tf_get_mfl (tf);
    mf = NULL;
    while ((mf = (MFP) get_list_next (mfl, mf))) {
      bess_mf_set_ap (mf, tf);
    }
  }
    
  mf = NULL;
  while ((mf = bess_next_mf (bess, mf))) {
    tf       = (TFP) bess_mf_get_ap (mf);
    surf_id  = bess_tf_get_id (tf);
    numtfu   = 0;
    mrid[0] = mrid[1] = 0;
    tfu = NULL;
    while ((tfu = bess_tf_get_next_tfu (tf, tfu))) {
      o        = bess_tfu_get_o (tfu);
      ts       = bess_tfu_get_ts (tfu);
      tr       = bess_ts_get_tr (ts);
      if (tr) {
        if (o == BESS_ORIENT_SAME) mrid[0] = bess_tr_get_mrid (tr);
        else                       mrid[1] = bess_tr_get_mrid (tr);
        o_for_one = o;
        numtfu++;
      }
    }
    if (numtfu < 1) { 
      /* VOID */  /* shell with no incident region; skip it */
    } else if (numtfu == 1) {
      write_bsf (mf, o_for_one, surf_id, mrid, stream);
    } else if (numtfu == 2) {
      write_bsf (mf, BESS_ORIENT_SAME, surf_id, mrid, stream);
    } else {
      char buf[128];
      sprintf (buf, "Too many tfu's for tf %d", surf_id);
      fprintf (stderr, "%s\n", buf);
    }
  }
}

static double
char_dim (LOC loc[3])
{
  double x[3], y[3], a;
  G3D_tri_to_tri2d (&loc[0], &loc[1], &loc[2], x, y);
  a = G3D_tri2d_area (x[0], y[0], x[1], y[1], x[2], y[2]);
  return sqrt (4.0 / sqrt (3.0) * a);
}


static double
default_mesh_size (BESSP bess)
{
  MFP mf;
  MVP mv[6];
  int supplied;
  LOC loc[3];
  double avgcd = 0;
  int nt = 0;

  mf = NULL;
  while ((mf = bess_next_mf (bess, mf))) {
    supplied = bess_mf_get_mv (mf, mv);
    bess_mv_get_loc (mv[0], &loc[0]);
    bess_mv_get_loc (mv[1], &loc[1]);
    bess_mv_get_loc (mv[2], &loc[2]);
    avgcd += char_dim (loc);
    nt++;
  }
  return avgcd / (nt > 0 ? nt : 1);
}

int
bess_to_vdt (BESSP bess, FILE *stream)
{
  fprintf (stream, "mesh_size %g\n", default_mesh_size (bess));
  fprintf (stream, "gen_interior_pts\n");
  fprintf (stream, "lp_size_limit 100\n");
  write_pts (bess, stream);
  write_faces (bess, stream);
  
  return TRUE;
}
