#include "bess.h"
#include "bessstubs.h"
#include "bessproto.h"
#include "bess2wasp.h"

static int
write_pts (BESSP bess, FILE *stream)
{
  MVP mv;
  LOC l;
  MESHP mesh;

  mesh = bess_get_mesh (bess);
 
  fprintf (stream, "!  ================================================ \n");
  fprintf (stream, "NODES\n");
  mv = NULL;
  while ((mv = tet_next_mv (mesh, mv))) {
    tet_mv_get_loc (mv, &l);
    fprintf (stream, "%d %g %g %g\n", tet_mv_get_id (mv), l.x, l.y, l.z);
  }
  mv = NULL;
  while ((mv = tet_next_intmv (mesh, mv))) {
    tet_mv_get_loc (mv, &l);
    fprintf (stream, "%d %g %g %g ! int\n", tet_mv_get_id (mv), l.x, l.y, l.z);
  }
  fprintf (stream, "&\n");
  return TRUE;
}


static int
write_region_tets (TRP tr,  MESHP mesh, FILE *stream)
{
  MTP mt;
  MVP mv[10];
  int supplied, i;
  MRID mtmrid;
  MRID mrid = bess_tr_get_mrid (tr);

  fprintf (stream, "! region %d\n", (int) mrid);
  fprintf (stream, "!  ================================================ \n");
  fprintf (stream, "MATERIAL\n");
  fprintf (stream, "6\n");
  fprintf (stream, "1\n");
  fprintf (stream, "5  ! 5=halquist, 3=rfi\n");
  fprintf (stream, "1   E 117000   NU 0.35 rho 8.93e-9  sigma_y 400 HARDENING_PARAM 100\n");
  fprintf (stream, "!  END MATERIAL\n");
  fprintf (stream, "&\n");
  fprintf (stream, "!  ================================================ \n");
  fprintf (stream, "ELEM\n8\n");
  mt = NULL;
  while ((mt = tet_next_mt (mesh, mt))) {
    mtmrid = tet_mt_get_region (mt);
    if (mtmrid == mrid) {
      fprintf (stream, "%d ", tet_mt_get_id (mt));
      supplied = tet_mt_get_mv (mt, mv);
      FOR (i, supplied) {
        fprintf (stream, "%d ", tet_mv_get_id (mv[i]));
      }
      fprintf (stream, "%d\n", (int) mrid);
    }
  }
  fprintf (stream, "&\n");
  fprintf (stream, "!  ================================================ \n");
  return TRUE;
}

static int
write_tets (BESSP bess, FILE *stream)
{
  TRP tr;
  MESHP mesh;

  mesh = bess_get_mesh (bess);
                        
  tr = NULL;
  while ((tr = bess_next_tr (bess, tr))) {
    write_region_tets (tr, mesh, stream);
  }
  return TRUE;
}


static void
wasp_header (FILE *stream)
{
  fprintf (stream, "! WASP file written by TETBESS\n");
  fprintf (stream, "! (C) 1998 Petr Krysl\n");
  fprintf (stream, "*** enter comment here ***\n");
  fprintf (stream, "0 0 0 0 1\n");
  fprintf (stream, "!  ===================================================\n");
  fprintf (stream, "Initial conditions\n");
  fprintf (stream, "ID 1\n");
  fprintf (stream, "  IV \"0\" \"0\" \"0\"\n");
  fprintf (stream, "END\n");
  fprintf (stream, "&\n");
  fprintf (stream, "!  ===================================================\n");
  fprintf (stream, "TIME_FUNCTIONS\n");
  fprintf (stream, "2  ! tabular\n");
  fprintf (stream, "! func#   #pairs\n");
  fprintf (stream, "1        3\n");
  fprintf (stream, "! time     val\n");
  fprintf (stream, "0        1\n");
  fprintf (stream, "1        1\n");
  fprintf (stream, "10000000 1\n");
  fprintf (stream, "!  END TIME_FUNCTIONS\n");
  fprintf (stream, "&\n");

}

int
bess_to_wasp (BESSP bess, FILE *stream)
{
  wasp_header (stream);
  write_pts (bess, stream);
  write_tets (bess, stream);
  
  return TRUE;
}
