#include "bess.h"
#include "bessstubs.h"
#include "bessproto.h"
#include "bess2gocad.h"


static int
add_pt (LIST l, dict_dictionary d, MVP mv, int id)
{
  MVP amv;
  amv = dict_find_obj_with_handle_name (d, mv);
  if (!amv) {
    dict_add_obj_handle_name (d, mv, mv);
    add_to_tail (l, mv);
    bess_mv_set_ap (mv, (void *) id);
    id++;
  }
  return id;
}

static int
wts (BESSP bess, TSP ts, FILE *stream)
{
  LIST mfl=0;
  MVP mv[6];
  MFP mf;
  LOC l;
  int supplied;
  TFUP tfu;
  TFP tf;
  dict_dictionary vd = dict_create_dictionary ();
  LIST vl = make_list ();
  int id;

  /* Create dictionary of points used by triangles of the surface */
  id  = 0;
  tfu = NULL;
  while ((tfu = bess_ts_get_next_tfu (ts, tfu))) {
    tf = bess_tfu_get_tf (tfu);
    mfl = bess_tf_get_mfl (tf);
    mf = NULL;
    while ((mf = (MFP) get_list_next (mfl, mf))) {
      supplied = bess_mf_get_mv (mf, mv);
      id = add_pt (vl, vd, mv[0], id);
      id = add_pt (vl, vd, mv[1], id);
      id = add_pt (vl, vd, mv[2], id);
    }
  }
  
  /* Write out the points */
  fprintf (stream, "Coordinate3 { point [\n");
  {
    MVP amv;
    amv = NULL;
    while ((amv = (MVP) get_list_next (vl, amv))) {
      bess_mv_get_loc (amv, &l);
      fprintf (stream, "%g %g %g,\n", l.x, l.y, l.z);
    }
  }
  fprintf (stream, "] }\n");

  /* Write out the triangles */
  fprintf (stream, "IndexedFaceSet { coordIndex [\n");
  tfu = NULL;
  while ((tfu = bess_ts_get_next_tfu (ts, tfu))) {
    tf = bess_tfu_get_tf (tfu);
    mfl = bess_tf_get_mfl (tf);
    mf = NULL;
    while ((mf = (MFP) get_list_next (mfl, mf))) {
      supplied = bess_mf_get_mv (mf, mv);
      fprintf (stream, "%d,%d,%d,-1,\n",
               (int) bess_mv_get_ap (mv[0]),
               (int) bess_mv_get_ap (mv[1]),
               (int) bess_mv_get_ap (mv[2])
               );
    }
  }
  fprintf (stream, "]}\n");
  
  /* Clean up */
  dict_destroy_dictionary (vd);
  free_list (vl, NOT_NODES);
  free_list (mfl, NOT_NODES);
  
  return TRUE;
}




static int
write_ts (BESSP bess, TSP ts, char *name)
{
  FILE *fp;
  int result;
  
  fp = fopen (name, "w");
  if (!fp) return FALSE;
  fprintf (fp, "#VRML V1.0 ascii\n# Produced by BESS (C) 1998-99 Petr Krysl\n");
  result = wts (bess, ts, fp);
  fclose (fp);
  return result;
}


int
bess_to_wrl (BESSP bess, int tsid, char *name)
{
  TSP ts;
  
  ts = bess_find_ts (bess, tsid);
  if (!ts) return FALSE;
  
  if (!write_ts (bess, ts, name)) return FALSE;
  
  return TRUE;
}
