#ifndef BESSPROTO_H
#  define BESSPROTO_H

/*
   Iterate triangles.
 */
MFP /* NULL if there are no more triangles */
BESS_STUB_NEXT_MF (MESHP st, /* surface triangulation */
                   MFP o /* start iteration by passing NULL */
                   );

/*
   Iterate edges.
 */
MEP /* NULL if there are no more edges */
BESS_STUB_NEXT_ME (MESHP st, /* surface triangulation */
                   MEP o /* start iteration by passing NULL */
                   );

/*
   Get triangle id
 */
int
BESS_STUB_MF_GET_ID (MFP t);


/*
   Set associated pointer.
 */
void *
BESS_STUB_MV_SET_AP (MVP v, void *ap);

/*
   Set associated pointer.
 */
void * /* previously stored pointer */
BESS_STUB_MF_SET_AP (MFP mf, /* triangle */
                     void *ap /* pointer */
                     );

/*
   Get associated pointer.
 */
void *
BESS_STUB_MF_GET_AP (MFP f /* mesh face */
                     );


/*
   Set associated pointer.
 */
void *
BESS_STUB_ME_SET_AP (MEP e,     /* mesh edge */
                     void *ap   /* assoc pointer */
                     );


/*
   Get mesh face normal.
 */
int /* TRUE if normal is valid; FALSE otherwise */
BESS_STUB_MF_GET_NORMAL (MFP t, /* mesh face */
                         double xi, double eta, /* area coordinates (the third = 1 - xi - eta) */
                         LOC *normal
                         );

/*
   Get incident vertices.
 */
int /* returns number of returned nodes; 3 for a linear triangle, 6 for quadratic */
BESS_STUB_MF_GET_MV (MFP t, /* triangle */
                     MVP v[6] /* vertices in ccw order when
                                 looking against the normal;
                                 vertex 0 is shared by edges 0 and 1,
                                 vertex 1 is shared by edges 1 and 2,
                                 vertex 2 is shared by edges 2 and 0.
                                 vertex 3 is the mid-side node of edge 1,
                                 vertex 4 is the mid-side node of edge 2,
                                 vertex 5 is the mid-side node of edge 0 (the last three
                                 only for the quadratic geometry)
                                 */
                     );

/*
   Get edges.
 */
void
BESS_STUB_MF_GET_ME (MFP t, /* triangle */
                     MEP e[3] /* edges */
                     );

/*
   Get regions on plus and minus side (plus side is where the normal points).
 */
int
BESS_STUB_MF_GET_REGIONS (MFP mf, MRID *mridplus, MRID *mridminus);


/*
   Get id of mesh edge.
 */
int /* edge id */
BESS_STUB_ME_GET_ID (MEP e /* edge */
                     );

/*
   Get associated pointer.
 */
void *
BESS_STUB_ME_GET_AP (MEP e /* edge */
                     );

/*
   Iterate triangles incident to edge.
 */
MFP /* next triangle or NULL if there are no more */
BESS_STUB_ME_GET_NEXT_MF (MEP me,
                          MFP mf
                          );

/*
   Get tangent vector to an edge at a vertex.
   Tangent points from the starting vertex along the edge to the end vertex.
 */
void
BESS_STUB_ME_GET_TANGENT (MEP me, MVP mv, LOC *tangent);

/*
   Count incident mesh faces at mesh edge.
 */
int
BESS_STUB_ME_GET_NUM_INC_MF (MEP e);

/*
   Get incident vertices.
 */
int /* number of vertices returned; 2 for linear geometry, 3 for quadratic */
BESS_STUB_ME_GET_MV (MEP e,
                     MVP v[3] /* starting, ending and mid-side node (the last for quadratic geometry only) */
                     );


/*
   Get vertex id
 */
int
BESS_STUB_MV_GET_ID (MVP v);

/*
   Get associated pointer.
 */
void *
BESS_STUB_MV_GET_AP (MVP v /* vertex */
                     );

/*
   Get vertex location.
 */
void
BESS_STUB_MV_GET_LOC (MVP v, /* vertex */
                      LOC *p /* location */
                      );

/*
   Iterate edges incident to vertex
 */
MEP
BESS_STUB_MV_GET_NEXT_ME (MVP v, MEP e);


#include "bess.pt"

#endif
