 
#include "bess.h"
#include "bessstubs.h"
#include "bess.pt"
#include "getoptP.h"
#include "egfacts.h"
#include "smesh.h"
#include "mic.h"
#include "show.h"
#include "bess2vdt.h"
#include "bess2gts.h"
#include "bess2gocad.h"
#include "bess2targe2.h"
#include "bess2wrl.h"
#include "bess2acis.h"
#include "bess2adlib.h"
#include "gleb.h"
#include "refpt.h"

static void 
usage (int argc, char **argv) 
{
  char *option, *usage;

  fprintf (stderr, "usage: %s + options below\n", argv[0]);
  
  if (ckit_opt_first_recognized (&option, &usage)) {
    fprintf (stderr, "%s %s\n",
             (option ? option : "?"),
             (usage ? usage  : "?"));
  }  
  while (ckit_opt_next_recognized (&option, &usage)) {
    fprintf (stderr, "%s %s\n",
             (option ? option : "?"),
             (usage ? usage  : "?"));
  }  
  exit (0);
}  


static char *cmd_file = NULL;
static int shuffle = FALSE;
static int netscape = FALSE;

static void
parse_cmd_line (int argc, char **argv) 
{
  int help = FALSE;
  int report_opts = FALSE;
#define SET_BOOL(name) \
    {                                                             \
    if (report_opts) { fprintf (stderr, "Set: %s\n", #name); }    \
    name = TRUE;                                                \
  }  
#define SET_WITH_ARG(name, format) \
    {                                                                       \
    if (report_opts) { fprintf (stderr, "Set: %s = %" #format "\n",       \
                                #name, name); }                             \
  }  
  
  GET_OPT_BOOL (argc, argv, "-report_opts", SET_BOOL (report_opts),
                "prints options as they are being set");
  
  GET_OPT_BOOL (argc, argv, "-help", SET_BOOL (help),
                "prints usage for each option recognized");
  
  GET_OPT_BOOL (argc, argv, "-shuffle", SET_BOOL (shuffle),
                "shuffle triangles as they are being added to the triangulation");
  
  GET_OPT_ARG (argc, argv, "-c", char, cmd_file,
              SET_WITH_ARG (cmd_file, s),
              "hexvue command file");

  if (help) 
    usage (argc, argv);
}  






static void
exec_cmd_file (char *name) 
{
  const char *linebuf;
  FILE *stream;
  ckit_iobuf_t IOB = NULL;

  stream = ckit_io_open_stream (name, "r");
  if (stream) {
    IOB    = ckit_io_new_buffer (stream);
    while ((linebuf = ckit_io_scan_line (IOB, FALSE)) ) {
      ESIHandleCmd ((char *) linebuf);
    }  
    ckit_io_delete_buffer (IOB);
    fclose (stream);
  }   else   {
    fprintf (stderr, "\nCould not open command file \"%s\\n", name);
    exit (1);
  }  
}  

static void
exec_cmd_line_opts (void) 
{
  if (cmd_file) {
    exec_cmd_file (cmd_file);
  }   else   {
    ESIHandleCmd ("NEWFRAME");
  }  
}  

int  
main (int argc, char **argv) 
{
  unsigned long m = ESI_GRAPHIC_EDITOR_MASK;
  
  parse_cmd_line (argc, argv);

  ESIBuildInterface (m, argc, argv);
  ESIPopup ();

  ESIHandleCmd ("EDGECOLOR white");
  ESIHandleCmd ("EDGEFLAG off");
  ESIHandleCmd ("COLOR yellow");
  ESIHandleCmd ("MSIZE 10");
  ESIHandleCmd ("MTYPE 1");
  ESIHandleCmd ("FONT \"9x15\"");
  ESIHandleCmd ("SELCRIT inside");

  exec_cmd_line_opts ();
  ESIPopupAndRun ();

  return 1;
}  



static smesh_triangulation_p mesh = NULL;
static bess_topology_p bess = NULL;
static mi_checker_p mic = NULL;
static int presplit = FALSE;
static int subdiv_split = FALSE;
static double max_edge_len = 1;
static char *h_expr = NULL;
static int use_ref_pts = FALSE;

static void 
do_destroy (XtPointer ptr) 
{
  smesh_destroy_triangulation (mesh);
  bess_destroy (bess);
  mesh = NULL;
  bess = NULL;
}  


static void
clear_all_cb (Widget w, XtPointer ptr, XtPointer call) 
{
  ESIPopupConfirmDialog (ESITopLevelWidget (),
                            "Do you really want to destroy mesh and topo?",
                         do_destroy, NULL, NULL, NULL);
}  




static void
do_vdt (void) 
{
  char *name=0;
  FILE *fp;
  char cmd[1024];
  
  if (TypeInGetTokensTotal () >= 2) {
    name = TypeInGetTokenAsString (2);
  } else {
    fp = ESISelFile ("Select output file", "OK", "CANCEL", NULL, NULL, "w", &name);
    if (fp) {
      fclose (fp);
    }
  }

  if (name) {
    fp = fopen (name, "w");
    if (fp == NULL) {
      ERptErrMessage (1, 1, "Failed to print the data", ERROR_GRADE);
      return;
    }  
  } 

  bess_to_vdt (bess, fp);

  fclose (fp);
  if (netscape) {
    sprintf (cmd, "netscape -remote 'openURL (file:%s) '", name);
    system (cmd);
  }
}  


static void
do_gts (void) 
{
  char *name=0;
  FILE *fp;
  char cmd[1024];
  
  if (TypeInGetTokensTotal () >= 2) {
    name = TypeInGetTokenAsString (2);
  } else {
    fp = ESISelFile ("Select output file", "OK", "CANCEL", NULL, NULL, "w", &name);
    if (fp) {
      fclose (fp);
    }
  }

  if (name) {
    fp = fopen (name, "w");
    if (fp == NULL) {
      ERptErrMessage (1, 1, "Failed to print the data", ERROR_GRADE);
      return;
    }  
  } 

  bess_to_gts (bess, fp);

  fclose (fp);
  if (netscape) {
    sprintf (cmd, "netscape -remote 'openURL (file:%s) '", name);
    system (cmd);
  }
}  



static void
do_acis (void) 
{
  char *name=0;
  FILE *fp;
  char cmd[1024];
  
  fp = ESISelFile ("Select output file", "OK", "CANCEL", NULL, NULL, "w", &name);
  if (fp) {
    fclose (fp);
    fp = fopen (name, "w");
    if (fp == NULL) {
      ERptErrMessage (1, 1, "Failed to print the data", ERROR_GRADE);
      return;
    }  
  } 

  bess_to_acis (bess, fp);

  fclose (fp);
  if (netscape) {
    sprintf (cmd, "netscape -remote 'openURL (file:%s) '", name);
    system (cmd);
  }
}  


static double default_h = 1;

static void
do_adlib (void)
{
  char *name=0;
  FILE *fp;
  char cmd[1024];
  
  fp = ESISelFile ("Select output file", "OK", "CANCEL", NULL, NULL, "w", &name);
  if (fp) {
    fclose (fp);
    fp = fopen (name, "w");
    if (fp == NULL) {
      ERptErrMessage (1, 1, "Failed to print the data", ERROR_GRADE);
      return;
    }  
  } 

  bess_to_adlib (bess, fp, default_h);

  fclose (fp);
  if (netscape) {
    sprintf (cmd, "netscape -remote 'openURL (file:%s) '", name);
    system (cmd);
  }
}



static void
handle_adlib (void)
{
  if (!bess) {
    ERptStatusMessage(1, 1, "no topology");
    return;
  }
  do_adlib ();
}




static int
do_gocad (int tsid, char *name) 
{
  return bess_to_gocad (bess, tsid, name);
}  

static int
do_targe2 (int tsid, char *name) 
{
  return bess_to_targe2 (bess, tsid, name);
}  

static int
do_wrl (int tsid, char *name) 
{
  return bess_to_wrl (bess, tsid, name);
}  

static  double vltol = 0;
static double prune_elongated = 0;
static double prune_small = 0;
static int variable_vltol = FALSE;
static int use_layer_as_surf_tag = FALSE;

static int
not_too_elongated (WCRec pts[3])
{
  double x[3], y[3];
  int good, type;
  double sigma, tau;

  if (prune_elongated <= 0) return TRUE;
  
  G3D_tri_to_tri2d (&pts[0], &pts[1], &pts[2], x, y);
  G3D_triangle_shape (x[0], y[0], x[1], y[1], x[2], y[2], 6, 3, &sigma, &tau,
                      &good, &type);

  return (sqrt (sigma * sigma + tau * tau) > prune_elongated);
}

static int
not_too_small (WCRec pts[3])
{
  double x[3], y[3];
  double a;

  if (prune_small <= 0) return TRUE;
  
  G3D_tri_to_tri2d (&pts[0], &pts[1], &pts[2], x, y);
  a = G3D_tri2d_area (x[0], y[0], x[1], y[1], x[2], y[2]);
  fprintf(stderr, "a=%g, ar=%g\n", a,  prune_small * prune_small / 2);
  return (a > prune_small * prune_small / 2);
}


static void
add_tri (LIST l, GraphicObj *go) 
{
  if (shuffle) {
    int randnum = rand ();
    if ((randnum % 2) == 0) { add_to_tail (l, go); }
    else                    { add_to_head (l, go); }
  } else {
    add_to_tail (l, go);
  }
}

static double computed_vltol = FLT_MAX;
static void
update_vltol (WCRec tpnts[3])
{
  if (vltol == 0) {
    G3D_box_t box;
    double r;
    G3D_INIT_BOX (box, tpnts[0]);
    G3D_UPDT_BOX (box, tpnts[1]);
    G3D_UPDT_BOX (box, tpnts[2]);
    r = (G3D_BOX_RANGE (box, x) + G3D_BOX_RANGE (box, y) + G3D_BOX_RANGE (box, z)) / 3 / 1000;
    computed_vltol = min (computed_vltol, r);
  }
}

static double
var_vltol (WCRec tpnts[3])
{
  CkitPoint3 v;
  double mind0, d0;
  
  VECMAC_3_VEC (v, =, tpnts[2], -, tpnts[0]);
  mind0 = d0 = VECMAC_DOT (v, v);
  VECMAC_3_VEC (v, =, tpnts[0], -, tpnts[1]);
  d0 = VECMAC_DOT (v, v);
  mind0 = min (d0, mind0);
  VECMAC_3_VEC (v, =, tpnts[1], -, tpnts[2]);
  d0 = VECMAC_DOT (v, v);
  mind0 = min (d0, mind0);

  return sqrt (mind0) / 100;
}

void 
convert_bcbezs3d_to_triangles (
		    WCRec         points[4][4],  
                    LIST          add_to,
		    int           tessel_intervals
                    )
{
  FPNum u, v, delta = 1.0/tessel_intervals;
  FPNum bu[4], bv[4];
  double tmp;
  int i, j, k, m;
  GraphicObj *newgo;
  CkitPoint3 p1, p2, p3, p4, tpnts[3];
    
  for (u = 0, i = 0; i < tessel_intervals; u += delta, i++) {
    v = 0;
    CubicBernstein (u, &bu[0], &bu[1], &bu[2], &bu[3]);
    CubicBernstein (v, &bv[0], &bv[1], &bv[2], &bv[3]);
    VECMAC_SCALAR (p1, =, 0);
    for (k = 0; k < 4; k++)
      for (m = 0; m < 4; m++) {
        tmp = bu[k] * bv[m];
        VECMAC_2_VEC (p1, +=, tmp * points[k][m]);
      }
    CubicBernstein (u+delta, &bu[0], &bu[1], &bu[2], &bu[3]);
    VECMAC_SCALAR (p2, =, 0);
    for (k = 0; k < 4; k++)
      for (m = 0; m < 4; m++) {
        tmp = bu[k] * bv[m];
        VECMAC_2_VEC (p2, +=, tmp * points[k][m]);
      }
    for (v = delta, j = 0; j < tessel_intervals; v += delta, j++) {
      CubicBernstein (u, &bu[0], &bu[1], &bu[2], &bu[3]);
      CubicBernstein (v, &bv[0], &bv[1], &bv[2], &bv[3]);
      VECMAC_SCALAR (p3, =, 0);
      for (k = 0; k < 4; k++)
        for (m = 0; m < 4; m++) {
          tmp = bu[k] * bv[m];
          VECMAC_2_VEC (p3, +=, tmp * points[k][m]);
        }
      CubicBernstein (u+delta, &bu[0], &bu[1], &bu[2], &bu[3]);
      VECMAC_SCALAR (p4, =, 0);
      for (k = 0; k < 4; k++)
        for (m = 0; m < 4; m++) {
          tmp = bu[k] * bv[m];
          VECMAC_2_VEC (p4, +=, tmp * points[k][m]);
        }
      tpnts[0] = p1; tpnts[1] = p2; tpnts[2] = p3;
      newgo = CreateTriangle3D (tpnts); update_vltol (tpnts);
      add_tri (add_to, newgo);
      tpnts[0] = p2; tpnts[1] = p4; tpnts[2] = p3;
      newgo = CreateTriangle3D (tpnts); update_vltol (tpnts);
      add_tri (add_to, newgo);
      p1 = p3; p2 = p4;
    } /* for (v = delta ...)  */
  } /* for (u = 0 ...) */
}


static LIST
shuffled_output (LIST output)
{
  if (shuffle) {
    int i;
    LIST newl1, newl2;
    void *go;
    int randnum;
    
    FOR (i, 13) {
      newl1 = make_list ();
      newl2 = make_list ();
      go = NULL;
      while (!list_empty (output)) {
        randnum = rand ();
        if (randnum / 2 * 2 == randnum) go = get_from_head (output);
        else                            go = get_from_tail (output);
        randnum = rand ();
        if (randnum / 2 * 2 == randnum) add_to_tail (newl1, go);
        else                            add_to_tail (newl2, go);
      }
      free_list (output, NOT_NODES);
      output = newl1;
      append_list (output, newl2);
    }
  }
  return output;
}

static LIST
convert_to_all_triangles (LIST sel)
{
  GraphicObj *go, *newgo;
  int np;
  LIST iter;
  LIST ggl;
  LIST output = make_list ();
#define CHKM(newgo)                                             \
  if (!newgo) {                                                 \
    ERptStatusMessage (1, 1, "Failed to create triangle");      \
    return NULL;                                                \
  }

  iter = copy_list (sel, NOT_NODES);
  free_list (sel, NOT_NODES);
  
  ggl = make_list ();

 retry:
  go = NULL;
  while ((go = (GraphicObj *) get_list_next (iter, go)) ) {
    if (   EGGraphicType (go) == EG_QUAD3D
        || EGGraphicType (go) == EG_QUADWD3D) {
      WCRec pts[4];
      WCRec tpnts[3];
      if (!EGGetGraphicsGeometry (go, pts, &np)) 
        ERptStatusMessage (1, 1, "EGGetGraphicsGeometry failed");
      tpnts[0] = pts[0]; tpnts[1] = pts[1]; tpnts[2] = pts[2];
      newgo = CreateTriangle3D (tpnts);  CHKM(newgo);
      add_tri (output, newgo); update_vltol (tpnts);
      tpnts[0] = pts[2]; tpnts[1] = pts[3]; tpnts[2] = pts[0];
      newgo = CreateTriangle3D (tpnts);  CHKM(newgo);
      add_tri (output, newgo); update_vltol (tpnts);
    } else if (EGGraphicType (go) == EG_BCBEZSURFACE3D) {
      WCRec pts[16];
      WCRec pts2[4][4];
      int i, j, k;
      if (!EGGetGraphicsGeometry (go, pts, &np)) 
        ERptStatusMessage (1, 1, "EGGetGraphicsGeometry failed");
      EGCopyAttributesToActiveSet (go);
      k = 0;
      FOR (i, 4) {
        FOR (j, 4) {
          pts2[i][j] = pts[k++];
        }
      }
      convert_bcbezs3d_to_triangles (pts2, output,
                                     EASValsGetTesselIntervals ());
    } else if (   EGGraphicType (go) == EG_TRIANGLE3D
               || EGGraphicType (go) == EG_TRIANGLEWD3D) {
      WCRec tpnts[3];
      if (!EGGetGraphicsGeometry (go, tpnts, &np)) 
        ERptStatusMessage (1, 1, "EGGetGraphicsGeometry failed");
      add_tri (output, go); update_vltol (tpnts);
    } else if (EGGraphicType (go) == EG_GGROUP) {
      LIST gol = GgroupGObjects (go);
      GraphicObj *ago;
      ago = NULL;
      while ((ago = (GraphicObj *) get_list_next (gol, ago)) ) {
        add_to_tail (ggl, ago);
      }
    }  
  }

  free_list (iter, NOT_NODES);

  if (list_length (ggl) > 0) {
    iter = copy_list (ggl, NOT_NODES);
    flush_list (ggl, NOT_NODES);
    goto retry;
  }

  free_list (ggl, NOT_NODES);
  
  return shuffled_output (output);
}



static double
h_hat (void *data, double x, double y, double z)
{
  int stat;
  if (h_expr) {
    expeval_set_value("x", x);
    expeval_set_value("y", y);
    expeval_set_value("z", z);
    stat = expeval_eval (h_expr, &max_edge_len);
    if (stat != EXPEVAL_E_OK) {
      max_edge_len = FLT_MAX;
    }
  } else if (use_ref_pts) {
    CkitPoint3 loc;
    loc.x = x; loc.y = y; loc.z = z;
    max_edge_len = refpt_eval_edge_len (&loc);
  }
  return max_edge_len;
}
     

static void
do_presplit_subdiv (LIST sel)
{
  GraphicObj *go;
  WCRec pts[6];
  int np;
  LIST newsel;
#define SUBTRI(n0, n1, n2) {points[0] = pts[n0];points[1] = pts[n1];points[2] = pts[n2];}
#define NEWTRI(n0, n1, n2)                                      \
  {                                                             \
    GraphicObj *newgo;                                          \
    WCRec points[3];                                            \
    SUBTRI(n0, n1, n2);                                         \
    newgo = CreateTriangle3D (points);                          \
    EGWithMaskChangeAttributes (ALL_ATTRIB_MASK, newgo);        \
    add_to_tail (newsel, newgo);                                \
  }

  newsel = make_list ();
  
  while ((go = (GraphicObj *) get_from_head (sel))) {
    if (!EGGetGraphicsGeometry (go, pts, &np)) 
      ERptStatusMessage (1, 1, "EGGetGraphicsGeometry failed");
    EGCopyAttributesToActiveSet (go);
    VECMAC_3_VEC (pts[3], =, 0.5 * pts[0], +, 0.5 * pts[1]);
    VECMAC_3_VEC (pts[4], =, 0.5 * pts[1], +, 0.5 * pts[2]);
    VECMAC_3_VEC (pts[5], =, 0.5 * pts[2], +, 0.5 * pts[0]);
    NEWTRI(0,3,5);
    NEWTRI(3,1,4);
    NEWTRI(4,2,5);
    NEWTRI(3,4,5);
  }

  go = NULL;
  while ((go = (GraphicObj *) get_list_next (newsel, go))) {
    add_to_tail (sel, go);
  }
  
  free_list (newsel, NOT_NODES);
}

static void
do_presplit_gleb (LIST sel)
{
  GraphicObj *go;
  WCRec pts[3];
  int np;
  GLEBP gleb;
  void *prev;

  gleb = gleb_new (h_hat, NULL);
  if (!gleb) {
    ERptStatusMessage (1, 1, "gleb_new failed");
    return;
  }
  
  while ((go = (GraphicObj *) get_from_head (sel))) {
    if (!EGGetGraphicsGeometry (go, pts, &np)) 
      ERptStatusMessage (1, 1, "EGGetGraphicsGeometry failed");
    if (!gleb_add (gleb, pts)) {
      ERptStatusMessage (1, 1, "gleb_add failed");
      return;
    }
  }

  gleb_split (gleb);

  prev = NULL;
  while ((prev = gleb_next (gleb, pts, prev))) {
    go = CreateTriangle3D (pts);
    add_to_tail (sel, go);
  }
  
  gleb_free (gleb);
}

static void
do_presplit (LIST sel)
{
  if (subdiv_split)
    do_presplit_subdiv (sel);
  else
    do_presplit_gleb (sel);
}


static void
shuffle_vertices (WCRec pts[3])
{
  int rot, i;
  WCRec p;

  rot = rand () % 3;
  FOR (i, rot) { /* rotate vertices by one */
    p = pts[0];
    pts[0] = pts[2];
    pts[2] = pts[1];
    pts[1] = p;
  }
}



static void 
do_construct_smesh (LIST sel) 
{
  GraphicObj *go;
  int np;
  WCRec pts[3];
  void *surf_tag = NULL; /* no further suggestions as to which surface I belong */
  
  if (list_length (sel) <= 0) {
    ERptErrMessage (1, 2, "No objects selected.", ERROR_GRADE);
    goto exit;
  }

  sel = convert_to_all_triangles (sel);

  if (presplit) do_presplit (sel);

  if (vltol == 0) vltol = computed_vltol;
  mesh = smesh_new_triangulation (NULL, vltol);

  go = NULL;
  while ((go = (GraphicObj *) get_list_next (sel, go)) ) {
    if (   EGGraphicType (go) == EG_TRIANGLE3D
        || EGGraphicType (go) == EG_TRIANGLEWD3D) {
      if (!EGGetGraphicsGeometry (go, pts, &np)) 
        ERptStatusMessage (1, 1, "EGGetGraphicsGeometry failed");
      if (shuffle) { shuffle_vertices (pts); }
      if (not_too_elongated (pts) && not_too_small (pts)) {
        if (variable_vltol) smesh_set_vltol (mesh, var_vltol (pts));
        surf_tag = NULL;/* no further suggestions as to which surface I belong */
        if (use_layer_as_surf_tag) {
          EGCopyAttributesToActiveSet (go);
          surf_tag = (void *) EASValsGetLayer ();
        }
        if (!smesh_add_poly (mesh, pts, surf_tag)) 
          ERptStatusMessage (1, 1, "smesh_add_poly failed");
      }
    }   else {
      ERptStatusMessage (1, 1, "Non-triangle go");
    }  
  }  

 exit:
  free_list (sel, NOT_NODES);  
}  




static void 
from_view_sel_construct_smesh (EView *v_p, caddr_t data_p) 
{
  LIST sel = make_list ();
  GraphicObj *go;

  if (!sel) {
    ERptStatusMessage (1, 1, "make_list () failed");
    return;
  }  
  while ((go = EVGetFromSelectionList ()) ) {
    add_to_tail (sel, go);
  }  

  do_construct_smesh (sel);
}  

static void
construct_smesh_cb (Widget w, XtPointer ptr, XtPointer call) 
{
  EVSetSelectActionFunction (from_view_sel_construct_smesh, NULL, NULL);
  EMInstallHandler (ESIModel (), EVSelectGraphicsHandler,
                    NULL, NULL,
                    NULL, NULL,
                    NULL, NULL,
                    NULL, NULL);
  ERptStatusMessage (1, 1, "Select triangles and quadrangles");
}  

static void
write_surf_mesh_as_tecplot_cb (Widget w, XtPointer ptr, XtPointer call) 
{
  MVP mv, tmv[3];
  MFP mf;
  LOC l;
  FILE *fp;
  int nnodes, ntris;
  char *name;
  
  fp = ESISelFile ("Select TECPLOT file", "OK", "CANCEL", NULL, NULL, "w", &name);
  if (fp) {
    fclose (fp);
    fp = fopen (name, "w");
    if (!fp) {
      ERptErrMessage (1, 1, "Could not open file", ERROR_GRADE);
      return; 
    }  
    fprintf (fp, "VARIABLES = n x y z\n");
    nnodes = 0; mv = NULL; while ((mv = smesh_next_v (mesh, mv)) ) { nnodes++; }  
    ntris  = 0; mf = NULL; while ((mf = smesh_next_t (mesh, mf)) ) { ntris++; }  
    fprintf (fp, "ZONE T=egf_export, N=%d, E=%d, F=fepoint, ET=triangle\n", nnodes, ntris);
    mv = NULL;
    while ((mv = smesh_next_v (mesh, mv)) ) {
      smesh_v_get_loc (mv, &l);
      fprintf (fp, "%d   \t %g %g %g\n", smesh_v_get_id (mv), l.x, l.y, l.z);
    }  
    mf = NULL;
    while ((mf = smesh_next_t (mesh, mf)) ) {
      smesh_t_get_v (mf, tmv);
      fprintf (fp, "%d %d %d\n",
               smesh_v_get_id (tmv[0]), smesh_v_get_id (tmv[1]), smesh_v_get_id (tmv[2]));
    }  
    fclose (fp);
  }  
}  

static void
handle_ref_pt (void) 
{
  CkitPoint3 loc;
  double h_0, a;
  if (TypeInGetTokensTotal () != 6)
    { fprintf (stderr, "Expected ref_pt x y z h_0 a\n");  exit (1); }
  if (TypeInGetTokenType (2) != NUMBER)
    { fprintf (stderr, "Expected ref_pt x y z h_0 a\n");  exit (1); }
  loc.x = TypeInGetTokenAsDouble (2);
  if (TypeInGetTokenType (3) != NUMBER)
    { fprintf (stderr, "Expected ref_pt x y z h_0 a\n");  exit (1); }
  loc.y = TypeInGetTokenAsDouble (3);
  if (TypeInGetTokenType (4) != NUMBER)
    { fprintf (stderr, "Expected ref_pt x y z h_0 a\n");  exit (1); }
  loc.z = TypeInGetTokenAsDouble (4);
  if (TypeInGetTokenType (5) != NUMBER)
    { fprintf (stderr, "Expected ref_pt x y z h_0 a\n");  exit (1); }
  h_0 = TypeInGetTokenAsDouble (5);
  if (TypeInGetTokenType (6) != NUMBER)
    { fprintf (stderr, "Expected ref_pt x y z h_0 a\n");  exit (1); }
  a = TypeInGetTokenAsDouble (6);
  if (!refpt_add (&loc, h_0, a))
    { fprintf (stderr, "Expected ref_pt x y z h_0 a\n");  exit (1); }
  use_ref_pts = TRUE;
}  

static double crease_angle = 60;
static double corner_angle = 45;

static void
handle_crease_angle (void) 
{
  if (TypeInGetTokensTotal () >= 2 && TypeInGetTokenType (2) == NUMBER) 
    crease_angle = TypeInGetTokenAsDouble (2);
}  

static void
handle_presplit (void) 
{
  if (TypeInGetTokensTotal () >= 2) {
    NEWP_DUPL_STRING (h_expr, TypeInGetTokenAsString (2),
                  { h_expr = NULL; max_edge_len = FLT_MAX; }  );
    TypeInParseLine (h_expr);
    /* constant number: optimize by evaluating and caching */
    if (   TypeInGetTokensTotal () == 1
        && TypeInGetTokenType (1) == NUMBER) {
      max_edge_len = TypeInGetTokenAsDouble (1);
      if (max_edge_len <= 0) {
        max_edge_len = FLT_MAX;
        ERptStatusMessage (1, 1, "max_edge_len must be > 0");
      }
      NEWP_FREE_OBJ (h_expr); h_expr = NULL;
    }
    presplit = TRUE; 
  }
}  

static void
handle_vltol (void) 
{
  if (TypeInGetTokensTotal () >= 2 && TypeInGetTokenType (2) == NUMBER) 
    vltol = TypeInGetTokenAsDouble (2);
}  

static void
handle_variable_vltol (void) 
{
  variable_vltol = TRUE;
}  

static void
handle_use_layer_as_surf_tag (void) 
{
  use_layer_as_surf_tag = TRUE;
}  

static void
handle_subdiv_split (void) 
{
  subdiv_split = TRUE;
}  

static void
handle_corner_angle (void) 
{
  if (TypeInGetTokensTotal () >= 2 && TypeInGetTokenType (2) == NUMBER) 
    corner_angle = TypeInGetTokenAsDouble (2);
}  

static void
handle_prune_elongated (void) 
{
  if (TypeInGetTokensTotal () >= 2 && TypeInGetTokenType (2) == NUMBER) 
    prune_elongated = TypeInGetTokenAsDouble (2);
}  

static void
handle_prune_small (void) 
{
  if (TypeInGetTokensTotal () >= 2 && TypeInGetTokenType (2) == NUMBER) 
    prune_small = TypeInGetTokenAsDouble (2);
}  

static void
handle_micheck (void) 
{
  if (!bess) {
    ERptStatusMessage (1, 1, "need topology to run mi check");
    return;
  }  

  if (!mic) mic = mi_create_checker (bess);
  mi_check (mic);
}  

static void
construct_topo_cb (Widget w, XtPointer ptr, XtPointer call) 
{
  bess_opts_t opts;
  if (!mesh) {
    ERptStatusMessage (1, 1, "smesh_orient failed");
    return;
  }  
  if (!smesh_orient (mesh)) 
    ERptStatusMessage (1, 1, "smesh_orient failed");

  bess = bess_new ();

  opts.option = BESS_CREASE_ANGLE;
  opts.value.d = crease_angle;
  bess_config (bess, &opts);

  opts.option = BESS_CORNER_ANGLE;
  opts.value.d = corner_angle;
  bess_config (bess, &opts);

  bess_build (bess, mesh);
}  


static void
handle_radial_order (void)
{
  TEP te;
  TVP tv[2];
  int teid;
  if (TypeInGetTokensTotal () >= 2 && TypeInGetTokenType (2) == NUMBER) 
    teid = TypeInGetTokenAsInt (2);
  te = bess_find_te (bess, teid);
  if (te) {
    LIST tfl;
    TFP tf;
    char buf[1024] = "./tmp";
    FILE *fp;

    tmpnam (buf);
    fp = fopen (buf, "w");
    if (!fp) fp = stdout;
    
    tfl = bess_tf_in_radial_order_around (te);
    bess_te_get_tv (te, tv);
    fprintf (fp, "Radial order around TE %d (TV {%d, %d}): < ", teid,
            bess_tv_get_id (tv[0]), bess_tv_get_id (tv[1]));
    tf = NULL;
    while (tf = (TFP) get_list_next (tfl, tf)) {
      fprintf (fp, "%d ", bess_tf_get_id (tf));
    }
    fprintf (fp, ">\n");

    tf = NULL;
    while (tf = (TFP) get_list_next (tfl, tf)) {
      int which_occurence = 1;
      TFP cw_tf, ccw_tf;
      do {
        bess_tf_cw_ccw (tfl, tf, te, which_occurence, &cw_tf, &ccw_tf);
        if (cw_tf && ccw_tf) {
          fprintf (fp, "TF %d, occurrence %d: CW TF %d, CCW TF %d\n",
                   bess_tf_get_id (tf), which_occurence,
                   bess_tf_get_id (cw_tf),
                   bess_tf_get_id (ccw_tf)
                  );
        }
        which_occurence++;
      } while (cw_tf && ccw_tf);
    }

    fclose (fp);
    ESIPopupInfoFromFile (buf);
  } /* if (te) */
}

static void
handle_vdt (void) 
{
  if (!bess) {
    ERptStatusMessage (1, 1, "no topology");
    return;
  }  
  do_vdt ();
}  


static void
handle_gts (void) 
{
  if (!bess) {
    ERptStatusMessage (1, 1, "no topology");
    return;
  }  
  do_gts ();
}  


static void
handle_acis (void) 
{
  if (!bess) {
    ERptStatusMessage (1, 1, "no topology");
    return;
  }  
  do_acis ();
}  



static void
handle_gocad (void) 
{
  int i, tsid;
  char *basename;
  char name[1024];
  char buf[L_tmpnam] = "./tmp";
  FILE *fp;
  
  if (!bess) {
    ERptStatusMessage (1, 1, "no topology");
    return;
  }  

  if (TypeInGetTokensTotal () < 3) {
    ERptStatusMessage (1, 1, "Expecting args: basename tsid1 [ tsid2 ...]");
    return;
  }  

  tmpnam (buf);
  fp = fopen (buf, "w");
  if (fp == NULL) {
    ERptErrMessage (1, 1, "Failed to print the data", ERROR_GRADE);
    return;
  }  

  basename = TypeInGetTokenAsString (2);

  for (i = 3; i <= TypeInGetTokensTotal (); i++) {
    tsid = TypeInGetTokenAsInt (i);
    sprintf (name, "%s-s%d", basename, tsid);
    if (do_gocad (tsid, name)) {
      fprintf (fp, "Wrote shell %d in the GOCAD format, files \"%s-f*.xy\"\n",
               tsid, name);
    }   else   {
      fprintf (fp, "**** Failed writing shell %d in the GOCAD format\n", tsid);
    }        
  }  
  
  fclose (fp);
  ESIPopupInfoFromFile (buf);
}  


static void
handle_targe2 (void) 
{
  int i, tsid;
  char *basename;
  char name[1024];
  char buf[L_tmpnam] = "./tmp";
  FILE *fp;
  
  if (!bess) {
    ERptStatusMessage (1, 1, "no topology");
    return;
  }  

  if (TypeInGetTokensTotal () < 3) {
    ERptStatusMessage (1, 1, "Expecting args: basename tsid1 [ tsid2 ...]");
    return;
  }  

  tmpnam (buf);
  fp = fopen (buf, "w");
  if (fp == NULL) {
    ERptErrMessage (1, 1, "Failed to print the data", ERROR_GRADE);
    return;
  }  

  basename = TypeInGetTokenAsString (2);

  for (i = 3; i <= TypeInGetTokensTotal (); i++) {
    tsid = TypeInGetTokenAsInt (i);
    sprintf (name, "%s-s%d", basename, tsid);
    if (do_targe2 (tsid, name)) {
      fprintf (fp, "Wrote shell %d in the TARGE2 format, files \"%s-f*.tg2\"\n",
               tsid, name);
    }   else   {
      fprintf (fp, "**** Failed writing shell %d in the TARGE2 format\n", tsid);
    }        
  }  
  
  fclose (fp);
  ESIPopupInfoFromFile (buf);
}  

static void
handle_wrl (void) 
{
  int i, tsid;
  char *basename;
  char name[1024];
  char buf[L_tmpnam] = "./tmp";
  FILE *fp;
  
  if (!bess) {
    ERptStatusMessage (1, 1, "no topology");
    return;
  }  

  if (TypeInGetTokensTotal () < 3) {
    ERptStatusMessage (1, 1, "Expecting args: basename tsid1 [ tsid2 ...]");
    return;
  }  

  tmpnam (buf);
  fp = fopen (buf, "w");
  if (fp == NULL) {
    ERptErrMessage (1, 1, "Failed to print the data", ERROR_GRADE);
    return;
  }  

  basename = TypeInGetTokenAsString (2);

  for (i = 3; i <= TypeInGetTokensTotal (); i++) {
    tsid = TypeInGetTokenAsInt (i);
    sprintf (name, "%s-s%d.wrl", basename, tsid);
    if (do_wrl (tsid, name)) {
      fprintf (fp, "Wrote shell %d in the VRML format, files \"%s-f*.wrl\"\n",
               tsid, name);
    }   else   {
      fprintf (fp, "**** Failed writing shell %d in the VRML format\n", tsid);
    }        
  }  
  
  fclose (fp);
  ESIPopupInfoFromFile (buf);
}  

static void
print_topo (void) 
{
  char buf[1024] = "./tmp";
  FILE *fp;
  int print_to_widget = TRUE;
  
  tmpnam (buf);
  if (TypeInGetTokensTotal () >= 2) { /* print to a file */
    sprintf (buf, "%s", TypeInGetTokenAsString (2));
    print_to_widget = FALSE;
  }
  
  fp = fopen (buf, "w");
  if (fp == NULL) {
    ERptErrMessage (1, 1, "Failed to print the data", ERROR_GRADE);
    return;
  }  

  bess_print (bess, fp);

  fclose (fp);
  if (print_to_widget)
    ESIPopupInfoFromFile (buf);  
}  


static char *
str_concat (char *string1, const char *string2)
{
  char *target;
  int len = 0;
  
  if (string1 == NULL) {
    len = strlen (string2) + 1;
    string1 = NEWP_REALLOC_OBJ_ARRAY (NULL, len, char);
    if (string1) {
      sprintf (string1, "%s", string2);
      return string1;
    } else {
      return NULL;
    }
  } else {
    char *t;
    len = strlen (string1) + 1 + strlen (string2) + 1;
    t = NEWP_REALLOC_OBJ_ARRAY (string1, len, char);
    if (t) {
      sprintf (t, "%s %s", string1, string2);
      return t;
    } else {
      return NULL;
    }
  }
}
 

static void
html_topo (void) 
{
  char buf[1024] = "./tmp";
  FILE *fp;
  char cmd[1024];
  char *options = NULL;

  tmpnam (buf);
  fp = fopen (buf, "w");
  if (fp == NULL) {
    ERptErrMessage (1, 1, "Failed to print the data", ERROR_GRADE);
    return;
  }  

  if (TypeInGetTokensTotal () >= 2) {
    int i;
    const char *s;
    for (i = 2; i <= TypeInGetTokensTotal (); i++) {
      s = TypeInGetTokenAsString (i);
      options = str_concat (options, s);
    }
  }

  bess_html (bess, fp, options);

  fclose (fp);
  sprintf (cmd, "netscape -remote 'openURL (file:%s) '", buf);
  if (system (cmd) != 0) {
    system ("netscape&");
    sleep (10);
    system (cmd);
  }
}  

static void
print_topo_cb (Widget w, XtPointer ptr, XtPointer call) 
{
  if (!bess) {
    ERptErrMessage (1, 1, "Construct topo first", ERROR_GRADE);
  }   else   {
    print_topo ();
  }    
}  


static void
html_topo_cb (Widget w, XtPointer ptr, XtPointer call) 
{
  if (!bess) {
    ERptErrMessage (1, 1, "Construct topo first", ERROR_GRADE);
  }   else   {
    html_topo ();
  }    
}  

static int 
add_yourself_to_sel (NODE ptr, NODE elem) 
{
  add_to_tail ((LIST) ptr, elem);
  return 1;
}  

static void
from_model_construct_smesh (void) 
{
  EModel *m_p = ESIModel ();
  LIST sel = make_list ();

  if (!sel) {
    ERptStatusMessage (1, 1, "make_list () failed");
    return;
  }  
  EMDispatchToGraphicObjects (m_p, add_yourself_to_sel, (caddr_t) sel);

  do_construct_smesh (sel);
}  



static void
handle_smesh (void) 
{
  from_model_construct_smesh ();
}  

static void
handle_quit (void)
{
  exit (0);
}

#define TIMED_CODE(code, usedtime)                      \
   {                                                    \
     double realtime, usertime, systime;                \
     ckit_get_times (&realtime, &usertime, &systime);   \
     usedtime = usertime;                               \
     { code };                                          \
     ckit_get_times (&realtime, &usertime, &systime);   \
     usedtime = usertime - usedtime;                    \
   }

static FILE *logfp = NULL;
static char logname[L_tmpnam];
static   char logbuf[1024];

static void
update_log (char *s)
{
  if (!logfp) {
    tmpnam (logname);
    logfp = fopen (logname, "w");
  }
  if (logfp)
    fprintf (logfp, "%s\n", s);
}



static void
popup_log (void)
{
  fclose (logfp);
  ESIPopupInfoFromFile (logname);  
}


static void
handle_bess (void) 
{
  bess_opts_t opts;
  double t;
  
  if (!mesh) {
    ERptStatusMessage (1, 1, "no mesh");
    return;
  }
  
  TIMED_CODE (
              {
                if (!smesh_orient (mesh)) {
                  ERptStatusMessage (1, 1, "smesh_orient failed");
                  return;
                }  
              }, t);
  sprintf (logbuf, "Mesh orientation took %g seconds", t);
  update_log (logbuf);
  
  bess = bess_new ();

  opts.option = BESS_CREASE_ANGLE;
  opts.value.d = crease_angle;
  bess_config (bess, &opts);

  opts.option = BESS_CORNER_ANGLE;
  opts.value.d = corner_angle;
  bess_config (bess, &opts);

  TIMED_CODE ( { bess_build (bess, mesh); }, t);
  sprintf (logbuf, "BRep was built in %g seconds", t);
  update_log (logbuf);
  sprintf (logbuf, "Surface mesh: %d triangles, %d edges",
           smesh_total_of_surf_triangles (mesh),
           smesh_total_of_surf_edges (mesh));
  update_log (logbuf);

  popup_log ();
}  


static void
handle_html (void) 
{
  if (!bess) {
    ERptStatusMessage (1, 1, "no topology");
    return;
  }  
  html_topo ();
}  

static void
handle_print (void) 
{
  if (!bess) {
    ERptStatusMessage (1, 1, "no topology");
    return;
  }  
  print_topo ();
}  





static EModel *tshow = NULL;

static void
set_model (void) 
{
  int   kp;

  if (tshow == NULL) tshow = ESINamedModel ("tshow");
  
  if ((kp = TypeInIsKeywordInLine (1, "MODEL")) > 0) {
    if (TypeInGetTokenType (kp+1) == STRING) {
      tshow = ESINamedModel (TypeInGetTokenAsString (kp+1));
    }  
  }  
}  

static void
handle_show (void) 
{
  int kp;
  dict_dictionary to_include = NULL;
  dict_dictionary to_exclude = NULL;
  char *crit = NULL;
#define COLLECT_IDS(dict, startat) \
    {                                                                             \
    int _n_, _i_;                                                               \
    if (!dict) dict = dict_create_dictionary ();                                \
    for (_i_ = startat, _n_ = TypeInGetTokensTotal (); _i_ <= _n_; _i_++) {     \
      if (TypeInGetTokenType (_i_) == NUMBER) {                                 \
        int n = TypeInGetTokenAsInt (_i_);                                      \
        dict_add_obj_int_name (dict, (void *) (abs (n) +1), n);                    \
      }                                                                           \
    }                                                                             \
  }  
#define GETARGS() \
    {                                                                             \
    int kp_;                                                                    \
    kp_ = TypeInIsKeywordInLine (2, "INCLUDE");                                 \
    if (kp_ > 0) {                                                              \
      COLLECT_IDS (to_include, kp_ + 1);                                        \
    }                                                                             \
    kp_ = TypeInIsKeywordInLine (2, "EXCLUDE");                                 \
    if (kp_ > 0) {                                                              \
      COLLECT_IDS (to_exclude, kp_ + 1);                                        \
    }                                                                             \
    kp_ = TypeInIsKeywordInLine (2, "CRIT");                                    \
    if (kp_ > 0) {                                                              \
      NEWP_DUPL_STRING (crit, TypeInGetTokenAsString (kp_ + 1),   { return; }  );   \
    }                                                                             \
  }  
  
  set_model ();

  if ((kp = TypeInIsKeywordInLine (1, "TV")) > 0) {
    GETARGS ();
    if (bess) show_tv (bess, to_include, to_exclude, crit, tshow);
    else      ERptErrMessage (1, 2, "No topo.", ERROR_GRADE);
  }  
  
  if ((kp = TypeInIsKeywordInLine (1, "TV_NUMS")) > 0) {
    GETARGS ();
    if (bess) show_tv_nums (bess, to_include, to_exclude, tshow);
    else      ERptErrMessage (1, 2, "No mesh.", ERROR_GRADE);      
  }  
  
  if ((kp = TypeInIsKeywordInLine (1, "TE_NUMS")) > 0) {
    GETARGS ();
    if (bess) show_te_nums (bess, to_include, to_exclude, tshow);
    else      ERptErrMessage (1, 2, "No mesh.", ERROR_GRADE);      
  }  
  
  if ((kp = TypeInIsKeywordInLine (1, "TE")) > 0) {
    GETARGS ();
    if (bess) show_te (bess, to_include, to_exclude, crit, tshow);
    else      ERptErrMessage (1, 2, "No topo.", ERROR_GRADE);
  }  
  
  if ((kp = TypeInIsKeywordInLine (1, "TL")) > 0) {
    GETARGS ();
    if (bess) show_tl (bess, to_include, to_exclude, crit, tshow);
    else      ERptErrMessage (1, 2, "No topo.", ERROR_GRADE);
  }  
  
  if ((kp = TypeInIsKeywordInLine (1, "TF")) > 0) {
    GETARGS ();
    if (bess) show_tf (bess, to_include, to_exclude, crit, tshow);
    else      ERptErrMessage (1, 2, "No topo.", ERROR_GRADE);      
  }  
  
  if ((kp = TypeInIsKeywordInLine (1, "TF_ORIENT")) > 0) {
    GETARGS ();
    if (bess) show_tf_orient (bess, to_include, to_exclude, crit, tshow);
    else      ERptErrMessage (1, 2, "No topo.", ERROR_GRADE);      
  }  
  
  if ((kp = TypeInIsKeywordInLine (1, "TS")) > 0) {
    GETARGS ();
    if (bess) show_ts (bess, to_include, to_exclude, crit, tshow);
    else      ERptErrMessage (1, 2, "No topo.", ERROR_GRADE);      
  }  
  
  if ((kp = TypeInIsKeywordInLine (1, "TS_ORIENT")) > 0) {
    GETARGS ();
    if (bess) show_ts_orient (bess, to_include, to_exclude, crit, tshow);
    else      ERptErrMessage (1, 2, "No topo.", ERROR_GRADE);      
  }  
  
  if ((kp = TypeInIsKeywordInLine (1, "MF")) > 0) {
    if (bess) show_mf (bess, tshow);
    else      ERptErrMessage (1, 2, "No topo.", ERROR_GRADE);
  }  
  
  if ((kp = TypeInIsKeywordInLine (1, "MF_INTERFERENCE")) > 0) {
    if (bess) show_mf_interference (bess, tshow, mic);
    else      ERptErrMessage (1, 2, "No topo.", ERROR_GRADE);
  }  
  
  if ((kp = TypeInIsKeywordInLine (1, "ME")) > 0) {
    if (bess) show_me (bess, tshow);
    else      ERptErrMessage (1, 2, "No topo.", ERROR_GRADE);
  }  
  
  if ((kp = TypeInIsKeywordInLine (1, "MV")) > 0) {
    if (bess) show_mv (bess, tshow);
    else      ERptErrMessage (1, 2, "No topo.", ERROR_GRADE);
  }  
  
  if ((kp = TypeInIsKeywordInLine (1, "MF_NUMS")) > 0) {
    if (bess) show_mf_nums (bess, tshow);
    else      ERptErrMessage (1, 2, "No topo.", ERROR_GRADE);      
  }  
  
  if ((kp = TypeInIsKeywordInLine (1, "ME_NUMS")) > 0) {
    if (bess) show_me_nums (bess, tshow);
    else      ERptErrMessage (1, 2, "No topo.", ERROR_GRADE);      
  }  
  
  if ((kp = TypeInIsKeywordInLine (1, "MV_NUMS")) > 0) {
    if (bess) show_mv_nums (bess, tshow);
    else      ERptErrMessage (1, 2, "No topo.", ERROR_GRADE);      
  }  

  if (crit) NEWP_FREE_OBJ (crit);
  if (to_include) dict_destroy_dictionary (to_include);
  if (to_exclude) dict_destroy_dictionary (to_exclude);
}  

static void
handle_clear (void) 
{
  set_model ();

  EMDestroyAllGraphics (tshow);
}  

static void
handle_toggle_ts_orient (void) 
{
  int tsid;

  if (TypeInGetTokensTotal () >= 2) {
    tsid = TypeInGetTokenAsInt (2);
    if (!bess_toggle_ts_orient (bess, tsid)) {
      ERptErrMessage (1, 2, "Reversal of shell orientation failed",
                      ERROR_GRADE);
    }  
  }  
}  


static void
handle_make_ts (void) 
{
  static int *tfids = NULL;
  int numtfids = 0;
  int nt;
  int i;

  nt = TypeInGetTokensTotal ();
  i = 2;
  while (i <= nt) {
    if (TypeInGetTokenType (i) == NUMBER) {
      tfids = NEWP_REALLOC_OBJ_ARRAY (tfids, numtfids+1, int);
      tfids[numtfids++] = TypeInGetTokenAsInt (i);
    }  
    i++;
  }   
  if (numtfids > 0) {
    if (!bess_make_ts (bess, tfids, numtfids)) {
      ERptErrMessage (1, 2, "Failed to make shell", ERROR_GRADE);
    }  
  }  
}  



static void
handle_copy_ts (void) 
{
  int tsid;

  if (TypeInGetTokensTotal () >= 2) {
    if (TypeInGetTokenType (2) == NUMBER) {
      tsid = TypeInGetTokenAsInt (2);
    }  
    if (!bess_copy_ts (bess, abs (tsid), (tsid < 0))) {
      ERptErrMessage (1, 2, "Failed to copy shell", ERROR_GRADE);
    }  
  }  
}  



static void
handle_make_tr (void) 
{
  static int *tsids = NULL;
  int numtsids = 0;
  int i;
  static int mrid = 1;
  i = 2;
  while (i <= TypeInGetTokensTotal ()) {
    if (TypeInGetTokenType (i) == NUMBER) {
      tsids = NEWP_REALLOC_OBJ_ARRAY (tsids, numtsids+1, int);
      tsids[numtsids++] = TypeInGetTokenAsInt (i);
    }  
    i++;
  }  
  if (numtsids > 0) {
    if (!bess_make_tr (bess, (MRID) mrid, tsids, numtsids)) {
      ERptErrMessage (1, 2, "Failed to make region", ERROR_GRADE);
    }  
    mrid++;
  }  
}  



static char
key_pressed (XEvent *ev)
{
  char buffer[12];
  KeySym ksym;
  int lngth;
  static XComposeStatus	c_stat;
  
  if (ev->type == KeyPress) {
    lngth = XLookupString (&(ev->xkey), buffer, 12, &ksym, &c_stat);
    buffer[lngth] = '\0';
    return buffer[0];
  } else {
    return '?';
  }
}

typedef struct imake_ts_data_t {
  TFP             currtf;
  int             ntfids;
  int            *tfids;
}              imake_ts_data_t;
static imake_ts_data_t imtsd = {NULL, 0, NULL};

static void
add_tfid (int tfid)
{
  char buf[128];
  int n = imtsd.ntfids + 1;
  imtsd.tfids = NEWP_REALLOC_OBJ_ARRAY (imtsd.tfids, n, int);
  if (!imtsd.tfids) {
    abort ();
  }
  imtsd.tfids[imtsd.ntfids++] = tfid;
  sprintf (buf, "Added TF %d", tfid);
  ERptStatusMessage (1, 4, buf);
}

static void
next_tf (void)
{
  imtsd.currtf = bess_next_tf (bess, imtsd.currtf);
  show_clear (ESIModel ());
  show_te (bess, NULL, NULL, NULL, ESIModel ());
  if (imtsd.currtf) {
    int tfid;
    char buf[132];
    dict_dictionary d;
    d = dict_create_dictionary ();
    tfid = bess_tf_get_id (imtsd.currtf);
    sprintf (buf, "Considering TF %d", tfid);
    ERptStatusMessage (1, 4, buf);
    dict_add_obj_int_name (d, (void *) (tfid + 1), tfid);
    show_tf (bess, d, NULL, NULL, ESIModel ());
    show_tf_orient (bess, d, NULL, NULL, ESIModel ());
    dict_destroy_dictionary (d);
  }
}

static void 
imake_ts_handler (Widget w, EView *v_p, XEvent *event)
{
  char key;
#define SKIP   0
#define ASIS  +1
#define REV   -1
#define BOTH   2
  int action = 0;

  switch(event->type) {
  case KeyPress:
    key = key_pressed (event);
    switch (key) {
    case '+': action = ASIS; break;
    case '-': action = REV;  break;
    case 's': action = SKIP; break;
    case 'b': action = BOTH; break;
    case 'x':
      EMUninstallHandler(EVGetModel(v_p));
      return;
    default:
      EVFastViewHandler (w, v_p, event);
      return;
    }
    if (action != SKIP) {
      if (action == BOTH) {
        add_tfid (+ bess_tf_get_id (imtsd.currtf));
        add_tfid (- bess_tf_get_id (imtsd.currtf));
      } else {
        if (imtsd.currtf)
          add_tfid (action * bess_tf_get_id (imtsd.currtf));
      }
    }
    next_tf ();
    EVRedrawContents (v_p);
    if (!imtsd.currtf) {
      if (imtsd.ntfids > 0) {
        if (!bess_make_ts (bess, imtsd.tfids, imtsd.ntfids)) {
          ERptErrMessage (1, 2, "Failed to make shell", ERROR_GRADE);
        }  
      }
      ERptStatusMessage (1, 4, "Created a new shell");
      EMUninstallHandler (EVGetModel (v_p));
    }
    break;
  default:
    EVFastViewHandler (w, v_p, event);
    break;
  }
}


static void
handle_imake_ts (void) 
{
  imtsd.currtf = NULL;
  imtsd.ntfids = 0;
  imtsd.tfids  = NEWP_REALLOC_OBJ_ARRAY (imtsd.tfids, 100, int);
  if (!imtsd.tfids) { abort (); }
  next_tf ();
  EMInstallHandler (ESIModel(), imake_ts_handler,
                    NULL, NULL,
                    NULL, NULL,
                    NULL, NULL,
                    NULL, NULL);
  ERptStatusMessage (1, 4, "'+' = as is, '-' = reversed, 'b' = both, 's' = skip, 'x' = exit");
}  




typedef struct imake_tr_data_t {
  int             mrid;
  TSP             currts;
  int             ntsids;
  int            *tsids;
}              imake_tr_data_t;
static imake_tr_data_t imtrd = {NULL, 0, NULL};

static void
add_tsid (int tsid)
{
  char buf[128];
  int n = imtrd.ntsids + 1;
  imtrd.tsids = NEWP_REALLOC_OBJ_ARRAY (imtrd.tsids, n, int);
  if (!imtrd.tsids) {
    abort ();
  }
  imtrd.tsids[imtrd.ntsids++] = tsid;
  sprintf (buf, "Added TS %d", tsid);
  ERptStatusMessage (1, 4, buf);
}

static void
next_ts (void)
{
  show_clear (ESIModel ());
  show_te (bess, NULL, NULL, NULL, ESIModel ());
  if (imtrd.currts == NULL) {
    ESIHandleCmd ("FIT");
  }
  imtrd.currts = bess_next_ts (bess, imtrd.currts);
  if (imtrd.currts) {
    int tsid;
    char buf[132];
    dict_dictionary d;
    d = dict_create_dictionary ();
    tsid = bess_ts_get_id (imtrd.currts);
    sprintf (buf, "Considering TS %d", tsid);
    ERptStatusMessage (1, 4, buf);
    dict_add_obj_int_name (d, (void *) (tsid + 1), tsid);
    show_ts (bess, d, NULL, NULL, ESIModel ());
    show_ts_orient (bess, d, NULL, NULL, ESIModel ());
    dict_destroy_dictionary (d);
  }
}

static void 
imake_tr_handler (Widget w, EView *v_p, XEvent *event)
{
  char key;
#define SKIP   0
#define INCL  +1
  int action = 0;
  
  switch(event->type) {
  case KeyPress:
    key = key_pressed (event);
    switch (key) {
    case 'i': action = INCL; break;
    case 's': action = SKIP; break;
    case 'x':
      EMUninstallHandler(EVGetModel(v_p));
      return;
    default:
      EVFastViewHandler (w, v_p, event);
      return;
    }
    if (action != SKIP) {
      if (imtrd.currts)
        add_tsid (bess_ts_get_id (imtrd.currts));
    }
    next_ts ();
    EVRedrawContents (v_p);
    if (!imtrd.currts) {
      if (imtrd.ntsids > 0) {
        if (!bess_make_tr (bess, (bess_mesh_region_id_t) imtrd.mrid,
                           imtrd.tsids, imtrd.ntsids)) {
          ERptErrMessage (1, 2, "Failed to make region", ERROR_GRADE);
        }  
      }
      ERptStatusMessage (1, 4, "Created a new region");
      EMUninstallHandler (EVGetModel (v_p));
    }
    break;
  default:
    EVFastViewHandler (w, v_p, event);
    break;
  }
}

static void
handle_imake_tr (void) 
{
  if (TypeInGetTokensTotal () >= 2 &&
      TypeInGetTokenType (2) == NUMBER) {
    imtrd.mrid = TypeInGetTokenAsInt (2);
    imtrd.currts = NULL;
    imtrd.ntsids = 0;
    imtrd.tsids  = NEWP_REALLOC_OBJ_ARRAY (imtrd.tsids, 100, int);
    if (!imtrd.tsids) { abort (); }
    next_ts ();
    EMInstallHandler (ESIModel(), imake_tr_handler,
                      NULL, NULL,
                      NULL, NULL,
                      NULL, NULL,
                      NULL, NULL);
    ERptStatusMessage (1, 4, "'i' = include, 's' = skip, 'x' = exit");
  } else {
    ERptErrMessage (1, 2, "Expected region id", ERROR_GRADE);
  }
}  





#define DEFCMDNAME(name) #name, sizeof (#name) 

static top_command ctable[] =   {
    {
    DEFCMDNAME (show),
    handle_show,
    " (show part of topological and geometrical model) "
  }  ,
    {
    DEFCMDNAME (clear),
    handle_clear,
    " (clear the visualizations of the topological model) "
  }  ,
    {
    DEFCMDNAME (smesh),
    handle_smesh,
    " (generate surface mesh from the whole Elixir model) "
  }  ,
    {
    DEFCMDNAME (quit),
    handle_quit,
    " (quit program) "
  }  ,
    {
    DEFCMDNAME (bess),
    handle_bess,
    " (generate topology from the current mesh) "
  }  ,
    {
    DEFCMDNAME (html),
    handle_html,
    " (generate HTML topology documentation) "
  }   ,
    {
    DEFCMDNAME (print),
    handle_print,
    " (generate topology printout) "
  }  ,
    {
    DEFCMDNAME (crease_angle),
    handle_crease_angle,
    "# (set limit crease angle) "
  }  ,
    {
    DEFCMDNAME (vltol),
    handle_vltol,
    "# (set limit vertex location tolerance [default 0]) "
  }   ,
    {
    DEFCMDNAME (variable_vltol),
    handle_variable_vltol,
    "(use variable (per poly) vertex location tolerance)"
  }   ,
    {
    DEFCMDNAME (use_layer_as_surf_tag),
    handle_use_layer_as_surf_tag,
    "(use layer as surface tag)"
  }    ,
    {
    DEFCMDNAME (subdiv_split),
    handle_subdiv_split,
    "(use subdivision to split)"
  } ,
    {
    DEFCMDNAME (prune_elongated),
    handle_prune_elongated,
    "# (set limit shape criterion for pruning elongated triangles) "
  }  ,
    {
    DEFCMDNAME (prune_small),
    handle_prune_small,
    "# (set limit area criterion for pruning too small triangles) "
  }  ,
    {
    DEFCMDNAME (corner_angle),
    handle_corner_angle,
    "# (set corner angle) "
  }  ,
    {
    DEFCMDNAME (micheck),
    handle_micheck,
    " (run mesh-interference check) "
  }  ,
    {
    DEFCMDNAME (vdt),
    handle_vdt,
    " (write vdt input) "
  }  ,
    {
    DEFCMDNAME (gts),
    handle_gts,
    " (write GTS input) "
  }  ,
    {
    DEFCMDNAME (acis),
    handle_acis,
    " (write ACIS input) "
  }  ,
    {
    DEFCMDNAME (gocad),
    handle_gocad,
    " (write gocad input) "
  }   ,
    {
    DEFCMDNAME (targe2),
    handle_targe2,
    " (write TARGE2 input) "
  }  ,
    {
    DEFCMDNAME (wrl),
    handle_wrl,
    " (write VRML input) "
  }  ,
    {
    DEFCMDNAME (adlib),
    handle_adlib,
    " (write ADLIB brep3d.dat) "
  }  ,
    {
    DEFCMDNAME (radial_order),
    handle_radial_order,
    "# (compute radial order of TF's around a TE) "
  } ,
    {
    DEFCMDNAME (toggle_ts_orient),
    handle_toggle_ts_orient,
    "# (toggle orientation of a shell) "
  }  ,
    {
    DEFCMDNAME (make_ts),
    handle_make_ts,
    "# (make a shell) "
  }  ,
    {
    DEFCMDNAME (copy_ts),
    handle_copy_ts,
    "# (copy a shell) "
  }  ,
    {
    DEFCMDNAME (make_tr),
    handle_make_tr,
    "# (make a region) "
  }  ,
    {
    DEFCMDNAME (imake_ts),
    handle_imake_ts,
    "# (interactively make a shell) "
  }   ,
    {
    DEFCMDNAME (imake_tr),
    handle_imake_tr,
    "# (interactively make a region) "
  }  ,
    {
    DEFCMDNAME (presplit),
    handle_presplit,
    "# (presplit graphic objects to specified edge length) "
  }   ,
    {
    DEFCMDNAME (ref_pt),
    handle_ref_pt,
    "# # # # # (define refinement point) "
  }     
}  ;


static smesh_triangulation_p level_mesh = NULL;

static void 
do_surf_level (EView *v_p, caddr_t data_p) 
{
  GraphicObj *go = NULL;
  int i, maxn, nv;
  WCRec pts[8], v;
  int np;
  int ngo;
  LOC oldloc, newloc, loc;
  MVP mv, mva[6], mvtomove;
  MFP mf;
  void *ap;
  
  if ((ngo = EVItemsInSelectionList ()) < 3) {
    ERptErrMessage (1, 2, "Not enough objects selected.", ERROR_GRADE);
    goto exit;
  } 

  if (level_mesh) 
    smesh_destroy_triangulation (level_mesh);

  level_mesh = smesh_new_triangulation (NULL, 0);
  
  while ((go = EVGetFromSelectionList ()) ) {
    if (EGGraphicType (go) == EG_TRIANGLE3D) {
      if (!EGGetGraphicsGeometry (go, pts, &np)) 
        ERptStatusMessage (1, 1, "EGGetGraphicsGeometry failed");
      smesh_add_poly_with_ap (level_mesh, pts, go, NULL);
    }  
  }
  
  mv = NULL;
  while ((mv = smesh_next_v (level_mesh, mv))) {
    smesh_v_set_ap (mv, 0);
  }

  mf = NULL;
  while ((mf = smesh_next_t (level_mesh, mf))) {
    smesh_t_get_v (mf, mva);
    for (i = 0; i < 3; i++) {
      ap = smesh_v_get_ap (mva[i]);
      smesh_v_set_ap (mva[i], (void *) ((int) ap + 1));
    }
  }

  mvtomove = NULL;
  maxn     = 0;
  nv       = 0;
  mv       = NULL;
  while ((mv = smesh_next_v (level_mesh, mv))) {
    ap = smesh_v_get_ap (mv);
    if ((int)ap > maxn) {
      maxn = (int)ap;
      mvtomove = mv;
    }
    nv++;
  }

  if (mvtomove) {
    nv = nv - 1;
    if (nv < 3) goto exit;
    VECMAC_SCALAR (newloc, =, 0);
    mv = NULL;
    while ((mv = smesh_next_v (level_mesh, mv))) {
      smesh_v_get_loc (mv, &loc);
      if (mv == mvtomove) { oldloc = loc; }
      else { VECMAC_2_VEC (newloc, +=, loc); }
    }
    VECMAC_SCALAR (newloc, *=, (1.0 / nv));
    smesh_v_set_loc (mvtomove, &newloc);

    mf = NULL;
    while ((mf = smesh_next_t (level_mesh, mf))) {
      go = (GraphicObj *) smesh_t_get_ap (mf);
      smesh_t_get_v (mf, mva);
      for (i = 0; i < 3; i++) {
        smesh_v_get_loc (mva[i], &pts[i]);
      }
      EGEraseGraphics (v_p, go);
      EGModifyGraphicsGeometry (go, pts);
    }
  }

  EVRedrawContents (v_p);
 exit :
  ERptStatusMessage (1, 1, "Done");
} 


static void
surf_level_cb (Widget w, XtPointer ptr, XtPointer call) 
{
  EVSetSelectActionFunction (do_surf_level, NULL, NULL);
  EMInstallHandler (ESIModel (), EVSelectGraphicsHandler,
		      NULL, NULL,
		      NULL, NULL,
		      NULL, NULL,
		      NULL, NULL);
  ERptStatusMessage (1, 1, "Select triangles");
}  




static double
quad_split_h_hat (void *data, double x, double y, double z)
{
  return *((double *)(data));
}



static double
get_mel (GraphicObj *go[2])
{
  double mel, el;
  WCRec pts[8], v;
  int i, np;
  mel = 0;
  FOR (i, 2) {
    EGGetGraphicsGeometry (go[i], pts, &np);
    VECMAC_3_VEC (v, =, pts[2], -, pts[0]);
    el = VECMAC_NORM (v);
    mel = max (mel, el);
    VECMAC_3_VEC (v, =, pts[0], -, pts[1]);
    el = VECMAC_NORM (v);
    mel = max (mel, el);
    VECMAC_3_VEC (v, =, pts[1], -, pts[2]);
    el = VECMAC_NORM (v);
    mel = max (mel, el);
  }
  return mel;
}

static double
sqdist (WCRec *p0, WCRec *p1)
{
  WCRec v;                                    
  VECMAC_3_VEC (v, =, (*p1), -, (*p0));             
  return  VECMAC_DOT (v, v);                      
}

typedef struct ptset_t {
  double tol;
  WCRec  pt[6];
  int    npt;
} ptset_t;

static int
in_ptset (ptset_t *set, WCRec *toadd)
{
  int i;
  FOR (i, set->npt) {
    if (sqdist (&set->pt[i], toadd) < set->tol) {
      return TRUE;
    }
  }
  return FALSE;
}

static void
add_to_ptset (ptset_t *set, WCRec *toadd)
{
  if (in_ptset (set, toadd))
    return;
  set->pt[set->npt] = *toadd;
  set->npt++;
}

static void
init_ptset (ptset_t *set, double tol)
{
  set->tol = tol;
  set->npt = 0;
}

static int
count_ptset (ptset_t *set)
{
  return set->npt;
}

static int
get_quad_verts (GraphicObj *go[2], WCRec sharedv[2], WCRec freev[2])
{
  WCRec pts[6];
  ptset_t freeset, sharedset;
  int np,i,j;
  double mel = get_mel (go);
  double tol = mel / 1000;
  EGGetGraphicsGeometry (go[0], &pts[0], &np);
  EGGetGraphicsGeometry (go[1], &pts[3], &np);

  init_ptset (&freeset, tol);
  init_ptset (&sharedset, tol);
  
  /* build the shared set */
  FOR (i, 6) {
    FOR (j, 6) {
      if (i != j) {
        if (sqdist (&pts[i], &pts[j]) < tol) {
          add_to_ptset (&sharedset, &pts[i]);
          break;
        }
      }
    }
  }
  if (count_ptset (&sharedset) != 2) return FALSE;

  /* build the free set */
  FOR (i, 6) {
    if (! in_ptset (&sharedset, &pts[i])) {
      add_to_ptset (&freeset, &pts[i]);
    }
  }
  if (count_ptset (&freeset) != 2) return FALSE;

  /* copy into buffers */
  FOR (i, 2) {
    sharedv[i] = sharedset.pt[i];
    freev[i]   = freeset.pt[i];
  }
  return TRUE;
}

     
static void 
do_quad_split (EView *v_p, caddr_t data_p) 
{
  GraphicObj *go[2] = {NULL,NULL}, *newgo;
  int i, np;
  WCRec sharedv[2], freev[2];
  
  if (EVItemsInSelectionList () != 2) {
    ERptErrMessage (1, 2, "Select just two triangles.", ERROR_GRADE);
    goto exit;
  } 

  FOR (i, 2) {go[i] = EVGetFromSelectionList (); }
  
  EGCopyAttributesToActiveSet (go[0]);

  if (get_quad_verts (go, sharedv, freev)) {
    WCRec modpts[3], center;
    VECMAC_3_VEC (center, =, 0.5 * sharedv[0], +, 0.5 * sharedv[1]);
    
    FOR (i, 2) { EGEraseGraphics (v_p, go[i]); EMDeleteGraphics (ESIModel (), go[i]); }
  
    modpts[0] = freev[0];
    modpts[1] = sharedv[1];
    modpts[2] = center;
    newgo = CreateTriangle3D (modpts);
    EGWithMaskChangeAttributes (ALL_ATTRIB_MASK, newgo);
    EMNoDrawAddGraphicsToModel (ESIModel (), newgo);
    
    modpts[0] = freev[1];
    modpts[1] = sharedv[1];
    modpts[2] = center;
    newgo = CreateTriangle3D (modpts);
    EGWithMaskChangeAttributes (ALL_ATTRIB_MASK, newgo);
    EMNoDrawAddGraphicsToModel (ESIModel (), newgo);

    modpts[0] = freev[1];
    modpts[1] = sharedv[0];
    modpts[2] = center;
    newgo = CreateTriangle3D (modpts);
    EGWithMaskChangeAttributes (ALL_ATTRIB_MASK, newgo);
    EMNoDrawAddGraphicsToModel (ESIModel (), newgo);

    modpts[0] = freev[0];
    modpts[1] = sharedv[0];
    modpts[2] = center;
    newgo = CreateTriangle3D (modpts);
    EGWithMaskChangeAttributes (ALL_ATTRIB_MASK, newgo);
    EMNoDrawAddGraphicsToModel (ESIModel (), newgo);
  }

  EVRedrawContents (v_p);
 exit :
  ERptStatusMessage (1, 1, "Done");
} 

static void
quad_split_cb (Widget w, XtPointer ptr, XtPointer call) 
{
  EVSetSelectActionFunction (do_quad_split, NULL, NULL);
  EMInstallHandler (ESIModel (), EVSelectGraphicsHandler,
		      NULL, NULL,
		      NULL, NULL,
		      NULL, NULL,
		      NULL, NULL);
  ERptStatusMessage (1, 1, "Select two triangles");
}  

     
static void 
do_quad_swap (EView *v_p, caddr_t data_p) 
{
  GraphicObj *go[2] = {NULL,NULL}, *newgo;
  int i, np;
  WCRec pts[8];
  void *prev;
  WCRec sharedv[2], freev[2];
  double mel;
  
  if (EVItemsInSelectionList () != 2) {
    ERptErrMessage (1, 2, "Select just two triangles.", ERROR_GRADE);
    goto exit;
  } 

  FOR (i, 2) {
    go[i] = EVGetFromSelectionList ();
  }
  
  if (get_quad_verts (go, sharedv, freev)) {
    WCRec modpts[3];
    modpts[0] = freev[0];
    modpts[1] = sharedv[1];
    modpts[2] = freev[1];
    EGEraseGraphics (v_p, go[0]);
    EGModifyGraphicsGeometry (go[0], modpts);
    modpts[0] = freev[0];
    modpts[1] = sharedv[0];
    modpts[2] = freev[1];
    EGEraseGraphics (v_p, go[1]);
    EGModifyGraphicsGeometry (go[1], modpts);
  }

  EVRedrawContents (v_p);
 exit :
  ERptStatusMessage (1, 1, "Done");
} 

static void
quad_swap_cb (Widget w, XtPointer ptr, XtPointer call) 
{
  EVSetSelectActionFunction (do_quad_swap, NULL, NULL);
  EMInstallHandler (ESIModel (), EVSelectGraphicsHandler,
		      NULL, NULL,
		      NULL, NULL,
		      NULL, NULL,
		      NULL, NULL);
  ERptStatusMessage (1, 1, "Select two triangles");
}  



void
ESICustomize (Widget parent_pane) 
{
  Arg args[3];
  int argn;
  Widget w, p;

  p = parent_pane;
  argn = 0;
  w    = ESIAddButton ("construct_surface_triangulation",
                      "Construct surface mesh",
                      commandWidgetClass,
                      p, args, argn, construct_smesh_cb, NULL);
  argn = 0;
  w    = ESIAddButton ("construct_topology",
                      "Construct topology",
                      commandWidgetClass,
                      p, args, argn, construct_topo_cb, NULL);
  argn = 0;
  w    = ESIAddButton ("print_topology",
                      "Print topology",
                      commandWidgetClass,
                      p, args, argn, print_topo_cb, NULL);
  argn = 0;
  w    = ESIAddButton ("html_topology",
                      "HTML topology",
                      commandWidgetClass,
                      p, args, argn, html_topo_cb, NULL);
  
  argn = 0;
  w    = ESIAddButton ("write_surf_mesh_as_tecplot",
                      "Write surf mesh for TECPLOT",
                      commandWidgetClass,
                      p, args, argn, write_surf_mesh_as_tecplot_cb, NULL);
  argn = 0;
  w    = ESIAddButton ("clear",
                      "Clear all",
                      commandWidgetClass,
                      p, args, argn, clear_all_cb, NULL);
  
  argn = 0;
  w    = ESIAddButton ("surf_levelling_tool",
                      "Surface levelling",
                      commandWidgetClass,
                      p, args, argn, surf_level_cb, NULL);
  
  argn = 0;
  w    = ESIAddButton ("quad_split_tool",
                      "Quad split",
                      commandWidgetClass,
                      p, args, argn, quad_split_cb, NULL);
  
  argn = 0;
  w    = ESIAddButton ("quad_diag_swap_tool",
                      "Quad diag swap",
                      commandWidgetClass,
                      p, args, argn, quad_swap_cb, NULL);
  
  TypeInAugmentCommandTable (ctable, XtNumber (ctable));
}  




