#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <limits.h>
#include <ioP.h>
#include <tokensP.h>
#include <newP.h>
#define VECMAC_VECTOR_TYPE 1
#include "vecmacP.h" 
#define WANT_VECTP_ALIASES
#include "vectP.h" /* to be able to use vector library routines. */



/*
cc -o pdftoegf -I../Ckit/include pdftoegf.c -L../Ckit/lib/mips-sgi-irix6.4 -lckit -lm
 */

static char *prog = NULL;

static int
had_v (int nv, double eps, CkitPoint3 *pt, CkitPoint3 *vertices)
{
  int i;
  for (i = 0; i < nv; i++) {
    if (   fabs (pt->x - vertices[i].x) < eps
        && fabs (pt->y - vertices[i].y) < eps
        && fabs (pt->z - vertices[i].z) < eps)
      return 1;
  }
  return 0;
}

#define ELIXIR_FILE_HEADER  "# ELIXIR file (C) 1994 Petr Krysl; version "
#define ELIXIR_FILE_VERSION 1
#ifndef EG_QUAD3D
#define EG_QUAD3D 15
#endif
#ifndef EG_TRIANGLE3D
#define EG_TRIANGLE3D 7
#endif


static void
egf_graphic_header (int timeid, int serial, int type, int layer)
{
  printf ("%lu %lu %d %d %d %d\n", timeid, serial, type, layer, 1, 0);
}

static CkitPoint3 *allverts = NULL;
static CkitPoint3 *verts = NULL;
static int np = 0;

static double
compute_eps (CkitPoint3 *v, int nv)
{
  double x, y, z;
  int i;
  if (nv < 2) return 0;
  x = fabs (v[nv-1].x - v[0].x);
  y = fabs (v[nv-1].y - v[0].y);
  z = fabs (v[nv-1].z - v[0].z);
  for (i = 1; i < nv-1; i++) {
    x = max (x, fabs (v[i+1].x - v[i].x));
    y = max (y, fabs (v[i+1].y - v[i].y));
    z = max (z, fabs (v[i+1].z - v[i].z));
  }
  return (x + y + z) / 1000;
}

static int
get_poly (tokens_parser parser)
{
  int i, n, nv;
  CkitPoint3 pt;
  double eps;
  
  nv = tokens_token_as_int (parser, 1);

  if (nv <= 0) {
    fprintf (stderr, "%s: Non-positive number of vertices, line %d\n",
             (prog ? prog : "?"),
             tokens_curr_line_num (parser));
    exit (1);
  }
  
  if (np < nv) {
    np = 2 * nv; 
    allverts = NEWP_REALLOC_OBJ_ARRAY (allverts, np, CkitPoint3);
    if (!allverts) {
      fprintf (stderr, "%s: Not enough memory\n",
             (prog ? prog : "?"));
      exit (1);
    }
    verts = NEWP_REALLOC_OBJ_ARRAY (verts, np, CkitPoint3);
    if (!verts) {
      fprintf (stderr, "%s: Not enough memory\n",
             (prog ? prog : "?"));
      exit (1);
    }
  }
  
  for (i = 0; i < nv; i++) {
    VECMAC_SCALAR (allverts[i], =, -FLT_MAX);
    if (!tokens_next_line (parser)) {
      fprintf (stderr, "%s: Failed to read coordinate data, line %d\n",
               (prog ? prog : "?"),
               tokens_curr_line_num (parser));
      exit (1);
    }
    if (tokens_total_of_tokens (parser) < 3) {
      fprintf (stderr, "%s: Expected 3 numbers, line %d.  Got \"%s\"\n",
               (prog ? prog : "?"),
               tokens_curr_line_num (parser),
               tokens_curr_line_contents (parser));
      exit (1);
    }
    allverts[i].x = tokens_token_as_double (parser, 1);
    allverts[i].y = tokens_token_as_double (parser, 2);
    allverts[i].z = tokens_token_as_double (parser, 3);
  }

  eps = compute_eps (allverts, nv);
  
  n = 0;
  for (i = 0; i < nv; i++) {
    if (!had_v (n, eps, &allverts[i], verts)) {
      VECMAC_2_VEC (verts[n], =, allverts[i]);
      n++;
    }
  }
  
  if (n == 3) {
    egf_graphic_header (0, 0, EG_TRIANGLE3D, 0);
    printf ("%f %f %f %f %f %f %f %f %f\n%d %d %d %f\n%s\n%s\n",
            verts[0].x, verts[0].y, verts[0].z, 
            verts[1].x, verts[1].y, verts[1].z,
            verts[2].x, verts[2].y, verts[2].z,
            1, 0, 1, 1.0,
            "white", "black");
  } else if (n == 4) {
    egf_graphic_header (0, 0, EG_QUAD3D, 0);
    printf ("%f %f %f %f %f %f %f %f %f %f %f %f\n%d %d %f %d\n%s\n%s\n",
            verts[0].x, verts[0].y, verts[0].z, 
            verts[1].x, verts[1].y, verts[1].z,
            verts[2].x, verts[2].y, verts[2].z,
            verts[3].x, verts[3].y, verts[3].z,
            1, 0, 1.0, 1,
            "white", "black");
  } else {
    CkitPoint3 c = {0,0,0}, pts[3];
    int i0, i1;
    for (i = 0; i < n; i++) {
      VECMAC_2_VEC (c, +=, verts[i]);
    }
    VECMAC_SCALAR (c, /=, n);
    VECMAC_2_VEC (pts[0], =, c);
    for (i = 0; i < n; i++) {
      i0 = i;
      i1 = (i0 + 1 == n ? 0 : i0 + 1);
      VECMAC_2_VEC (pts[1], =, verts[i0]);
      VECMAC_2_VEC (pts[2], =, verts[i1]);
      egf_graphic_header (0, 0, EG_TRIANGLE3D, 0);
      printf ("%f %f %f %f %f %f %f %f %f\n%d %d %d %f\n%s\n%s\n",
              verts[0].x, verts[0].y, verts[0].z, 
              verts[1].x, verts[1].y, verts[1].z,
              verts[2].x, verts[2].y, verts[2].z,
              1, 0, 1, 1.0,
              "white", "black");
    }
  }

  return 1;
}

int
main (int argc, char **argv)
{
  tokens_parser parser;

  prog = argv[0];
  
  parser = tokens_new_parser ();

  printf (ELIXIR_FILE_HEADER "%d\n", ELIXIR_FILE_VERSION);

  if (!tokens_open_file (parser, "-")) {
    fprintf (stderr, "\n%s: Cannot read stdin\n",
               (prog ? prog : "?"));
    exit (1);
  }

  tokens_next_line (parser);    /* skip the bounding box */
  tokens_next_line (parser);    /* skip the total of polygons */

  while (tokens_next_line (parser)) {
    if (tokens_total_of_tokens (parser) == 1) {
      get_poly (parser);
    }
  }
}
