
#include "bess.h"
#include "refpt.h"

static LIST refpts = NULL;

typedef struct refpt_t {
  CkitPoint3 loc;
  double h_0, a;
}              refpt_t;

/*
  Add a refinement point.  The edge length is defined as
  h(x) = h_0 + a * r^2
 */
int
refpt_add (CkitPoint3 *loc, double h_0, double a)
{
  refpt_t *rp;
  if (!refpts) refpts = make_list ();
  rp = NEWP_OBJ (refpt_t);
  if (!rp)
    return FALSE;
  rp->loc = *loc;
  rp->h_0 = h_0;
  rp->a   = a;
  add_to_tail (refpts, rp);
  return TRUE;
}

/*
  Evaluate edge length.
 */
double
refpt_eval_edge_len (CkitPoint3 *loc)
{
  refpt_t *rp;
  double minel = FLT_MAX, el, r2;
  CkitPoint3 d;
  
  rp = NULL;
  while ((rp = (refpt_t *) get_list_next (refpts, rp))) {
    VECMAC_3_VEC (d, =, (*loc), -, rp->loc);
    r2 = VECMAC_DOT (d, d);
    el = rp->h_0 + rp->a * r2;
    minel = min (minel, el);
  }
  return minel;
}
