
#include "bess.h"
#include "bessstubs.h"
#include "bess.pt"
#include "mic.h"
#include "show.h"
#include "color.h"

static int
include_entity (dict_dictionary to_include, dict_dictionary to_exclude, int entity_id)
{
  int should_be_included = FALSE;
  if (to_include) {
    if (dict_find_obj_with_int_name (to_include, entity_id))
      should_be_included = TRUE;
    else
      should_be_included = FALSE;
  }
  if (to_exclude) {
    if (dict_find_obj_with_int_name (to_exclude, entity_id))
      should_be_included = FALSE;
    else
      should_be_included = TRUE;
  }
  return should_be_included;
}

static void
tv_grep (TVP tv, EModel *model)
{
  WCRec pts[1];
  GraphicObj *go;
  
  bess_mv_get_loc (bess_tv_get_mv (tv), &pts[0]);
  go = CreateMarker3D (pts);
  EGWithMaskChangeAttributes (COLOR_MASK | MSIZE_MASK | MTYPE_MASK, go);
  EMNoDrawAddGraphicsToModel (model, go);
}

int
show_tv (BESSP topo, dict_dictionary to_include, dict_dictionary to_exclude, char *crit, EModel *model)
{
  TVP tv;

  ESIHandleCmd ("color orange");
  ESIHandleCmd ("msize 6");
  ESIHandleCmd ("mtype 3");

  tv = NULL;
  while ((tv = bess_next_tv (topo, tv))) {
    if (to_include || to_exclude) {
      if (include_entity (to_include, to_exclude, bess_tv_get_id (tv))) 
        tv_grep (tv, model);
    } else {
      tv_grep (tv, model);
    }
  }

  return 1;
}

static void
te_grep (TEP te, EModel *model)
{
  LIST_OF (MEP) mel;
  EPixel p = color_from_id (bess_te_get_id (te));
  WCRec pts[3], pnts[2];
  MEP me;
  GraphicObj *go;
  int supplied;

  EASValsSetColor (p);
  mel = bess_te_get_mel (te);
  loop_through_list (mel, me, MEP) {
    supplied = bess_me_get_loc (me, pts);
    if (supplied == 3) {
      pnts[0] = pts[0];
      pnts[1] = pts[2];
      go = CreateLine3D (pnts);
      EGWithMaskChangeAttributes (COLOR_MASK, go);
      EMNoDrawAddGraphicsToModel (model, go);
      pnts[0] = pts[2];
      pnts[1] = pts[1];
      go = CreateLine3D (pnts);
      EGWithMaskChangeAttributes (COLOR_MASK, go);
      EMNoDrawAddGraphicsToModel (model, go);
    } else {
      go = CreateLine3D (pts);
      EGWithMaskChangeAttributes (WIDTH_MASK | COLOR_MASK, go);
      EMNoDrawAddGraphicsToModel (model, go);
    }
  }
  free_list (mel, NOT_NODES);
}

int 
show_clear (EModel *model)
{
  EMDestroyAllGraphics (model);
  return 1;
}


static int
te_pass_crit (char *crit, TEP te, BESSP topo)
{
  static tokens_parser P = NULL;

  if (!P) P = tokens_new_parser ();

  if (!crit) return TRUE;
  
  if (!tokens_parse_line (P, crit)) return TRUE;

  if (tokens_have_keyword (P, 1, "MANIFOLD")) {
    return bess_te_is_manifold (te);
  }
  if (tokens_have_keyword (P, 1, "NON-MANIFOLD")) {
    return bess_te_is_non_manifold (te);
  }
  if (tokens_have_keyword (P, 1, "SHEET")) {
    return bess_te_is_sheet (te);
  }
  return TRUE;
}

int
show_te (BESSP topo, dict_dictionary to_include, dict_dictionary to_exclude, char *crit, EModel *model)
{
  TEP te;

  te = NULL;
  while ((te = bess_next_te (topo, te))) {
    if (to_include || to_exclude) {
      if (include_entity (to_include, to_exclude, bess_te_get_id (te)))
        te_grep (te, model);
    } else {
      if (te_pass_crit (crit, te, topo))
        te_grep (te, model);
    }
  }

  return 1;
}


static void
tf_grep (TFP tf, EModel *model, TOE o)
{
  LIST_OF (MFP) mfl;
  EPixel p = color_from_id (bess_tf_get_id (tf));
  WCRec pts[6];
  MFP mf;
  GraphicObj *go;
  int supplied;
#define SWAPLOC(l0, l1) { LOC al = (l0); (l0) = (l1); (l1) = al; }

  EASValsSetColor (p);
  EASValsSetFillStyle (FILL_SOLID);
  mfl = bess_tf_get_mfl (tf);
  loop_through_list (mfl, mf, MFP) {
    supplied = bess_mf_get_loc (mf, pts);
    if (supplied == 6) {
      go = CreateTri6 (pts);
      if (o == BESS_ORIENT_REVERSED) {
        SWAPLOC (pts[1], pts[2]);
        SWAPLOC (pts[3], pts[5]);
      }
    } else {
      if (o == BESS_ORIENT_REVERSED) {
        SWAPLOC (pts[1], pts[2]);
      }
      go = CreateTriangle3D (pts);
    }
    EGWithMaskChangeAttributes (COLOR_MASK | SHRINK_MASK | FILL_MASK, go);
    EMNoDrawAddGraphicsToModel (model, go);
  }
  free_list (mfl, NOT_NODES);
}



static int
tf_pass_crit (char *crit, TFP tf, BESSP topo)
{
  static tokens_parser P = NULL;

  if (!P) P = tokens_new_parser ();

  if (!crit) return TRUE;
  
  if (!tokens_parse_line (P, crit)) return TRUE;

  if (tokens_have_keyword (P, 1, "MANIFOLD")) {
    return bess_tf_is_manifold (tf);
  }
  if (tokens_have_keyword (P, 1, "NON-MANIFOLD")) {
    return bess_tf_is_non_manifold (tf);
  }
  if (tokens_have_keyword (P, 1, "SHEET")) {
    return bess_tf_is_sheet (tf);
  }
  return TRUE;
}


int
show_tf (BESSP topo, dict_dictionary to_include, dict_dictionary to_exclude, char *crit, EModel *model)
{
  TFP tf;

  tf = NULL;
  while ((tf = bess_next_tf (topo, tf))) {
    if (to_include || to_exclude) {
      if (include_entity (to_include, to_exclude, bess_tf_get_id (tf)))
        tf_grep (tf, model, BESS_ORIENT_SAME);
    } else {
      if (tf_pass_crit (crit, tf, topo))
        tf_grep (tf, model, BESS_ORIENT_SAME);
    }
  }

  return 1;
}

static void
get_t_bbox (MFP mf, G3D_box_t *box)
{
  LOC loc[6];
  bess_mf_get_loc (mf, loc);
  G3D_INIT_BOX ((*box), loc[0]);
  G3D_UPDT_BOX ((*box), loc[1]);
  G3D_UPDT_BOX ((*box), loc[2]);
}

static void
tf_orient_grep (TFP tf, TOE o, EModel *model)
{
  LIST_OF (MFP) mfl;
  EPixel p = color_from_id (bess_tf_get_id (tf));
  WCRec vpts[2], tpnts[6];
  MFP mf;
  GraphicObj *go;
  G3D_box_t bbox;
  double dim;

  EASValsSetColor (p);
  EASValsSetFillStyle (FILL_SOLID);
  mfl = bess_tf_get_mfl (tf);
  loop_through_list (mfl, mf, MFP) {
    bess_mf_get_loc (mf, tpnts);
    vpts[0].x = (tpnts[0].x + tpnts[1].x + tpnts[2].x) / 3;
    vpts[0].y = (tpnts[0].y + tpnts[1].y + tpnts[2].y) / 3;
    vpts[0].z = (tpnts[0].z + tpnts[1].z + tpnts[2].z) / 3;
    bess_mf_get_normal (mf, 0.33333, 0.33333, &vpts[1]);
    get_t_bbox (mf, &bbox);
    dim = (G3D_BOX_RANGE (bbox, x) + G3D_BOX_RANGE (bbox, y) + G3D_BOX_RANGE (bbox, z)) / 3 / 2;
    VECMAC_SCALAR (vpts[1], *=, o * dim);
    VECMAC_2_VEC (vpts[0], +=, vpts[1]);
    go = CreateVector3D (vpts);
    EGWithMaskChangeAttributes (COLOR_MASK, go);
    EMNoDrawAddGraphicsToModel (model, go);
  }
  free_list (mfl, NOT_NODES);
}

int
show_tf_orient (BESSP topo, dict_dictionary to_include, dict_dictionary to_exclude, char *crit, EModel *model)
{
  TFP tf;

  tf = NULL;
  while ((tf = bess_next_tf (topo, tf))) {
    if (to_include || to_exclude) {
      if (include_entity (to_include, to_exclude, bess_tf_get_id (tf))) 
        tf_orient_grep (tf, BESS_ORIENT_SAME, model);
    } else {
      if (tf_pass_crit (crit, tf, topo))
        tf_orient_grep (tf, BESS_ORIENT_SAME, model);
    }
  }
  
  return 1;
}


static void
ts_grep (TSP ts, EModel *model)
{
  TFUP tfu;
  TFP tf;
  tfu = NULL;
  while ((tfu = bess_ts_get_next_tfu (ts, tfu))) {
    tf = bess_tfu_get_tf (tfu);
    tf_grep (tf, model, bess_tfu_get_o (tfu));
  }
}

int
show_ts (BESSP topo, dict_dictionary to_include, dict_dictionary to_exclude, char *crit, EModel *model)
{
  TSP ts;

  ts = NULL;
  while ((ts = bess_next_ts (topo, ts))) {
    if (to_include || to_exclude) {
      if (include_entity (to_include, to_exclude, bess_ts_get_id (ts)))
        ts_grep (ts, model);
    } else {
      ts_grep (ts, model);
    }
  }

  return 1;
}


static void
ts_orient_grep (TSP ts, EModel *model)
{
  TFUP tfu;
  TFP tf;
  tfu = NULL;
  while ((tfu = bess_ts_get_next_tfu (ts, tfu))) {
    tf = bess_tfu_get_tf (tfu);
    tf_orient_grep (tf, bess_tfu_get_o (tfu), model);
  }
}

int
show_ts_orient (BESSP topo, dict_dictionary to_include, dict_dictionary to_exclude, char *crit, EModel *model)
{
  TSP ts;

  ESIHandleCmd ("color green");

  ts = NULL;
  while ((ts = bess_next_ts (topo, ts))) {
    if (to_include || to_exclude) {
      if (include_entity (to_include, to_exclude, bess_ts_get_id (ts))) 
        ts_orient_grep (ts, model);
    } else {
      ts_orient_grep (ts, model);
    }
  }

  return 1;
}




static void
tv_num_grep (TVP tv, EModel *model)
{
  WCRec pts[1];
  GraphicObj *go;
  char buf[64];
  
  bess_mv_get_loc (bess_tv_get_mv (tv), &pts[0]);
  sprintf (buf, " %d", bess_tv_get_id (tv));
  go = CreateAnnText3D (pts, buf);
  EGWithMaskChangeAttributes (COLOR_MASK | FONT_MASK, go);
  EMNoDrawAddGraphicsToModel (model, go);
}

int
show_tv_nums (BESSP topo, dict_dictionary to_include, dict_dictionary to_exclude, EModel *model)
{
  TVP tv;

  ESIHandleCmd ("color grey");
  ESIHandleCmd ("font \"9x15\"");

  tv = NULL;
  while ((tv = bess_next_tv (topo, tv))) {
    if (to_include || to_exclude) {
      if (include_entity (to_include, to_exclude, bess_tv_get_id (tv))) 
        tv_num_grep (tv, model);
    } else {
      tv_num_grep (tv, model);
    }
  }

  return TRUE;
}

static MVP
common_mv (MEP me, MEP ame)
{
  MVP mv[2], amv[2];
  bess_me_get_mv (me, mv);
  bess_me_get_mv (ame, amv);
  if      (mv[0] == amv[0] || mv[0] == amv[1]) return mv[0];
  else if (mv[1] == amv[0] || mv[1] == amv[1]) return mv[1];
  else                                         return NULL;
}

static void
e_mid_point (TEP te, WCRec pts[1])
{
  LIST_OF (MEP) mel;
  int n, n2;

  mel = bess_te_get_mel (te);
  n = list_length (mel); n2 = n / 2;

  if (n2 * 2 == n) {
    MEP me, nme;
    MVP mv;
    me = (MEP) get_list_pos (mel, n2);
    nme = (MEP) get_list_pos (mel, n2+1);
    mv = common_mv (me, nme);
    bess_mv_get_loc (mv, pts);
  } else {
    MEP me;
    MVP mv[2];
    LOC p0, p1;
    me = (MEP) get_list_pos (mel, n2+1);
    bess_me_get_mv (me, mv);
    bess_mv_get_loc (mv[0], &p0);
    bess_mv_get_loc (mv[1], &p1);
    VECMAC_3_VEC (pts[0], =, 0.5 * p0, +, 0.5 * p1);
  }
  free_list (mel, NOT_NODES);
}

static void
te_num_grep (TEP te, EModel *model)
{
  EPixel p = color_from_id (bess_te_get_id (te));
  WCRec pts[1];
  GraphicObj *go;
  char buf[64];
  
  EASValsSetColor (p);
  e_mid_point (te, pts);
  sprintf (buf, " %d", bess_te_get_id (te));
  go = CreateAnnText3D (pts, buf);
  EGWithMaskChangeAttributes (COLOR_MASK | FONT_MASK, go);
  EMNoDrawAddGraphicsToModel (model, go);
}

int
show_te_nums (BESSP topo, dict_dictionary to_include, dict_dictionary to_exclude, EModel *model)
{
  TEP te;

  ESIHandleCmd ("color grey");
  ESIHandleCmd ("font \"9x15\"");

  te = NULL;
  while ((te = bess_next_te (topo, te))) {
    if (to_include || to_exclude) {
      if (include_entity (to_include, to_exclude, bess_te_get_id (te))) 
        te_num_grep (te, model);
    } else {
      te_num_grep (te, model);
    }
  }

  return TRUE;
}


static void
mf_grep (MFP mf, EModel *model)
{
  WCRec pts[6];
  GraphicObj *go;
  int supplied;

  supplied = bess_mf_get_loc (mf, pts);
  if (supplied == 6)
    go = CreateTri6 (pts);
  else
    go = CreateTriangle3D (pts);
  EGWithMaskChangeAttributes (COLOR_MASK |
                              FILL_MASK |
                              EDGE_COLOR_MASK |
                              EDGE_FLAG_MASK |
                              LAYER_MASK, go);
  EMNoDrawAddGraphicsToModel (model, go);
}

int
show_mf (BESSP bess, EModel *model)
{
  MFP mf;
  ESIHandleCmd ("fill solid");
  ESIHandleCmd ("color red");
  ESIHandleCmd ("edgecolor white");
  ESIHandleCmd ("edgeflag on");
  ESIHandleCmd ("layer active 0");
  mf = NULL;
  while ((mf = bess_next_mf (bess, mf))) {
    mf_grep (mf, model);
  }
  return TRUE;
}

static int
is_adj_to_interfering (MFP mf, mi_checker_p mic)
{
  MFP nmf;
  MEP me[3];
  int i;

  bess_mf_get_me (mf, me);
  FOR (i, 3) {
    nmf = NULL;
    while ((nmf = bess_me_get_next_mf (me[i], nmf))) {
      if (nmf != mf && mi_does_interfere (mic, nmf)) {
        return TRUE;
      }
    }
  }

  return FALSE;
}


#define RESET { ESIHandleCmd ("layer active 0");      \
                ESIHandleCmd ("color white");         \
              }
#define INTRF { ESIHandleCmd ("layer active 13");     \
                ESIHandleCmd ("color red");           \
              }
#define INTRE { ESIHandleCmd ("layer active 14");     \
                ESIHandleCmd ("color red");           \
              }
#define ADJ   { ESIHandleCmd ("layer active 12");     \
                ESIHandleCmd ("color green");         \
              }

int
show_mf_interference (BESSP bess, EModel *model, mi_checker_p mic)
{
  MFP mf;

  ESIHandleCmd ("fill solid");
  ESIHandleCmd ("edgecolor black");
  ESIHandleCmd ("edgeflag on");
  RESET;

  mf = NULL;
  while ((mf = bess_next_mf (bess, mf))) {
    if (mic) {
      if (mi_does_interfere (mic, mf)) {
        INTRF;
        mf_grep (mf, model);
        RESET;
      } else if (is_adj_to_interfering (mf, mic)) {
        ADJ;
        mf_grep (mf, model);
        RESET;
      } else {
        mf_grep (mf, model);
      }
    } else {
      mf_grep (mf, model);
    }
  }
  return TRUE;
}


static void
me_grep (MEP me, EModel *model)
{
  WCRec pts[3], pnts[2];
  GraphicObj *go;
  int supplied;
  
  supplied = bess_me_get_loc (me, pts);
  if (supplied == 3) {
    pnts[0] = pts[0];
    pnts[1] = pts[2];
    go = CreateLine3D (pnts);
    EGWithMaskChangeAttributes (COLOR_MASK, go);
    EMNoDrawAddGraphicsToModel (model, go);
    pnts[0] = pts[2];
    pnts[1] = pts[1];
    go = CreateLine3D (pnts);
    EGWithMaskChangeAttributes (COLOR_MASK, go);
    EMNoDrawAddGraphicsToModel (model, go);
  } else {
    go = CreateLine3D (pts);
    EGWithMaskChangeAttributes (COLOR_MASK, go);
    EMNoDrawAddGraphicsToModel (model, go);
  }
}

int
show_me (BESSP bess, EModel *model)
{
  MEP me;
  ESIHandleCmd ("color yellow");
  me = NULL;
  while ((me = bess_next_me (bess, me))) {
    me_grep (me, model);
  }
  return TRUE;
}

static void
mv_grep (MVP mv, EModel *model)
{
  WCRec pts[1];
  GraphicObj *go;
  
  bess_mv_get_loc (mv, &pts[0]);
  go = CreateMarker3D (pts);
  EGWithMaskChangeAttributes (COLOR_MASK | MSIZE_MASK | MTYPE_MASK, go);
  EMNoDrawAddGraphicsToModel (model, go);
}

int
show_mv (BESSP bess, EModel *model)
{
  MVP mv;
  ESIHandleCmd ("color yellow");
  ESIHandleCmd ("msize 3");
  ESIHandleCmd ("mtype 3");
  mv = NULL;
  while ((mv = bess_next_mv (bess, mv))) {
    mv_grep (mv, model);
  }
  return TRUE;
}

static void
mf_num_grep (MFP mf, EModel *model)
{
  WCRec pts[6];
  WCRec pt[1];
  char buf[64];
  GraphicObj *go;

  bess_mf_get_loc (mf, pts);
  VECMAC_2_VEC (pt[0],  =, pts[0]);
  VECMAC_2_VEC (pt[0], +=, pts[1]);
  VECMAC_2_VEC (pt[0], +=, pts[2]);
  VECMAC_SCALAR (pt[0], *=, 0.333333);
  sprintf (buf, " %d", bess_mf_get_id (mf));
  go = CreateAnnText3D (pt, buf);
  EGWithMaskChangeAttributes (COLOR_MASK | FONT_MASK, go);  
  EMNoDrawAddGraphicsToModel (model, go);
}

int
show_mf_nums (BESSP bess, EModel *model)
{
  MFP mf;
  ESIHandleCmd ("color cyan");
  ESIHandleCmd ("font \"9x15\"");
  mf = NULL;
  while ((mf = bess_next_mf (bess, mf))) {
    mf_num_grep (mf, model);
  }
  return TRUE;
}

static void
me_num_grep (MEP me, EModel *model)
{
  WCRec pts[3];
  WCRec pt[1];
  GraphicObj *go;
  char buf[64];
 
  bess_me_get_loc (me, pts);
  VECMAC_3_VEC (pt[0], =, 0.5 * pts[0], +, 0.5 * pts[1]);
  sprintf (buf, " %d", bess_me_get_id (me));
  go = CreateAnnText3D (pt, buf);
  EGWithMaskChangeAttributes (COLOR_MASK | FONT_MASK, go);
  EMNoDrawAddGraphicsToModel (model, go);
}

int
show_me_nums (BESSP bess, EModel *model)
{
  MEP me;
  ESIHandleCmd ("color green");
  ESIHandleCmd ("font \"9x15\"");
  me = NULL;
  while ((me = bess_next_me (bess, me))) {
    me_num_grep (me, model);
  }
  return TRUE;
}

static void
mv_num_grep (MVP mv, EModel *model)
{
  WCRec pts[1];
  GraphicObj *go;
  char buf[64];
  
  bess_mv_get_loc (mv, &pts[0]);
  sprintf (buf, " %d", bess_mv_get_id (mv));
  go = CreateAnnText3D (pts, buf);
  EGWithMaskChangeAttributes (COLOR_MASK | FONT_MASK, go);
  EMNoDrawAddGraphicsToModel (model, go);
}

int
show_mv_nums (BESSP bess, EModel *model)
{
  MVP mv;
  ESIHandleCmd ("color hotpink");
  ESIHandleCmd ("font \"9x15\"");
  mv = NULL;
  while ((mv = bess_next_mv (bess, mv))) {
    mv_num_grep (mv, model);
  }
  return TRUE;
}


static void
tl_teu_grep (TEUP teu, EModel *model)
{
  LIST_OF (MEP) mel;
  TEP te = bess_teu_get_te (teu);
  TOE to = bess_teu_get_o (teu);
  EPixel p = color_from_id (bess_te_get_id (te));
  WCRec pts[2];
  MEP me;
  GraphicObj *go;
  MVP smv, omv, emv[2];
  TVP tv[2];

  bess_te_get_tv (te, tv);
  smv = (to == BESS_ORIENT_SAME ? bess_tv_get_mv (tv[0]) : bess_tv_get_mv (tv[1]));
  EASValsSetColor (p);
  mel = bess_te_get_mel (te);
  if (to == BESS_ORIENT_SAME) {
    me = NULL;
    while ((me = (MEP) get_list_next (mel, me))) {
      bess_mv_get_loc (smv, &pts[1]);
      bess_me_get_mv (me, emv);
      omv = (emv[0] == smv ? emv[1] : emv[0]);
      bess_mv_get_loc (omv, &pts[0]);
      VECMAC_3_VEC (pts[1], =, pts[0], -, pts[1]);
      go = CreateVector3D (pts);
      EGWithMaskChangeAttributes (COLOR_MASK, go);
      EMNoDrawAddGraphicsToModel (model, go);
      smv = omv;
    }
  } else {
    me = NULL;
    while ((me = (MEP) get_list_prev (mel, me))) {
      bess_mv_get_loc (smv, &pts[1]);
      bess_me_get_mv (me, emv);
      omv = (emv[0] == smv ? emv[1] : emv[0]);
      bess_mv_get_loc (omv, &pts[0]);
      VECMAC_3_VEC (pts[1], =, pts[0], -, pts[1]);
      go = CreateVector3D (pts);
      EGWithMaskChangeAttributes (COLOR_MASK, go);
      EMNoDrawAddGraphicsToModel (model, go);
      smv = omv;
    }
  }
  free_list (mel, NOT_NODES);
}

static void
tl_grep (TLP tl, EModel *model)
{
  TEUP teu;
  teu = NULL;
  while ((teu = bess_tl_get_next_teu (tl, teu))) {
    tl_teu_grep (teu, model);
  }
}

int
show_tl (BESSP topo, dict_dictionary to_include, dict_dictionary to_exclude, char *crit, EModel *model)
{
  TLP tl;

  tl = NULL;
  while ((tl = bess_next_tl (topo, tl))) {
    if (to_include || to_exclude) {
      if (include_entity (to_include, to_exclude, bess_tl_get_id (tl))) 
        tl_grep (tl, model);
    } else {
      tl_grep (tl, model);
    }
  }

  return 1;
}


static MVP
other_mv (MEP e, MVP av)
{
  MVP v[3];
  bess_me_get_mv(e, v);
  if      (v[0] == av)
    return v[1];
  else if (v[1] == av)
    return v[0];
  else
    return NULL;
}

static void
remesh2(MEP me1, MEP me2, double lambda1, double lambda2, EModel *model)
{
  LOC cvl, ov1l, ov2l, s1, s2, vl[3];
  MVP cv = common_mv(me1, me2);
  MVP ov1 = other_mv(me1, cv);
  MVP ov2 = other_mv(me2, cv);
  if ((cv == 0) || (ov1 == 0) || (ov2 == 0))
    ERptErrMessage (1, 2, "Wrong.", ERROR_GRADE);
  bess_mv_get_loc(cv, &cvl);
  bess_mv_get_loc(ov1, &ov1l);
  bess_mv_get_loc(ov2, &ov2l);
  bess_me_get_loc(me1, vl);
  VECMAC_LERP(s1, lambda1, vl[0], vl[1]); 
  bess_me_get_loc(me2, vl);
  VECMAC_LERP(s2, lambda2, vl[0], vl[1]);
  {
    WCRec pts[6];
    GraphicObj *go;
    INTRF;
    pts[0]=cvl; pts[1]=s1; pts[2]=s2;
    go = CreateTriangle3D (pts);
    EGWithMaskChangeAttributes (COLOR_MASK |
                                FILL_MASK |
                                EDGE_COLOR_MASK |
                                EDGE_FLAG_MASK |
                                LAYER_MASK, go);
    EMNoDrawAddGraphicsToModel (model, go);
    pts[0]=s1; pts[1]=ov1l; pts[2]=ov2l; pts[3]=s2;
    go = CreateQuad3D (pts);
    EGWithMaskChangeAttributes (COLOR_MASK |
                                FILL_MASK |
                                EDGE_COLOR_MASK |
                                EDGE_FLAG_MASK |
                                LAYER_MASK, go);
    EMNoDrawAddGraphicsToModel (model, go);
    INTRE;
    pts[0]=s1; 
    go = CreateMarker3D (pts);
    EGWithMaskChangeAttributes (COLOR_MASK | MSIZE_MASK | MTYPE_MASK, go);
    EMNoDrawAddGraphicsToModel (model, go);
    pts[0]=s2; 
    go = CreateMarker3D (pts);
    EGWithMaskChangeAttributes (COLOR_MASK | MSIZE_MASK | MTYPE_MASK, go);
    EMNoDrawAddGraphicsToModel (model, go);
    pts[0]=s1; pts[1]=s2; 
    go = CreateLine3D (pts);
    EGWithMaskChangeAttributes (COLOR_MASK | WIDTH_MASK | LAYER_MASK, go);
    EMNoDrawAddGraphicsToModel (model, go);
  }
}

static void
remesh_mf (EModel *model, mi_checker_p mic, MFP mf)
{
  MEP mes[3];
  double lambda[3]={0,0,0};
  int intersected[3]={FALSE,FALSE,FALSE};
  int nintersected=0;
  int i;

  bess_mf_get_me (mf, mes);
  FOR (i, 3) {
    intersected[i]=mi_does_intersect (mic, mes[i], &lambda[i]);
    if (intersected[i]) nintersected++;
  }

  switch (nintersected) {
  case 2:
    if      (intersected[0] && intersected[1])
      remesh2(mes[0], mes[1], lambda[0], lambda[1], model);
    else if (intersected[0] && intersected[2]) 
      remesh2(mes[0], mes[2], lambda[0], lambda[2], model);
    else if (intersected[1] && intersected[2]) 
      remesh2(mes[1], mes[2], lambda[1], lambda[2], model);
    else
      /* void */;
    break;
  default:                      /* can't handle anything else */
    break;
  }
}

static void
show_remesh (EModel *model, mi_checker_p mic, MEP me, dict_dictionary remeshed)
{
  MFP nmf = 0;
  
  nmf = NULL;
  while ((nmf = bess_me_get_next_mf (me, nmf))) {
    if (mi_does_interfere (mic, nmf)) {
      if ((MFP) dict_find_obj_with_handle_name (remeshed, nmf) == 0) {
        remesh_mf (model, mic, nmf);
        dict_add_obj_handle_name (remeshed, nmf, nmf);
      }
    }
  }
}

int
show_mf_remeshed (BESSP bess, EModel *model, mi_checker_p mic)
{
  MFP mf;
  MEP me;
  dict_dictionary remeshed = dict_create_dictionary();
  
  if (mic == 0) return FALSE;

  if (remeshed == 0) return FALSE;
  
  ESIHandleCmd ("fill solid");
  ESIHandleCmd ("edgecolor black");
  ESIHandleCmd ("edgeflag on");
  ESIHandleCmd ("msize 6");
  ESIHandleCmd ("mtype 1");
  RESET;

  mf = NULL;
  while ((mf = bess_next_mf (bess, mf))) {
    if (mi_does_interfere (mic, mf)) {
      /* ignore */
    } else if (is_adj_to_interfering (mf, mic)) {
      ADJ;
      mf_grep (mf, model);
      RESET;
    } else {
      mf_grep (mf, model);
    }
  }

  INTRF;
  me = NULL;
  while ((me = bess_next_me (bess, me))) {
    double  lambda=0;
    if (mi_does_intersect (mic, me, &lambda)) {
      show_remesh (model, mic, me, remeshed);
    }
  }
  RESET;
  
  dict_destroy_dictionary (remeshed);
  
  return TRUE;
}
