
#include "Esimple.h"
#include "bess.h"
#include "bessstubs.h"
#include "tetmesh.h"
#include "bess.pt"
#include "t3din.h"


static int nnodes, nedges, ntriangles, ntetras;
static int etype, edegree;
static tokens_parser PARSER = NULL;

#define GETLINE()                                                               \
  {                                                                             \
    if (!tokens_next_line (PARSER)) {                                           \
      fprintf (stderr, "Error in %s (l.%d)\n", __FILE__, __LINE__);             \
      fprintf (stderr, "Input: \"%s\", failed to get next line at l.%d\n",      \
               file, tokens_curr_line_num (PARSER));                            \
      return FALSE;                                                             \
    }                                                                           \
  }

#define EXPECT_TOT(tot)                                                 \
  {                                                                     \
    if (tokens_total_of_tokens (PARSER) < tot) {                        \
      fprintf (stderr, "Error in %s (l.%d)\n", __FILE__, __LINE__);     \
      fprintf (stderr, "Expected %d tokens, got \"%s\"\n",              \
               tot, tokens_curr_line_contents (PARSER));                \
      return FALSE;                                                     \
    }                                                                   \
  }


static void
drawtet (int id, MESHP st, EModel *model, int layer)
{
  WCRec pts[10];
  MVP mv[10];
  int i;
  GraphicObj *go;
  MTP mt;
  int supplied;

  mt = tet_find_mt_with_id (st, id);
  
  supplied = tet_mt_get_mv (mt, mv);
  FOR (i, supplied) {
    tet_mv_get_loc (mv[i], &pts[i]);
  }
  if (supplied == 4)
    go = CreateTetra (pts);
  else
    go = CreateTet10 (pts);
  EASValsSetLayer (layer);
  EASValsSetColor (color_from_id (layer));
  EGWithMaskChangeAttributes (  COLOR_MASK
                              | EDGE_COLOR_MASK
                              | EDGE_FLAG_MASK
                              | FILL_MASK
                              | LAYER_MASK, go);
  EMNoDrawAddGraphicsToModel (model, go);
}

int
t3din_read (char *file, MESHP st, EModel *model)
{
  int i, j, id;
  G3D_loc_t loc;
  int mvids[10];
  int regid, regidat;
  
  if (!PARSER) PARSER = tokens_new_parser ();

  if (!tokens_open_file (PARSER, file))
    return FALSE;

  /* nnodes nedges ntriangles ntetras */
  GETLINE ();
  EXPECT_TOT (4);
  nnodes     = tokens_token_as_int (PARSER, 1);
  nedges     = tokens_token_as_int (PARSER, 2);
  ntriangles = tokens_token_as_int (PARSER, 3);
  ntetras    = tokens_token_as_int (PARSER, 4);

  /* etype edegree */
  GETLINE ();
  EXPECT_TOT (2);
  etype   = tokens_token_as_int (PARSER, 1);
  edegree = tokens_token_as_int (PARSER, 2);

  /* nodes */
  for (i = 0; i < nnodes; i++) {
    GETLINE ();
    EXPECT_TOT (4);
    id    = tokens_token_as_int (PARSER, 1);
    loc.x = tokens_token_as_double (PARSER, 2);
    loc.y = tokens_token_as_double (PARSER, 3);
    loc.z = tokens_token_as_double (PARSER, 4);
    if (!tet_add_node (st, id, &loc))
      return FALSE;
  }

  /* skip edges */
  for (i = 0; i < nedges; i++) {
    GETLINE ();
  }
  
  /* skip triangles */
  for (i = 0; i < ntriangles; i++) {
    GETLINE ();
  }
  
  /* tetras */
  ESIHandleCmd ("color yellow");
  ESIHandleCmd ("edgecolor black");
  ESIHandleCmd ("edgeflag on");
  ESIHandleCmd ("fill solid");
  regidat = (edegree == 1 ? 7 : 13);
  for (i = 0; i < ntetras; i++) {
    GETLINE ();
    EXPECT_TOT (regidat);
    id       = tokens_token_as_int (PARSER, 1);
    mvids[0] = tokens_token_as_int (PARSER, 2);
    mvids[1] = tokens_token_as_int (PARSER, 3);
    mvids[2] = tokens_token_as_int (PARSER, 4);
    mvids[3] = tokens_token_as_int (PARSER, 5);
    for (j = 4; j < 10; j++) mvids[j] = -1;
    if (edegree == 2) {
      mvids[4] = tokens_token_as_int (PARSER,  6);
      mvids[5] = tokens_token_as_int (PARSER,  7);
      mvids[6] = tokens_token_as_int (PARSER,  8);
      mvids[7] = tokens_token_as_int (PARSER,  9);
      mvids[8] = tokens_token_as_int (PARSER, 10);
      mvids[9] = tokens_token_as_int (PARSER, 11);
    }
    regid    = tokens_token_as_int (PARSER, regidat);
    if (!tet_add_tet (st, id, mvids, regid))
      return FALSE;
    drawtet (id, st, model, regid);
  }
  
  tokens_close_curr_file (PARSER);

  return TRUE;
}
