
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <limits.h>
#include <sys/stat.h>
#include <unistd.h>
#include <float.h>

#if HAVE_STRING_H
#      include <string.h>
#else
#  if HAVE_STRINGS_H
#         include <strings.h>
#  endif
#endif

#include "getoptP.h"

#include "Esimple.h"
#include "tetmesh.h"


#ifndef FOR
#  define FOR(i_, n_) for (i_ = 0; i_ < n_; i_++)
#endif

static void 
usage (int argc, char **argv)
{
  char *option, *usage;

  fprintf (stderr, "usage: %s t3d_file_name\n", argv[0]);
  
  exit (0);
}


#define MESH     tet_triangulation_t
#define MESHP    tet_triangulation_p
#define MT       tet_tet_t
#define MF       tet_face_t
#define ME       tet_edge_t
#define MV       tet_vertex_t
#define MTP      tet_tet_p
#define MFP      tet_face_p
#define MEP      tet_edge_p
#define MVP      tet_vertex_p


static int nnodes, nedges, ntriangles, ntetras;
static int etype, edegree;
static tokens_parser PARSER = NULL;

#define GETLINE()                                                               \
  {                                                                             \
    if (!tokens_next_line (PARSER)) {                                           \
      fprintf (stderr, "Error in %s (l.%d)\n", __FILE__, __LINE__);             \
      fprintf (stderr, "Input: \"%s\", failed to get next line at l.%d\n",      \
               file, tokens_curr_line_num (PARSER));                            \
      return FALSE;                                                             \
    }                                                                           \
  }

#define EXPECT_TOT(tot)                                                 \
  {                                                                     \
    if (tokens_total_of_tokens (PARSER) < tot) {                        \
      fprintf (stderr, "Error in %s (l.%d)\n", __FILE__, __LINE__);     \
      fprintf (stderr, "Expected %d tokens, got \"%s\"\n",              \
               tot, tokens_curr_line_contents (PARSER));                \
      return FALSE;                                                     \
    }                                                                   \
  }


static void
drawtet (int id, MESHP st, EModel *model, int layer)
{
  WCRec pts[10];
  MVP mv[10];
  int i;
  GraphicObj *go;
  MTP mt;
  int supplied;

  mt = tet_find_mt_with_id (st, id);
  
  supplied = tet_mt_get_mv (mt, mv);
  FOR (i, supplied) {
    tet_mv_get_loc (mv[i], &pts[i]);
  }
  if (supplied == 4)
    go = CreateTetra (pts);
  else
    go = CreateTet10 (pts);
  EASValsSetLayer (layer);
  EASValsSetColor (color_from_id (layer));
  EGWithMaskChangeAttributes (  COLOR_MASK
                              | EDGE_COLOR_MASK
                              | EDGE_FLAG_MASK
                              | FILL_MASK
                              | LAYER_MASK, go);
  EMNoDrawAddGraphicsToModel (model, go);
}

static int
t3din_read (char *file, MESHP st, EModel *model)
{
  int i, j, id;
  G3D_loc_t loc;
  int mvids[10];
  int regid, regidat;
  
  if (!PARSER) PARSER = tokens_new_parser ();

  if (!tokens_open_file (PARSER, file))
    return FALSE;

  /* nnodes nedges ntriangles ntetras */
  GETLINE ();
  EXPECT_TOT (4);
  nnodes     = tokens_token_as_int (PARSER, 1);
  nedges     = tokens_token_as_int (PARSER, 2);
  ntriangles = tokens_token_as_int (PARSER, 3);
  ntetras    = tokens_token_as_int (PARSER, 4);

  /* etype edegree */
  GETLINE ();
  EXPECT_TOT (2);
  etype   = tokens_token_as_int (PARSER, 1);
  edegree = tokens_token_as_int (PARSER, 2);

  /* nodes */
  for (i = 0; i < nnodes; i++) {
    GETLINE ();
    EXPECT_TOT (4);
    id    = tokens_token_as_int (PARSER, 1);
    loc.x = tokens_token_as_double (PARSER, 2);
    loc.y = tokens_token_as_double (PARSER, 3);
    loc.z = tokens_token_as_double (PARSER, 4);
    if (!tet_add_node (st, id, &loc))
      return FALSE;
  }

  /* skip edges */
  for (i = 0; i < nedges; i++) {
    GETLINE ();
  }
  
  /* skip triangles */
  for (i = 0; i < ntriangles; i++) {
    GETLINE ();
  }
  
  /* tetras */
  ESIHandleCmd ("color yellow");
  ESIHandleCmd ("edgecolor black");
  ESIHandleCmd ("edgeflag on");
  ESIHandleCmd ("fill solid");
  regidat = (edegree == 1 ? 7 : 13);
  for (i = 0; i < ntetras; i++) {
    GETLINE ();
    EXPECT_TOT (regidat);
    id       = tokens_token_as_int (PARSER, 1);
    mvids[0] = tokens_token_as_int (PARSER, 2);
    mvids[1] = tokens_token_as_int (PARSER, 3);
    mvids[2] = tokens_token_as_int (PARSER, 4);
    mvids[3] = tokens_token_as_int (PARSER, 5);
    for (j = 4; j < 10; j++) mvids[j] = -1;
    if (edegree == 2) {
      mvids[4] = tokens_token_as_int (PARSER,  6);
      mvids[5] = tokens_token_as_int (PARSER,  7);
      mvids[6] = tokens_token_as_int (PARSER,  8);
      mvids[7] = tokens_token_as_int (PARSER,  9);
      mvids[8] = tokens_token_as_int (PARSER, 10);
      mvids[9] = tokens_token_as_int (PARSER, 11);
    }
    regid    = tokens_token_as_int (PARSER, regidat);
    if (!tet_add_tet (st, id, mvids, regid))
      return FALSE;
    drawtet (id, st, model, regid);
  }
  
  tokens_close_curr_file (PARSER);

  return TRUE;
}

void
ESICustomize (Widget parent_pane)
{
 /* VOID */
}

int
main (int argc, char **argv)
{
  static tet_triangulation_p mesh = NULL;
  unsigned long m = ESI_GRAPHIC_EDITOR_MASK;
  if (argc != 2) {
    usage (argc, argv);
  }
  mesh = tet_new_triangulation (NULL);
  if (!mesh) {
    return 1;
  }
  ESIBuildInterface (m, argc, argv);
  if (!t3din_read (argv[1], mesh, ESIModel ())) {
    ERptStatusMessage(1, 1, "reading failed");
    return 0;
  }
  ESIPopupAndRun ();

  return 0;
}

