 
#include "bess.h"
#include "bessstubs.h"
#include "bess.pt"
#include "getoptP.h"
#include "tetacts.h"
#include "tetmesh.h"
#include "mic.h"
#include "show.h"
#include "t3din.h"
#include "tetin.h"
#include "bess2adlib.h"
#include "bess2vdt.h"
#include "bess2wasp.h"

static FILE *logfp = NULL;
static char logname[L_tmpnam];
static char logbuf[1024];
static int netscape = FALSE;

static void
update_log (char *s)
{
  if (!logfp) {
    tmpnam (logname);
    logfp = fopen (logname, "w");
  }
  if (logfp)
    fprintf (logfp, "%s\n", s);
}


static void
popup_log (void)
{
  fclose (logfp);
  ESIPopupInfoFromFile (logname);  
}



static void 
usage (int argc, char **argv)
{
  char *option, *usage;

  fprintf (stderr, "usage: %s + options below\n", argv[0]);
  
  if (ckit_opt_first_recognized (&option, &usage)) {
    fprintf (stderr, "%s %s\n",
             (option != NULL ? option : "?"),
             (usage  != NULL ? usage  : "?"));
  }
  while (ckit_opt_next_recognized (&option, &usage)) {
    fprintf (stderr, "%s %s\n",
             (option != NULL ? option : "?"),
             (usage  != NULL ? usage  : "?"));
  }
  exit (0);
}


static char *cmd_file = NULL;
static char *mesh_file = NULL;


static void
parse_cmd_line (int argc, char **argv)
{
  int help = FALSE;
  int report_opts = FALSE;
#define SET_BOOL(name)                                          \
  {                                                             \
    if (report_opts) { fprintf (stderr, "Set: %s\n", #name); }  \
    name = TRUE;                                                \
  }
#define SET_WITH_ARG(name, format)                                        \
  {                                                                       \
    if (report_opts) { fprintf (stderr, "Set: %s = %" #format "\n",       \
                                #name, name); }                           \
  }
  
  GET_OPT_BOOL (argc, argv, "-report_opts", SET_BOOL(report_opts),
                "prints options as they are being set");
  
  GET_OPT_BOOL (argc, argv, "-help", SET_BOOL(help),
                "prints usage for each option recognized");
  
  GET_OPT_ARG (argc, argv, "-c", char, cmd_file,
              SET_WITH_ARG (cmd_file, s),
              "command file");

  GET_OPT_ARG (argc, argv, "-m", char, mesh_file,
              SET_WITH_ARG (mesh_file, s),
              "mesh file in t3d format");

  if (help)
    usage (argc, argv);
}






static void
exec_cmd_file (char *name)
{
  const char *linebuf;
  FILE *stream;
  ckit_iobuf_t IOB = NULL;

  stream = ckit_io_open_stream (name, "r");
  if (stream) {
    IOB    = ckit_io_new_buffer (stream);
    while ((linebuf = ckit_io_scan_line (IOB, FALSE))) {
      ESIHandleCmd ((char *)linebuf);
    }
    ckit_io_delete_buffer (IOB);
    fclose(stream);
  } else {
    fprintf (stderr, "\nCould not open command file \"%s\\n", name);
    exit (1);
  }
}

static void
exec_cmd_line_opts (void)
{
  if (cmd_file) {
    exec_cmd_file (cmd_file);
  } else if (mesh_file) {
    char buf[512];
    sprintf (buf, "run \"%s\"", mesh_file);
    ESIHandleCmd (buf);
    ESIHandleCmd ("NEWFRAME");
  } else {
    ESIHandleCmd ("NEWFRAME");
  }
}


#if 0
static void
test_tri6 (void)
{
  GraphicObj *go;
  WCRec pts[6] = {
    {0,0,0},
    {1,0,0},
    {0,1,0},
    {0.5,0,0},
    {0.5, 0.5,0.2},
    {0,0.5,0.1}
  };
  go = CreateTri6 (pts);
  EGWithMaskChangeAttributes (ALL_ATTRIB_MASK, go);
  EMNoDrawAddGraphicsToModel (ESIModel(), go);
}
#endif



int  
main (int argc, char **argv)
{
  unsigned long m = ESI_GRAPHIC_EDITOR_MASK;
  
  parse_cmd_line (argc, argv);

  ESIBuildInterface (m, argc, argv);
  ESIPopup ();

  ESIHandleCmd ("EDGECOLOR white");
  ESIHandleCmd ("EDGEFLAG off");
  ESIHandleCmd ("COLOR yellow");
  ESIHandleCmd ("MSIZE 10");
  ESIHandleCmd ("MTYPE 1");
  ESIHandleCmd ("FONT \"9x15\"");
  ESIHandleCmd ("SELCRIT inside");

  exec_cmd_line_opts ();

  ESIPopupAndRun ();

  return 1;
}



static tet_triangulation_p mesh = NULL;
static bess_topology_p bess = NULL;








static double crease_angle = 60;
static double corner_angle = 45;

static void
handle_crease_angle (void)
{
  if (TypeInGetTokensTotal () >= 2 && TypeInGetTokenType (2) == NUMBER)
    crease_angle = TypeInGetTokenAsDouble (2);
}

static void
handle_corner_angle (void)
{
  if (TypeInGetTokensTotal () >= 2 && TypeInGetTokenType (2) == NUMBER)
    corner_angle = TypeInGetTokenAsDouble (2);
}



static void
print_topo (void)
{
  char buf[L_tmpnam] = "./tmp";
  FILE *fp;
  
  tmpnam (buf);
  fp = fopen (buf, "w");
  if (fp == NULL) {
    ERptErrMessage (1, 1, "Failed to print the data", ERROR_GRADE);
    return;
  }

  bess_print (bess, fp);

  fclose (fp);
  ESIPopupInfoFromFile (buf);  
}

static char *
str_concat (char *string1, const char *string2)
{
  char *target;
  int len = 0;
  
  if (string1 == NULL) {
    len = strlen (string2) + 1;
    string1 = NEWP_REALLOC_OBJ_ARRAY (NULL, len, char);
    if (string1) {
      sprintf (string1, "%s", string2);
      return string1;
    } else {
      return NULL;
    }
  } else {
    char *t;
    len = strlen (string1) + 1 + strlen (string2) + 1;
    t = NEWP_REALLOC_OBJ_ARRAY (string1, len, char);
    if (t) {
      sprintf (t, "%s %s", string1, string2);
      return t;
    } else {
      return NULL;
    }
  }
}
 

static void
html_topo (void) 
{
  char buf[1024] = "./tmp";
  FILE *fp;
  char cmd[1024];
  char *options = NULL;

  tmpnam (buf);
  fp = fopen (buf, "w");
  if (fp == NULL) {
    ERptErrMessage (1, 1, "Failed to print the data", ERROR_GRADE);
    return;
  }  

  if (TypeInGetTokensTotal () >= 2) {
    int i;
    const char *s;
    for (i = 2; i <= TypeInGetTokensTotal (); i++) {
      s = TypeInGetTokenAsString (i);
      options = str_concat (options, s);
    }
  }

  bess_html (bess, fp, options);

  fclose (fp);
  sprintf (cmd, "netscape -remote 'openURL (file:%s) '", buf);
  if (system (cmd) != 0) {
    system ("netscape&");
    sleep (10);
    system (cmd);
  }
}  


#define TIMED_CODE(code, usedtime)                      \
   {                                                    \
     double realtime, usertime, systime;                \
     ckit_get_times (&realtime, &usertime, &systime);   \
     usedtime = usertime;                               \
     { code };                                          \
     ckit_get_times (&realtime, &usertime, &systime);   \
     usedtime = usertime - usedtime;                    \
   }


static void
handle_run (void)
{
  char *meshfile;
  double t;
  int separate_unconnected_tets = FALSE;
  int make_quadratic = FALSE;
  
  bess_opts_t opts;

  separate_unconnected_tets = TypeInIsKeywordInLine (1, "SEPARATE_UNCONNECTED_TETS");
  make_quadratic = TypeInIsKeywordInLine (1, "MAKE_QUADRATIC");
  if (TypeInGetTokensTotal () < 2) {
    ERptErrMessage (1, 1, "Expecting mesh file name", FATAL_ERROR_GRADE);
    return;
  }
  
  meshfile = TypeInGetTokenAsString (2);
  
  mesh = tet_new_triangulation (NULL);
  if (ckit_io_file_has_suffix (meshfile, ".t3d")) {
    sprintf (logbuf, "T3D mesh file: \"%s\"", meshfile);
    update_log (logbuf);
    ESIHandleCmd ("model");
    TIMED_CODE ( {
      if (!t3din_read (meshfile, mesh, ESIModel ())) {
        ERptStatusMessage(1, 1, "reading failed");
        return;
      }
    }, t);
    sprintf (logbuf, "Tet mesh reading took %g seconds", t);
    update_log (logbuf);
  } else {
    sprintf (logbuf, "Elixir mesh file: \"%s\"", meshfile);
    update_log (logbuf);
    ESIHandleCmd ("model");
    TIMED_CODE ( {
      if (!tetin_read (meshfile, mesh, ESIModel ())) {
        ERptStatusMessage(1, 1, "reading failed");
        return;
      }
    }, t);
    sprintf (logbuf, "Tet mesh reading took %g seconds", t);
    update_log (logbuf);
  }

  sprintf (logbuf, "Tet mesh: %d nodes, %d %s tetrahedra,",
           tet_total_of_nodes (mesh), 
           tet_total_of_tets (mesh),
           (tet_geom_degree (mesh) == 2
            ? "quadratic"
            : "linear"));
  update_log (logbuf);
  
  TIMED_CODE ({ tet_prepare (mesh, separate_unconnected_tets, make_quadratic); }, t);
  sprintf (logbuf, "Tet mesh preparation took %g seconds", t);
  update_log (logbuf);

  sprintf (logbuf, "Surface mesh: %d triangles, %d edges",
           tet_total_of_surf_triangles (mesh),
           tet_total_of_surf_edges (mesh));
  update_log (logbuf);
  
  TIMED_CODE ({
    bess = bess_new ();
    
    opts.option = BESS_CREASE_ANGLE;
    opts.value.d = crease_angle;
    bess_config (bess, &opts);
    
    opts.option = BESS_CORNER_ANGLE;
    opts.value.d = corner_angle;
    bess_config (bess, &opts);
    
    bess_build (bess, mesh);
  }, t);
  sprintf (logbuf, "BRep was built in %g seconds", t);
  update_log (logbuf);

  popup_log ();
}


static void
handle_html (void)
{
  if (!bess) {
    ERptStatusMessage(1, 1, "no topology");
    return;
  }
  html_topo ();
}



static void
handle_print (void)
{
  if (!bess) {
    ERptStatusMessage(1, 1, "no topology");
    return;
  }
  print_topo ();
}


static double default_h = 1;

static void
do_bess_to_adlib (void)
{
  char buf[L_tmpnam] = "./tmp";
  FILE *fp;
  char cmd[1024];
  
  tmpnam (buf);
  fp = fopen (buf, "w");
  if (fp == NULL) {
    ERptErrMessage (1, 1, "Failed to print the data", ERROR_GRADE);
    return;
  }

  bess_to_adlib (bess, fp, default_h);

  fclose (fp);
  sprintf (cmd, "netscape -remote 'openURL(file:%s)'", buf);
  system (cmd);
}



static void
handle_bess_to_adlib (void)
{
  if (!bess) {
    ERptStatusMessage(1, 1, "no topology");
    return;
  }
  do_bess_to_adlib ();
}



static void
do_bess_to_vdt (void)
{
  char buf[L_tmpnam] = "./tmp";
  FILE *fp;
  char cmd[1024];
  
  tmpnam (buf);
  fp = fopen (buf, "w");
  if (fp == NULL) {
    ERptErrMessage (1, 1, "Failed to print the data", ERROR_GRADE);
    return;
  }

  bess_to_vdt (bess, fp);

  fclose (fp);
  sprintf (cmd, "netscape -remote 'openURL(file:%s)'", buf);
  system (cmd);
}



static void
handle_bess_to_vdt (void)
{
  if (!bess) {
    ERptStatusMessage(1, 1, "no topology");
    return;
  }
  do_bess_to_vdt ();
}


static void
do_bess_to_wasp (void)
{
  char *name=0;
  FILE *fp=0;

  if (TypeInGetTokensTotal () >= 2) {
    name = TypeInGetTokenAsString (2);
  } else {
    fp = ESISelFile ("Select output file", "OK", "CANCEL", NULL, NULL, "w", &name);
    if (fp) {
      fclose (fp);
    }
  }

  if (name) {
    fp = fopen (name, "w");
    if (fp == NULL) {
      ERptErrMessage (1, 1, "Failed to print the data", ERROR_GRADE);
      return;
    }  
  } 

  bess_to_wasp (bess, fp);

  fclose (fp);
}


static void
handle_bess_to_wasp (void)
{
  if (!bess) {
    ERptStatusMessage(1, 1, "no topology");
    return;
  }
  do_bess_to_wasp ();
}

static void
handle_default_h (void)
{
  if (TypeInGetTokensTotal () >= 2 && TypeInGetTokenType (2) == NUMBER)
    default_h = TypeInGetTokenAsDouble (2);
  default_h = fabs (default_h);
  if (default_h <= 0) default_h = 1;
}





static EModel *tshow = NULL;

static void
set_model (void)
{
  int   kp;

  if (tshow == NULL) tshow = ESINamedModel ("tshow");
  
  if ((kp = TypeInIsKeywordInLine (1, "MODEL")) > 0)  {
    if (TypeInGetTokenType (kp+1) == STRING) {
      tshow = ESINamedModel (TypeInGetTokenAsString (kp+1));
    }
  }
}

static void
handle_show (void)
{
  int kp;
#define COLLECT_IDS(dict, startat)                                              \
  {                                                                             \
    int _n_, _i_;                                                               \
    if (!dict) dict = dict_create_dictionary ();                                \
    for (_i_ = startat, _n_ = TypeInGetTokensTotal (); _i_ <= _n_; _i_++) {     \
      if (TypeInGetTokenType (_i_) == NUMBER) {                                 \
        int n = TypeInGetTokenAsInt (_i_);                                      \
        dict_add_obj_int_name (dict, (void *)(abs(n)+1), n);                    \
      }                                                                         \
    }                                                                           \
  }
#define GETARGS()                                                               \
  {                                                                             \
    int kp_;                                                                    \
    kp_ = TypeInIsKeywordInLine (2, "INCLUDE");                                 \
    if (kp_ > 0) {                                                              \
      COLLECT_IDS (to_include, kp_ + 1);                                        \
    }                                                                           \
    kp_ = TypeInIsKeywordInLine (2, "EXCLUDE");                                 \
    if (kp_ > 0) {                                                              \
      COLLECT_IDS (to_exclude, kp_ + 1);                                        \
    }                                                                           \
    kp_ = TypeInIsKeywordInLine (2, "CRIT");                                    \
    if (kp_ > 0) {                                                              \
      NEWP_DUPL_STRING (crit, TypeInGetTokenAsString (kp_ + 1), { return; });   \
    }                                                                           \
  }
  dict_dictionary to_include = NULL;
  dict_dictionary to_exclude = NULL;
  char *crit = NULL;
  
  set_model ();

  if ((kp = TypeInIsKeywordInLine (1, "TV")) > 0)  {
    GETARGS();
    if (bess) show_tv (bess, to_include, to_exclude, crit, tshow);
    else      ERptErrMessage(1, 2, "No topo.", ERROR_GRADE);
  }
  
  if ((kp = TypeInIsKeywordInLine (1, "TV_NUMS")) > 0)  {
    GETARGS();
    if (bess) show_tv_nums (bess, to_include, to_exclude, tshow);
    else      ERptErrMessage(1, 2, "No mesh.", ERROR_GRADE);      
  }
  
  if ((kp = TypeInIsKeywordInLine (1, "TE_NUMS")) > 0)  {
    GETARGS();
    if (bess) show_te_nums (bess, to_include, to_exclude, tshow);
    else      ERptErrMessage(1, 2, "No mesh.", ERROR_GRADE);      
  }
  
  if ((kp = TypeInIsKeywordInLine (1, "TE")) > 0)  {
    GETARGS();
    if (bess) show_te (bess, to_include, to_exclude, crit, tshow);
    else      ERptErrMessage(1, 2, "No topo.", ERROR_GRADE);
  }
  
  if ((kp = TypeInIsKeywordInLine (1, "TL")) > 0)  {
    GETARGS();
    if (bess) show_tl (bess, to_include, to_exclude, crit, tshow);
    else      ERptErrMessage(1, 2, "No topo.", ERROR_GRADE);
  }
  
  if ((kp = TypeInIsKeywordInLine (1, "TF")) > 0)  {
    GETARGS();
    if (bess) show_tf (bess, to_include, to_exclude, crit, tshow);
    else      ERptErrMessage(1, 2, "No topo.", ERROR_GRADE);      
  }
  
  if ((kp = TypeInIsKeywordInLine (1, "TF_ORIENT")) > 0)  {
    GETARGS();
    if (bess) show_tf_orient (bess, to_include, to_exclude, crit, tshow);
    else      ERptErrMessage(1, 2, "No topo.", ERROR_GRADE);      
  }
  
  if ((kp = TypeInIsKeywordInLine (1, "TF_ORIENT")) > 0)  {
    GETARGS();
    if (bess) show_tf_orient (bess, to_include, to_exclude, crit, tshow);
    else      ERptErrMessage(1, 2, "No topo.", ERROR_GRADE);      
  }
  
  if ((kp = TypeInIsKeywordInLine (1, "TS")) > 0)  {
    GETARGS();
    if (bess) show_ts (bess, to_include, to_exclude, crit, tshow);
    else      ERptErrMessage(1, 2, "No topo.", ERROR_GRADE);      
  }
  
  if ((kp = TypeInIsKeywordInLine (1, "TS_ORIENT")) > 0)  {
    GETARGS();
    if (bess) show_ts_orient (bess, to_include, to_exclude, crit, tshow);
    else      ERptErrMessage(1, 2, "No topo.", ERROR_GRADE);      
  }
  
  if ((kp = TypeInIsKeywordInLine (1, "MF")) > 0)  {
    if (bess) show_mf (bess, tshow);
    else      ERptErrMessage(1, 2, "No mesh.", ERROR_GRADE);
  }
  
  if ((kp = TypeInIsKeywordInLine (1, "MF_INTERFERENCE")) > 0)  {
    if (bess) show_mf_interference (bess, tshow, NULL);
    else      ERptErrMessage(1, 2, "No mesh.", ERROR_GRADE);
  }
  
  if ((kp = TypeInIsKeywordInLine (1, "ME")) > 0)  {
    if (bess) show_me (bess, tshow);
    else      ERptErrMessage(1, 2, "No mesh.", ERROR_GRADE);
  }
  
  if ((kp = TypeInIsKeywordInLine (1, "MV")) > 0)  {
    if (bess) show_mv (bess, tshow);
    else      ERptErrMessage(1, 2, "No mesh.", ERROR_GRADE);
  }
  
  if ((kp = TypeInIsKeywordInLine (1, "MF_NUMS")) > 0)  {
    if (bess) show_mf_nums (bess, tshow);
    else      ERptErrMessage(1, 2, "No mesh.", ERROR_GRADE);      
  }
  
  if ((kp = TypeInIsKeywordInLine (1, "ME_NUMS")) > 0)  {
    if (bess) show_me_nums (bess, tshow);
    else      ERptErrMessage(1, 2, "No mesh.", ERROR_GRADE);      
  }
  
  if ((kp = TypeInIsKeywordInLine (1, "MV_NUMS")) > 0)  {
    if (bess) show_mv_nums (bess, tshow);
    else      ERptErrMessage(1, 2, "No mesh.", ERROR_GRADE);      
  }

  if (crit) NEWP_FREE_OBJ (crit);
  if (to_include) dict_destroy_dictionary (to_include);
  if (to_exclude) dict_destroy_dictionary (to_exclude);
}


static void
handle_clear (void)
{
  set_model ();

  EMDestroyAllGraphics (tshow);
}



#define DEFCMDNAME(name) #name, sizeof (#name)

static top_command ctable[] = {
  {
    DEFCMDNAME (show),
    handle_show,
    " (show part of topological and geometrical model)"
  },
  {
    DEFCMDNAME (clear),
    handle_clear,
    " (clear the visualizations of the topological model)"
  },
  {
    DEFCMDNAME (run),
    handle_run,
    " (generate surface mesh from a tet mesh)"
  },
  {
    DEFCMDNAME (html),
    handle_html,
    " (generate HTML topology documentation)"
  },
  {
    DEFCMDNAME (print),
    handle_print,
    " (generate printed topology documentation)"
  },
  {
    DEFCMDNAME (crease_angle),
    handle_crease_angle,
    " (set limit crease angle)"
  },
  {
    DEFCMDNAME (corner_angle),
    handle_corner_angle,
    "# (set corner angle)"
  },
  {
    DEFCMDNAME (default_h),
    handle_default_h,
    "# (set default mesh size)"
  },
  {
    DEFCMDNAME (bess_to_adlib),
    handle_bess_to_adlib,
    "# (write topology in adlib format)"
  },
  {
    DEFCMDNAME (bess_to_vdt),
    handle_bess_to_vdt,
    "(write vdt input)"
  },
  {
    DEFCMDNAME (bess_to_wasp),
    handle_bess_to_wasp,
    "(write WASP input)"
  },
};


void
ESICustomize (Widget parent_pane)
{
  TypeInAugmentCommandTable (ctable, XtNumber (ctable));
}




