#include "bess.h"
#include "bessstubs.h"
#include "tetmesh.h"
#include "bess.pt"
#include "tetin.h"
#include "g3dbox.h"
#include "pirat.h"

static int
locate_node (MESHP st, pirat_front_end_t *pirat, WCRec *loc, double tol, int *maxid)
{
  int id = *maxid;
  G3D_box_t range;
  G3D_INIT_BOX (range, (*loc));
  G3D_INFLATE_BOX (range, tol); 
  if (!pirat_search_range (pirat, &range, 1)) {
    fprintf (stderr, "Error: Failed to search in %s (l.%d)\n",
             __FILE__, __LINE__);
    exit (1);
  }
  if (pirat->num_of_points_found > 0) {
    id = (int) pirat_first_in_range (pirat);
    /*    fprintf (stderr, "Found node %d\n", id); */
  } else {
    id = *maxid;
    pirat_add_pt (pirat, (void *)id, loc);
    /*    fprintf (stderr, "Adding node %d at %g %g %g\n", id, loc->x, loc->y, loc->z); */
    if (!tet_add_node (st, id, loc))
          fprintf (stderr, "Error adding node %d at %g %g %g in %s (l.%d)\n",
                   id, loc->x, loc->y, loc->z, __FILE__, __LINE__);
    (*maxid)++;
  }    
  return id;
}


typedef struct {
  MESHP st;
  pirat_front_end_t *pirat;
  int maxnid;
  int maxtid;
} tetin_context_t;

static int
add_tet (NODE vctx, NODE vgo)
{
  tetin_context_t *ctx = (tetin_context_t *)vctx;
  GraphicObj *go = (GraphicObj *)vgo;
  if ((EGGraphicType (go) == EG_TETRAHEDRON) ||
      (EGGraphicType (go) == EG_TETRAHEDRONWD) ||
      (EGGraphicType (go) == EG_TET10)) {
#define SAME(i1,i2,tol) ((fabs(pts[i1].x-pts[i2].x) < tol)&&(fabs(pts[i1].y-pts[i2].y) < tol)&&(fabs(pts[i1].z-pts[i2].z) < tol))
    int mvids[10] = {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1};
    int np;
    int j;
    int regid;
    double tol;
    WCRec pts[10];
    G3D_box_t box;
    if (!EGGetGraphicsGeometry(go, pts, &np)) {
      ERptStatusMessage(1, 1, "EGGetGraphicsGeometry failed");
    }
    G3D_INIT_BOX (box, pts[0]); G3D_UPDT_BOX (box, pts[1]);
    G3D_UPDT_BOX (box, pts[2]); G3D_UPDT_BOX (box, pts[3]);
    tol=(G3D_BOX_RANGE (box, x) +
         G3D_BOX_RANGE (box, y) +
         G3D_BOX_RANGE (box, z)) / 10000;
    if (!( /* skip if the tetrahedron is denegerated */
          SAME(0,1,tol)||SAME(0,2,tol)||SAME(0,3,tol)||
          SAME(1,2,tol)||SAME(1,3,tol)||
          SAME(2,3,tol))) {
      for (j = 0; j < 4; j++) {
        mvids[j] = locate_node (ctx->st, ctx->pirat, &pts[j],
                                tol,
                                &ctx->maxnid);
      }
      EGCopyAttributesToActiveSet (go);
      regid = EASValsGetLayer ();
      /*
        fprintf (stderr, "Adding tet %d: nodes %d %d %d %d (region %d)\n",
        ctx->maxtid,
        mvids[0], 
        mvids[1], 
        mvids[2], 
        mvids[3], 
        regid);
      */
      if (!tet_add_tet (ctx->st, ctx->maxtid, mvids, regid))
        return FALSE;
      ctx->maxtid++;
    }
  }
  return 1;
}

int
tetin_read (char *file, MESHP st, EModel *model)
{
  tetin_context_t ctx;

  OnDiskSetFileName(file);
  EMRetrieveAllGraphics (model);

  ctx.st = st;
  ctx.pirat = pirat_new_pirat (10, 1e-6);
  ctx.maxnid = 1;
  ctx.maxtid = 1;
  EMDispatchToGraphicObjects (model, add_tet, (NODE)&ctx);
  pirat_destroy_pirat (ctx.pirat);

  return TRUE;
}
