#include <stdio.h>
#include <stdlib.h>

typedef struct loc_t {
  double x, y, z;
}              loc_t;

static int
write_hexvue_hex_ascii (FILE *s,
                        loc_t *p[8],
                        int num_of_values,
                        double *values[8]
                        )
{
  int j, i, n = num_of_values;
  
  fprintf (s, "HV\n");
  for (j = 0; j < 8; j++) {
    fprintf (s, "%g %g %g\n", p[j]->x, p[j]->y, p[j]->z);
    for (i = 0; i < n; i++) {
      fprintf (s, "%g\n", values[j][i]);
    }      
  }
  return 1;
}

static double 
radius_squared_func (double x, double y, double z)
{
  return (x*x + y*y + z*z);
}

typedef struct lump_t {
  loc_t  center;
  double radius;
}              lump_t;


#define NLUMPS 4
#if NLUMPS == 2
static lump_t lumps[] = {
    {
      { 0.4, 0.4, 0.4 },
      0.1
    },
    {
      { 0.6, 0.6, 0.6 },
      0.1
    },
  };
#elif NLUMPS == 3
static lump_t lumps[] = {
    {
      { 0.5, 0.4, 0.3 },
      0.1
    },
    {
      { 0.3, 0.1, 0.4 },
      0.1
    },
    {
      { 0.1, 0.6, 0.7 },
      0.1
    },
  };
#elif NLUMPS == 4
static lump_t lumps[] = {
    {
      { 0.5, 0.4, 0.3 },
      0.1
    },
    {
      { 0.3, 0.1, 0.4 },
      0.1
    },
    {
      { 0.1, 0.1, 0. },
      0.1
    },
    {
      { 0.1, 0.6, 0.7 },
      0.1
    },
  };
#else
#error NLUMPS not defined
#endif

static double 
n_lumps (double x, double y, double z)
{
  double rx, ry, rz, r2;
  int i, n;
  double val = 0;
  n = sizeof (lumps) / sizeof (lumps[0]);
  for (i = 0; i < n; i++) {
    rx = x - lumps[i].center.x;
    ry = y - lumps[i].center.y;
    rz = z - lumps[i].center.z;
    r2 = (rx*rx + ry*ry + rz*rz) / (lumps[i].radius * lumps[i].radius);
    val += 1 / (1 + r2);
  }
  return val;
}

#define FUNC n_lumps

int
main (int argc, char **argv)
{
  const int nint = 20;
  int i, j, k, m;
  double x, y, z;
  double d = 1.0 / (double) nint;
  double v[8];
  double *vp[8];
  loc_t p[8];
  loc_t *pp[8];
  FILE *s = stdout;

  for (i = 0; i < 8; i++) {
    pp[i] = &p[i];
    vp[i] = &v[i];
  }
  
  fprintf (s, "! HEXVUE file written by %s. (C) Petr Krysl.\n", argv[0]);
  fprintf (s, "Visualization of scalar function on the unit cube\n");
  fprintf (s, "1 ! 1 value per vertex\n");
  fprintf (s, "\"f_value\" ! function value\n");
  
  x = 0;
  for (i = 0; i < nint; i++) {
    y = 0;
    for (j = 0; j < nint; j++) {
      z = 0;
      for (k = 0; k < nint; k++) {
        p[0].x = x  ; p[0].y = y  ; p[0].z = z  ;
        p[1].x = x+d; p[1].y = y  ; p[1].z = z  ;
        p[2].x = x+d; p[2].y = y+d; p[2].z = z  ;
        p[3].x = x  ; p[3].y = y+d; p[3].z = z  ;
        p[4].x = x  ; p[4].y = y  ; p[4].z = z+d;
        p[5].x = x+d; p[5].y = y  ; p[5].z = z+d;
        p[6].x = x+d; p[6].y = y+d; p[6].z = z+d;
        p[7].x = x  ; p[7].y = y+d; p[7].z = z+d;
        for (m = 0; m < 8; m++)
          v[m] = FUNC (p[m].x, p[m].y, p[m].z);
        write_hexvue_hex_ascii (stdout, pp, 1, vp);
        z += d;
      }
      y += d;
    }
    x += d;
  }
  fprintf (s, "END\n");
}

